local __ = require '__'
local log = require 'log'
local http_parser = require 'http_parser'
local class = require 'class'
local sio = require 'sio'
local net = require 'net_ffi'
local ffi = require 'ffi'
local C = ffi.C
local ev = require 'ev'

ffi.cdef[[
    in_addr_t inet_addr(const char *cp);
    char *inet_ntoa(struct in_addr in);
]]

local m = class()

function m:init(settings)
    if settings.data ~= nil then
        self.camera = settings.data
    else
        self.camera = {
            debug = false,
            state = 'VIDEO_SHUTDOWN',
            socket = -1,
            frame_count = 0,
            frame_rate = 0,
            frame_len = 0,
            name = settings.vendor .. '-' ..  settings.model .. '-' ..
                settings.serial:sub(7),
            vendor = settings.vendor,
            model = settings.model,
            serial = settings.serial,
            firmware = settings.firmware,
            spec_frame_rate = '10',
            spec_image_size = '320x240',
            spec_image_compression = '30',
            host = settings.host,
            port = settings.port,
            channel_write = nil,
        }
    end
end

function m:setup(data, model, serial, host, firmware, port)
    if type(data) == 'table' then
        return m:new(self, { data = data })
    end
    return m:new(self, {
        vendor = data,
        model = model,
        serial = serial,
        host = host,
        firmware = firmware,
        port = port or '80'
    })
end

function m:start(codec)
    if self.camera.state ~= 'VIDEO_SHUTDOWN' then
        self:stop()
    end
    self:state('VIDEO_CONNECT')
--[[
    if not pthread:create(video[codec], self.camera) then
        return false
    end
]]
    local wait = 0
    while self.camera.state ~= 'VIDEO_STREAMING' do
        if wait == 100 then
            self:stop()
            return false
        end
        wait = wait + 1
        C.usleep(10000)
    end
    return true
end

function m:stop()
    self:state('VIDEO_SHUTDOWN')
    local wait = 0
    while self.camera.state ~= 'VIDEO_SHUTDOWN' do
        if wait == 100 then
            break
        end
        wait = wait + 1
        C.usleep(10000)
    end
    self.camera.frame_count = 0
    self.camera.frame_rate = 0
    self.camera.frame_len = 0
    self:disconnect()
--    pthread:join()
end

function m:api_write(x, len)
    if not x and not len then
        self.camera.api_write = nil
    elseif x and not len then
        self.camera.api_write = x
    elseif self.camera.api_write then
        self.camera.api_write(x, len)
    end
end

function m:channel_write(x, len)
    if not x and not len then
        self.camera.channel_write = nil
    elseif x and not len then
        self.camera.channel_write = x
    elseif self.camera.channel_write then
        self.camera.channel_write(x, len)
    end
end

function m:data()
    return self.camera
end

function m:debug(x)
    if x then
        self.camera.debug = x
    else
        return self.camera.debug
    end
end

function m:state(x)
    if x then
        self.camera.state = x
    else
        return self.camera.state
    end
end

function m:socket(x)
    if x then
        self.camera.socket = x
    else
        return self.camera.socket
    end
end

function m:frame_length(x)
    if x then
        self.camera.frame_len = x
    else
        return self.camera.frame_len
    end
end

function m:frame_count(x)
    if x then
        self.camera.frame_count = x
    else
        return self.camera.frame_count
    end
end

function m:frame_rate(x)
    if x then
        self.camera.frame_rate = x
    else
        return self.camera.frame_rate
    end
end

function m:host(x)
    if x then
        self.camera.host = x
    else
        return self.camera.host
    end
end

function m:port(x)
    if x then
        self.camera.port = x
    else
        return self.camera.port
    end
end

function m:name(x)
    if x then
        self.camera.name = x
    else
        return self.camera.name
    end
end

function m:vendor(x)
    if x then
        self.camera.vendor = x
    else
        return self.camera.vendor
    end
end

function m:model(x)
    if x then
        self.camera.model = x
    else
        return self.camera.model
    end
end

function m:serial(x)
    if x then
        self.camera.serial = x
    else
        return self.camera.serial
    end
end

function m:firmware(x)
    if x then
        self.camera.firmware = x
    else
        return self.camera.firmware
    end
end

function m:page()
    return self.camera.page
end

function m:query()
    return self.camera.query
end

function m:spec_frame_rate(x)
    if x then
        self.camera.spec_frame_rate = x
    else
        return self.camera.spec_frame_rate
    end
end

function m:spec_image_size(x)
    if x then
        self.camera.spec_image_size = x
    else
        return self.camera.spec_image_size,
            self.camera.spec_image_size:match('(%d+)x(%d+)')
    end
end

function m:spec_image_compression(x)
    if x then
        self.camera.spec_image_compression = x
    else
        return self.camera.spec_image_compression,
            tonumber(self.camera.spec_image_compression) ~= 0 and
            tonumber(self.camera.spec_image_compression) or 1
    end
end

function m:lock()
end

function m:unlock()
end

function m:update()
    return true
end

--[[

function m:write(sd, buf, len)
    local n = -1
    local sent = 0
    local data = ffi.cast('char *', buf)
    while sent < len do
        n = C.send(sd, data + sent, len - sent, 0)
        if n == -1 then
            log:error(__"camera:write(): {0} on {1}",
                      ffi.string(C.strerror(ffi.errno())), self:name())
            break
        end
        sent = sent + n
    end
    return n
end

function m:request()
    local n = ffi.new('int', -1)
    local rc = ffi.new('int', -1)
    local sent = ffi.new('int', 0)
    local query = "GET /" .. self:page() .. self:query() .. " HTTP/1.0\r\n" ..
                  "Host: " .. self:host() .. "\r\n" ..
                  "User-Agent: HBM CX23R\r\n\r\n"
    local sbuf = ffi.new('char [?]', #query + 1)
    ffi.copy(sbuf, query, #query)
    while sent < #query do
        local wfds = ffi.new('fd_set [1]')
        local tv = ffi.new('struct timeval [1]')
        sio.FD_ZERO(wfds)
        sio.FD_SET(self.camera.socket, wfds)
        tv[0].tv_sec = ffi.new('long', 1)
        tv[0].tv_usec = 0
        rc = C.select(self.camera.socket + 1, nil, wfds, nil, tv)
        if rc > 0 then
            n = C.send(self.camera.socket, sbuf + sent, #query - sent, 0)
            if n == -1 then
                io.stderr:write(self:name() .. ":request(): send " ..
                                ffi.string(C.strerror(ffi.errno())) .. "\n")
                break
            end
            sent = sent + n
        elseif rc == -1 then
            io.stderr:write(self:name() .. ":request(): select " ..
                            ffi.string(C.strerror(ffi.errno())) .. "\n")
            n = -1
            break
        else
            io.stderr:write(self:name() .. ":request(): timeout\n")
            n = -1
            break
        end
    end
    return n
end

function m:connect()
    local hints = ffi.new('struct addrinfo [1]')
    local res = ffi.new('struct addrinfo *[1]')
    local ressave = ffi.new('struct addrinfo *[1]')
    hints[0].ai_family = C.AF_UNSPEC
    hints[0].ai_socktype = C.SOCK_STREAM
    local rc = C.getaddrinfo(self.camera.host, self.camera.port,
                             hints, res)
    if rc ~= 0 then
        io.stderr:write(self:name() .. ":connect(): " ..
                 ffi.string(C.strerror(ffi.errno())) .. "\n")
        return sd
    end
    ressave = res
    while res ~= nil do
        self.camera.socket = C.socket(res[0].ai_family, res[0].ai_socktype,
                                         res[0].ai_protocol)
        if self.camera.socket >= 0 then
            rc = C.connect(self.camera.socket, res[0].ai_addr,
                           res[0].ai_addrlen)
            if rc ~= 0 then
                io.stderr:write(self:name() .. ":connect(): " ..
                         ffi.string(C.strerror(ffi.errno())) .. "\n")
                C.close(self.camera.socket)
                self.camera.socket = -1
            else
                -- socket is connected
                break
            end
        end
        res = res[0].ai_next
    end
    C.freeaddrinfo(ressave[0])
    return self.camera.socket
end

function m:disconnect()
    net.shutdown(self.camera.socket)
    local o = { camera = ffi.new('camera_data_t [1]') }
    ffi.copy(o.camera, self.camera, ffi.sizeof(o.camera))
    self.camera = nil
    collectgarbage()
    self.camera = o.camera
end
]]

function m:upnp(loop, timeout, cb)
    local host = ffi.new('char [16]', '239.255.255.250')
	local rc = ffi.new('int')
    local addr = ffi.new('struct sockaddr_in [1]')
    -- setup socket
	local sd = ffi.new('int')
    sd = C.socket(C.AF_INET, C.SOCK_DGRAM, 0)
    if sd < 0 then
        log:log(log.ERR, string.format("camera:upnp(): %s",
                 ffi.string(C.strerror(ffi.errno()))))
        cb{}
        return
    end
    ffi.fill(addr, ffi.sizeof('struct sockaddr_in'), 0)
    addr[0].sin_family = C.AF_INET
    addr[0].sin_port = C.htons(ffi.new('uint16_t', 1900))
    addr[0].sin_addr.s_addr = C.inet_addr(host)
    -- build query
    local msearch = 'M-SEARCH * HTTP/1.1\r\n' ..
                    'HOST: ' .. ffi.string(host) .. ':1900\r\n' ..
                    'ST: "ssdp:all"\r\n' ..
                    'MAN: "ssdp:discover"\r\n' ..
                    'MX: 3\r\n\r\n'
	local query = ffi.new('unsigned char [?]', #msearch)
    ffi.copy(query, msearch, #msearch)

    ev.Timer.new(function(l, t, e)
            local info = {}
            local slen = ffi.new('int[1]')
            local buf = ffi.new('char[8192]')
            repeat
                slen[0] = ffi.sizeof('struct sockaddr_in')
                local amt = C.recvfrom(sd, buf, ffi.sizeof(buf),
                    C.MSG_DONTWAIT, ffi.cast('struct sockaddr *', addr), slen)
                if amt > 0 then
                    info[#info + 1] = {
                        type = 'vivotek',
                        ip = ffi.string(C.inet_ntoa(addr[0].sin_addr)),
                        serial =  'blah',
                    }
                end
            until amt <= 0
            C.close(sd)
            cb(info)
        end, timeout, 0):start(loop)

    -- send query to destination address
    rc = C.sendto(sd, query, ffi.sizeof(query), 0,
                  ffi.cast('struct sockaddr *', addr), ffi.sizeof(addr))
    if rc < 0 then
        log:log(log.ERR, string.format("camera:upnp(): sendto %s",
                 ffi.string(C.strerror(ffi.errno()))))
        C.close(sd)
        cb{}
        return
    end
end

function m:probe(a, b, c)
    self:upnp(a, b, c)
end

--[[
function m:info_old(ip, query, rbuf)
    local port
    local host
    if self.camera then
        port = ffi.cast('char *', self.camera.port)
        host = ffi.cast('char *', self.camera.host)
    else
        port = ffi.new('char [3]', '80')
        host = ffi.new('char [?]', #ip + 1)
        ffi.copy(host, ip, #ip)
    end
    local rc = ffi.new('int')
    local sd = ffi.new('int')
    -- connect
    local hints = ffi.new('struct addrinfo [1]')
    local res = ffi.new('struct addrinfo *[1]')
    local ressave = ffi.new('struct addrinfo *[1]')
    hints[0].ai_family = C.AF_UNSPEC
    hints[0].ai_socktype = C.SOCK_STREAM
    rc = C.getaddrinfo(host, port, hints, res)
    if rc ~= 0 then
        io.stderr:write("camera:info(): getaddrinfo " ..
                 ffi.string(C.strerror(ffi.errno())) .. "\n")
        return false
    end
    ressave = res
    while res ~= nil do
        sd = C.socket(res[0].ai_family, res[0].ai_socktype,
                                         res[0].ai_protocol)
        if sd >= 0 then
            rc = C.connect(sd, res[0].ai_addr,
                           res[0].ai_addrlen)
            if rc ~= 0 then
                io.stdout:write("camera:info(): connect " ..
                         ffi.string(C.strerror(ffi.errno())) .. "\n")
                net.shutdown(sd)
                sd = -1
            else
                -- socket is connected
                break
            end
        end
        res = res[0].ai_next
    end
    C.freeaddrinfo(ressave[0])
    if sd < 0 then
        return false
    end
    -- send
    local n = ffi.new('int')
    local sent = ffi.new('int')
    local rfds = ffi.new('fd_set [1]')
    local wfds = ffi.new('fd_set [1]')
    local tv = ffi.new('struct timeval [1]')
    local sbuf = ffi.new('char [?]', #query + 1)
    ffi.copy(sbuf, query, #query)
    while sent < #query do
        sio.FD_ZERO(wfds)
        sio.FD_SET(sd, wfds)
        tv[0].tv_sec = ffi.new('long', 1)
        tv[0].tv_usec = 0
        rc = C.select(sd + 1, nil, wfds, nil, tv)
        if rc > 0 then
            n = C.send(sd, sbuf + sent, #query - sent, 0)
            if n == -1 then
                io.stderr:write("camera:info(): send " ..
                                ffi.string(C.strerror(ffi.errno())) .. "\n")
                net.shutdown(sd)
                return false
            end
            sent = sent + n
        elseif rc == -1 then
            io.stderr:write("camera:info(): select " ..
                            ffi.string(C.strerror(ffi.errno())) .. "\n")
            net.shutdown(sd)
            return false
        else
            io.stderr:write("camera:info(): send timeout\n")
            net.shutdown(sd)
            return false
        end
    end
    -- recv
    n = 0
    local http = ffi.new('char *')
    local http_response = {
        version = ffi.new('char [4]'),
        code = ffi.new('char [4]'),
        reason = ffi.new('char [33]')
    }
    repeat
        sio.FD_ZERO(rfds)
        sio.FD_SET(sd, rfds)
        tv[0].tv_sec = ffi.new('long', 3)
        tv[0].tv_usec = ffi.new('long', 0)
        rc = C.select(sd + 1, rfds, nil, nil, tv)
        if rc > 0 then
            n = C.recv(sd, rbuf + n, ffi.sizeof(rbuf) - n, 0)
            if n < 0 then
                io.stderr:write("camera:info(): recv " ..
                         ffi.string(C.strerror(ffi.errno())) .. "\n")
                break
            end
            if http == nil then
                http = C.strstr(rbuf, "HTTP/")
                if http ~= nil then
                    local ptr = C.strstr(http, "\r\n")
                    if ptr ~= nil then
                        ptr[0] = 0
                        C.sscanf(http, "HTTP/%s %s %32[^\n]s",
                                        http_response.version,
                                        http_response.code,
                                        http_response.reason)
                        ptr[0] = 13
                        if ffi.string(http_response.code) ~= '200' then
                            io.stderr:write("video: HTTP/" ..
                                ffi.string(http_response.version) .. " " ..
                                ffi.string(http_response.code) .. " " ..
                                ffi.string(http_response.reason) .. "\n")
                            net.shutdown(sd)
                            return false
                        end
                    end
                end
            end
        elseif rc == -1 then
            io.stderr:write("camera:info(): select " ..
                 ffi.string(C.strerror(ffi.errno())) .. "\n")
            break
        else
            io.stderr:write("camera:info(): recv timeout\n")
            break
        end
    until n <= 0
    net.shutdown(sd)
    return true
end
]]

function m:http_request(http_info, callback, data)
    local s = net.socket{
        ip = http_info.host,
        port = http_info.port or 80,
        family = C.AF_INET,
        type = C.SOCK_STREAM,
        flags = 0}

    local ok, socket = pcall(s.connect, s)
    if not ok then
        log:log(log.ERR, string.format("Can not connect to %s on port %s: %s",
            http_info.host, http_info.port or 80, socket))
        callback()
        return
    end
    local sbuf = net.StreamBuffer:new(socket)
    sbuf.read_cb = function(loop)
        local avail, buf = sbuf:read()
        if avail then
            local response, endpos =
                http_parser.response:match(ffi.string(buf, avail))
            if response then
                if response.response_code and
                    response.response_code == '200' then
                    local rh = response.headers
                    if rh and rh.content_length then
                        local cl = tonumber(rh.content_length)
                        if cl + endpos > avail + 1 then return 0 end
                        sbuf.read_cb = nil
                        callback(ffi.string(buf + endpos -1, cl))
                    elseif avail > (endpos - 1) then
                        callback(ffi.string(buf+endpos-1, avail-endpos-1))
                    else
                        return 0
                    end
                    sbuf:close()
                    return avail
                else
                    callback()
                    sbuf:close()
                end
            elseif ffi.string(buf, avail):find('HTTP/1%.[01] 200 OK') then
                callback(true)
                sbuf:close()
                return avail
            end
        end
        return 0
    end
    sbuf:start(ev.Loop.default)
    local req
    if http_info.method and http_info.method == 'POST' then
        req = string.format('%s %s HTTP/1.1\r\n' ..
            'Content-Length: %d\r\n' ..
            'Content-Type:  application/json\r\n\r\n' ..
            '%s\r\n',
            http_info.method, http_info.absolute_path,
            #data + 2, data)
    elseif http_info.method and http_info.method ~= 'GET' then
        req = string.format('%s %s HTTP/1.1\r\n\r\n%s\r\n',
            http_info.method, http_info.absolute_path, data)
    else
        if http_info.extra_header then
            req = string.format('GET %s HTTP/1.1\r\n%s\r\n',
                http_info.absolute_path, http_info.extra_header)
        else
            req = string.format('GET %s HTTP/1.1\r\n\r\n',
                http_info.absolute_path)
        end
    end
    sbuf:write(req)
end

function m:snapshot(proxy, file)
    log:notice(__"camera:snapshot unimplemented for {0}", proxy.proxy_hostname)
end

return m
