#!/usr/bin/luajit

local log = require 'log'

-- Defer everything until network is connected
local nf = require 'net_functions'
nf.wait_carrier('any', 'ignore', {'eth1', 'eth2'})

local ev = require 'ev'
local dbus = require 'dbus'
local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()
local master = assert(bus:auto_proxy('com.somat.armada.Master',
                                     '/com/somat/armada/Master'))
local net = require 'net_ffi'
local ffi = require 'ffi'
local C = ffi.C

ffi.cdef([[
int sscanf(const char *str, const char *format, ...);
char *strstr(const char *haystack, const char *needle);
struct _IO_FILE;
typedef struct _IO_FILE FILE;
FILE *popen(const char *command, const char *type);
int pclose(FILE *stream);
FILE *fopen(const char *path, const char *mode);
int fclose(FILE *fp);
char *fgets(char *s, int size, FILE *stream);
void rewind(FILE *stream);
typedef int32_t pid_t;
int kill(pid_t, int);
/* Pinger stuff */
struct iphdr {
    uint8_t       ihl_version;
    uint8_t       tos;
    uint16_t      tot_len;
    uint16_t      id;
    uint16_t      frag_off;
    uint8_t       ttl;
    uint8_t       protocol;
    uint16_t      check;
    uint32_t      saddr;
    uint32_t      daddr;
} __attribute__((packed))__;

struct icmphdr {
    uint8_t       type;
    uint8_t       code;
    uint16_t      checksum;
    union {
        struct {
            uint16_t      id;
            uint16_t      sequence;
        } echo;
        uint32_t   gateway;
        struct {
            uint16_t      __unused;
            uint16_t      mtu;
        } frag;
    } un;
    uint8_t       data[56];
} __attribute__((packed))__;

enum {
    ICMP_ECHO = 8,
};
]])

local camera = require 'camera'
local vendor_list = {
    require 'axis_camera',
    require 'vivotek_camera'
}

local pid_list = {}
local master_list = {}
local ethint = {}
local arp = ffi.new('FILE *')

local update_rate = arg[1]
update_rate = tonumber(update_rate) or 15.0
if update_rate < 10.0 then
    update_rate = 10.0
end

-- Start a camera-proxy for the discovered device
local function camera_proxy_start(dev)
    local added = true
    local p = ffi.new('pid_t')
    for _, proxy in ipairs(pid_list) do
        if proxy.serial == dev.serial then
            added = false
            break
        end
    end
    if added then
        p = master:Spawn('/usr/bin/camera-proxy', {
                        'camera-proxy',
                        dev.vendor,
                        dev.model,
                        dev.serial,
                        dev.ip,
                        dev.firmware,
                        'stdout_level:none',
                    })
        table.insert(pid_list, { pid = p, ip = dev.ip, serial = dev.serial })
    end
end

-- Terminate a camera-proxy
local function camera_proxy_stop(dev)
    local pl = {}
    for _, proxy in ipairs(pid_list) do
        if proxy.serial == dev.serial then
            C.kill(proxy.pid, 3)
            C.kill(proxy.pid, 15)
            C.kill(proxy.pid, 9)
        else
            table.insert(pl, proxy)
        end
    end
    pid_list = pl
end

-- Remove unreachable devices
local function validate(m)
    if m then
        local ml = {}
        for _, dev in ipairs(m) do
            if net.Pinger:ping(dev.ip, 3, 'ethint',
                                   ev.Loop.default, ethint) then
                table.insert(ml, dev)
            else
                camera_proxy_stop(dev)
            end
        end
        if #ml > 0 then
            return ml
        end
    end
    return nil
end

-- Find existing camera-proxies and add them (startup-only)
local function camera_proxy_search()
    local fp = ffi.new('FILE *')
    local dev = ffi.new('char *')
    local pid = ffi.new('char [17]')
    local line = ffi.new('char [257]')
    local cmd = ffi.new('char [37]', 'ps | grep luajit | grep camera-proxy')
    fp = C.popen(cmd, 'r')
    if fp == nil then
        log:log(log.ERR,
            string.format("camera_proxy_search(): popen error: %s",
                          ffi.string(C.strerror(ffi.errno()))))
        return
    end
    while C.fgets(line, ffi.sizeof(line), fp) ~= nil do
        local begin = ffi.cast('char *', line)
        repeat
            if begin[0] ~= 0x20 then
                break
            end
            begin = begin + 1
        until begin[0] ~= 0x20
        C.sscanf(begin, "%16[^ ]", pid)
        dev = C.strstr(line, "camera-proxy")
        if dev ~= nil then
            local vendor = ffi.new('char [33]')
            local model = ffi.new('char [33]')
            local serial = ffi.new('char [13]')
            local ip = ffi.new('char [16]')
            local firmware = ffi.new('char [33]')
            C.sscanf(dev,
                     "camera-proxy %32[^ ] %32[^ ] %12[^ ] %15[^ ] %s",
                     vendor, model, serial, ip, firmware)
            if ffi.string(vendor) ~= '' and
                                    ffi.string(model) ~= '' and
                                    ffi.string(serial) ~= '' and
                                    ffi.string(ip) ~= '' and
                                    ffi.string(firmware) ~= '' then
                local added = true
                for _, proxy in ipairs(pid_list) do
                    if proxy.serial == ffi.string(serial) then
                        added = false
                        break
                    end
                end
                if added then
                    table.insert(pid_list, {
                                 pid = ffi.new('pid_t', tonumber(ffi.string(pid))),
                                 ip = ffi.string(ip),
                                 serial = ffi.string(serial),
                             })
                    table.insert(master_list, {
                                 ip = ffi.string(ip),
                                 serial = ffi.string(serial),
                                 vendor = ffi.string(vendor),
                                 firmware = ffi.string(firmware),
                                 model = ffi.string(model),
                             })
                end
            end
        end
    end
    master_list = validate(master_list)
    C.pclose(fp)
end

-- Cleanup terminated/dead camera-proxies
local function camera_proxy_reap()
    if #pid_list == 0 then
        return
    end
    local pl = {}
    local rc = ffi.new('int')
    for _, proxy in ipairs(pid_list) do
        if C.kill(proxy.pid, 0) < 0 then
            -- camera-proxy is dead (remove from pid_list)
        else
            -- camera-proxy is alive (add to pid_list)
            table.insert(pl, proxy)
        end
    end
    pid_list = pl
end

-- Run each camera vendor's discovery method and start 'camera-proxy'
local function camera_discovery()
    local al = {}
    local dl = {}
    local iface = ffi.new('char [7]', 'ethint')
    local ptr = ffi.new('char *')
    local line = ffi.new('char [81]')
    local ip = ffi.new('char [16]')
    local mac = ffi.new('char [18]')
    local serial = ffi.new('char [13]')
    -- mDNS/UPnP/SSDP on ethint
    for _, vendor in ipairs(vendor_list) do
        vendor:probe()
    end
    -- Check ARP entries
    C.rewind(arp)
    while C.fgets(line, ffi.sizeof(line), arp) ~= nil do
        if C.strstr(line, iface) ~= nil then
            C.sscanf(line, "%17[^ ]", ip)
            ptr = C.strstr(line, ":")
            ptr = ptr - 2
            C.sscanf(ptr, "%2s:%2s:%2s:%2s:%2s:%3[^ ]",
                     serial,
                     serial + 2,
                     serial + 4,
                     serial + 6,
                     serial + 8,
                     serial + 10)
            C.sscanf(ptr, "%18[^ ]", mac)
            if net.Pinger:ping(ffi.string(ip), 3, 'ethint',
                                   ev.Loop.default, ethint) then
                table.insert(al,{
                             ip = ffi.string(ip),
                             serial = ffi.string(serial)
                         })
            end
        end
    end
    -- Run vendor specific discovery method on ARP list
    for _, vendor in ipairs(vendor_list) do
        local list = vendor:discover(al)
        if #list > 0 then
            for _, dev in ipairs(list) do
                table.insert(dl, dev)
            end
        end
    end
    local old_list = master_list
    master_list = validate(dl)
    -- Start camera-proxy for discovered devices
    if master_list then
        if old_list then
            -- Stop camera-proxy for nonexistent devices
            for _, odev in ipairs(old_list) do
                local removed = true
                for _, dev in ipairs(master_list) do
                    if dev.serial == odev.serial then
                        removed = false
                        break
                    end
                end
                if removed then
                    camera_proxy_stop(odev)
                end
            end
        end
        for _, dev in ipairs(master_list) do
            camera_proxy_start(dev)
        end
    end
end

-- Get 'ethint' IP Zeroconf address
local function get_zcip_addr()
    local addrs = net.getifaddrs('ethint')
    for _, addr in ipairs(addrs) do
        if addr:find('^169%.254%.') then
            return addr
        end
    end
end

local ll = ev.Loop.new()
ev.Timer.new(function(l, t, e)
        ethint = get_zcip_addr()
        if ethint then
            l:unloop()
        end
    end, 1, 1):start(ll)

-- Wait here for ZCIP assigned address to show up!
ll:loop()

-- Open Linux ARP file
arp = C.fopen('/proc/net/arp', 'r')
if arp == nil then
    log:log(log.ERR,
        string.format("camera_manager: fopen error: %s",
                      ffi.string(C.strerror(ffi.errno()))))
    os.exit(1)
end

-- Check for existing 'camera-proxy' instances
camera_proxy_search()

-- Start periodic discovery
ev.Timer.new(function(loop, timer, events)
        master_list = validate(master_list)
        camera_proxy_reap()
        camera_discovery()
    end, 0.1, update_rate):start(ev.Loop.default)

-- Main Event Loop
ev.Loop.default:loop()
