local __ = require '__'
local log = require 'log'

local ev = require 'ev'
local ffi = require 'ffi'
local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_new = ffi.new
local ffi_string = ffi.string
local ts = require 'timestamps'
local u = require 'unix'
local log = require 'log'
local config = require 'system_config'

local sample_clock = require 'sample_clock'

local channel_proxy = require 'channel_proxy'
local cca_handler = require 'cca_handler'
local message_can = require 'message_can'

local can = require 'can'

local C = ffi.C

require 'strict'

local update_rate = 10

local dumper = require 'data_dumper'

local can_buf_count = 1024

ffi.cdef[[
    ssize_t write(int fd, const void *buf, size_t count);

    typedef struct __attribute__((packed)) can_sink {
        uint64_t base_timestamp;
        int socket;
        int max_count;
        int count;
        struct {
            uint32_t len;
            message_can msg[1024];
        } buf;
    } can_sink;

    char *strerror(int errnum);
]]

local out_of_space_msg = "Out of buffer space for CAN messages to be sent " ..
    "to the test engine.  This error indicates that the test engine is not " ..
    "able to keep up with reading the CAN message data.  A common cause " ..
    "for this is using a start time delay that is too short for the test.  " ..
    "Increase the start time delay for the test and try again.  This error " ..
    "causes the flagship proxy to exit and the system must be rebooted " ..
    "to restart it."

local can_sink = ffi.metatype('can_sink', { __index = {
    flush = function(self, pkt)
        if self.count <= 0 then
            return
        end
        self.buf.len = self.count * ffi.sizeof'message_can'
        local towrite = ffi.sizeof'uint32_t' + self.buf.len
        local wrote, estr, e =
            u.write(self.socket, ffi_string(self.buf, towrite))
        if not wrote then
            if e == u.EAGAIN then
                return
            elseif e ~= u.EPIPE and e ~= u.ECONNRESET then
                log:log(log.ERR,
                    string.format("Write error on CAN sink: %s", estr))
                os.exit(-5)
            end
        elseif wrote ~= towrite then
            log:log(log.ERR, string.format("Short write on CAN sink %d -> %d",
                towrite, wrote))
            os.exit(-6)
        end
        self.count = 0
    end,
    enqueue = function(self, msg)
        if self.count >= self.max_count then
            log:log(log.ERR, out_of_space_msg)
            os.exit(-7)
        elseif self.count >= (self.max_count / 2) then
            self:flush()
        end
        local dmsg = self.buf.msg + self.count
        dmsg.timestamp[0] = msg.timestamp[0] - self.base_timestamp
        ffi.copy(dmsg.value, msg.value, 13)
        self.count = self.count + 1
        -- If this is a heartbeat, flush the buffer
        -- NOTE:  This requires an FPGA change to work!
        if bit.band(dmsg.value[12], 0x20) == 0x20 then
            self:flush()
        end
    end,
}})

----------------------------------------------------------------------

local can_proxy = class('can_proxy', channel_proxy)

function can_proxy:initialize(name, devname)
    self.proxy_type = 'can'
    self.devname = devname
    self.device = string.format(
        "CAN%d", tonumber(devname:match('/dev/cca(%d+)')) + 1)
    cca_handler.new(devname, self, can,
        'CAN_' .. tonumber(devname:match('/dev/cca(%d+)')) ..
        '.bin', sample_clock.master_rate)

    self:cmd_block('txdr_passthru_power_on')    -- Need to fix someday
    self.channel_sinks = {}
    channel_proxy.initialize(self, name)
    self.connector = '@' .. self.proxy_hostname
end

-- Function used to decode raw data from the FPGA
function can_proxy:buffer(msg)
    if msg.cmd == C.CAN_RX_START then
        local buf = ffi_cast('can_rxtx_buffer_t *', msg.bufdata)
        for i = 1, msg.element_count do
            for _, sink in pairs(self.channel_sinks) do
                local m = ffi_cast(message_can.ptr_type, buf)
                local err = sink:enqueue(m)
                if err then
                    self.md_sink{
                        error_message =
                            string.format("Error sending to test: %s", err),
                    }
                    return
                end
            end
            buf = buf + 1
        end
    elseif msg.cmd == C.CAN_EVENT_START then
        local buf = ffi_cast('can_ev_buffer_t *', msg.bufdata)
        for i = 1, msg.element_count do
            if buf.events.bus_off ~= 0 then
                self.md_sink{
                    error_message = 'CAN Bus is OFF',
                }
                return
            end
            buf = buf + 1
        end
    end
end

local spec_translate = {
    bit_rate = 'baud_rate',
}

-- Function used to start CAN data from the FPGA
function can_proxy:start_stream(w)
    w.proxy = self

    local device_spec = {}
    for k, v in pairs(self.spec) do
        device_spec[spec_translate[k] or k] = v
    end
    if not self.spec.internal_termination then
        self:cmd('disable_terminator')
    else
        self:cmd('enable_terminator')
    end
    self:cmd('bus_setup', device_spec)

    self:cmd('event_setup', {bus_off=true, bus_off_deassert=true})
    self:cmd('rx_start', self.start_stamp, update_rate, 16, 128)
    self:cmd('event_start', self.start_stamp, update_rate, 16, 128)
end

-- Function used to stop the FPGA from sending CAN data
function can_proxy:stop_stream()
    self:cmd('event_stop')
    self:cmd('rx_stop')
end

-- Function used to assign a comet sink to channels
function can_proxy:start_comet_stream(w, channels)
end

-- Function called to start acquisiton on a channel
function can_proxy:start_channel_stream(w, spec, md_sink, sink)
    -- Don't allow more than one test to use this channel
    for _, _ in pairs(self.channel_sinks) do
        md_sink{
            error_message =
                "Channel can not be used in more than one test at a time",
        }
        return
    end
    -- Dont allow the control channel to be used in a test
    if spec.connector:find('^control') then
        md_sink{
            error_message = "Control channels can not be used in a test",
        }
        return
    end

    self:cmd_block('txdr_passthru_power_on')    -- Need to fix someday

    -- Initial channel setup
    self.start_stamp = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
    -- Make it 1 second early
    self.start_stamp = self.start_stamp - sample_clock.master_rate
    spec.ts = self.start_stamp
    self.spec = spec

    -- Make sure the start time being supplied is doable!
    local cts = self:cmd_block('get_current_ts')
    if cts and cts.ts and cts.ts > self.start_stamp then
        md_sink{
            error_message = "Test start time in the past",
        }
        return
    end

    -- Special 'sink' for raw CAN
    -- NOTE:  Back out start time adjustment so we get timestamps relative to
    -- the actual test start time.
    sink = ffi_new('can_sink', self.start_stamp + sample_clock.master_rate,
        w.socket, can_buf_count, 0)

    -- Insert into Channel Sink table
    self.channel_sinks[w] = sink

    -- Cleanup function called when a test disconnects
    w.cleanup = function()
        self.channel_sinks[w] = nil
        for _, _ in pairs(self.channel_sinks) do return end
        self:stop_stream()
    end

    self:start_stream(w)
    self.md_sink = md_sink
    md_sink(spec)
end

function can_proxy:available_channels(hostname)
    return {
        {
            name = 'raw_' .. hostname:gsub("%W", "_"),
            connector = '@' .. hostname,
            description = "Raw CAN messages",
            data_type = 'message_can',
            category = 'source/hbm_cx23r/msg_can',
            message_format = 'CAN',
        },
        {
            name = 'control_' .. hostname:gsub("%W", "_"),
            connector = 'control.status@' .. hostname,
            description = 'CAN Control Channel',
            data_type = 'message_raw',
            category = 'source/hbm_cx23r/msg_can',
        }
    }
end

return can_proxy
