--! \file

local armada = require 'armada'
local dbus = require 'dbus'
local ev = require 'ev'
local http_server = require 'http_server'
local u = require 'unix'
local cjson = require 'cjson.safe'
local sqlite = require 'sqlite_ffi'
local ts = require 'timestamps'
local data_dumper = require 'data_dumper'
local net = require 'net'
local ffi = require 'ffi'
local C = ffi.C
local sys_notify = require 'sys_notify'
local config = require 'system_config'
local log = require 'log'

ffi.cdef[[
struct dhdr {
    unsigned int        len;
    long long           ts;
    unsigned char       data[0];
} __attribute__((packed));

struct my_iovec {
    void        *iov_base;
    size_t      iov_len;
};

typedef int ssize_t;
ssize_t send(int sockfd, const void *buf, size_t len, int flags);
ssize_t writev(int fd, const struct my_iovec *, int num);

union htonll_u {
    double d;
    uint64_t li;
    uint32_t i[2];
};
]]

local iovec = ffi.new('struct my_iovec[2]')
local dhdr = ffi.new('struct dhdr[1]')

ffi.fill(iovec, ffi.sizeof(iovec))

iovec[0].iov_len = ffi.sizeof('struct dhdr')
iovec[0].iov_base = dhdr

local dhdr_size = ffi.sizeof('struct dhdr')

--! The base channel proxy implementation.
--? class channel_proxy
local m = class('channel_proxy')
--? alias m channel_proxy

-- Function used to retrieve the ZCIP assigned IPv4 address of the internal
-- network interface.  (This should probably be put in a utilities or net file?)
local function get_zcip_addr()
    local addrs = net.getifaddrs('ethint')
    for _, addr in ipairs(addrs) do
        if addr:find('^169%.254%.') then
            return addr
        end
    end
end 

-- Initialize channel_proxy object.
function m:initialize(name, no_module_hostname)
    self.proxy_name = name
    assert(self.proxy_type)

    -- If any channel_proxy needs to read the system preferences from
    -- the database, enable this line of code.
    --self:system_preferences()

    ev.Signal.new(function() -- Ignore SIGPIPE -- KLUDGE move to server.lua
--        print('SIGPIPE')
    end, u.SIGPIPE):start(ev.Loop.default)
    ev.Signal.new(function() -- Allow single Ctrl-C to kill
        print('SIGINT')
        ev.Loop.default:unloop()
    end, u.SIGINT):start(ev.Loop.default)

    self.module_hostname = armada.network:GetModuleHostname()
    self.module_ip = armada.network:GetModuleIP()
    self.channel_listen_ip = armada.network:GetModuleLinkLocalIP()
    self.channel_setup_ip = self.channel_listen_ip:gsub("%%.*$", "")

    if not no_module_hostname then
        self.proxy_hostname = string.format("%s.%s",
            self.proxy_name, self.module_hostname)
    else
        self.proxy_hostname = self.proxy_name
    end

--[[
    self:logf("notice", "%s channel_proxy starting up: %s",
        self.proxy_type, self.proxy_hostname)
]]

    self.proxy_address = self.module_ip

    self.http = http_server:new{
        ev_loop = ev.Loop.default,
        handler = self.http_handler,
        add_to_watchers = {
            channel_proxy = self,
        }
    }
    self.http_w = self.http:listen{address = self.proxy_address}
    self.proxy_port = self.http_w.port

    -- Add a listener on a UNIX domain socket (for local connection support)
    self.http_local = http_server:new{
        ev_loop = ev.Loop.default,
        handler = self.http_handler,
        add_to_watchers = {
            channel_proxy = self,
        }
    }
    self.http_local_w =
        self.http:listen{path = '/tmp/proxy/' .. self.proxy_hostname}
    self.proxy_path = self.http_local_w.path
--[[
    self:logf("notice", "Listening for HTTP on %s %d",
        self.proxy_address, http_w.port)
]]

    local ch_list = self:available_channels(self.proxy_hostname)

    local host = self.proxy_hostname
    self.old_available_channels = cjson.encode(ch_list)
    armada.network:ModifyRRs({}, {
        {'_armada-connectors', 'PTR', host},
        {host, 'SRV', '0 0 ' .. self.http_w.port .. ' ' .. self.module_hostname},
        {host, 'TXT', 'type: ' .. self.proxy_type},
        {host, 'TXT', 'available_channels: ' .. self.old_available_channels},
    })

    if type(self["handle_server"]) == "function" and
       type(self["get_server_port"]) == "function" then
-- FIX - This needs to use published IP addresses when the method for
-- doing so is finally settled & implemented.  This currently only works
-- for on-box communications, which is adequate for milestone 10 (first
-- sales release).
        local proxy = self
        -- Create listener socket for communications TO the proxy.
        -- The net.lua listening socket utility is used;
        -- storage.lua was used as an example.
        local internal_addr = get_zcip_addr()
        self.server =
            net.socket{ip = internal_addr,
                       port = self:get_server_port()}:listen()

        -- Function to handle stream sockets returned when connection
        -- listener processes a connection request.  Parameters:
        -- loop:  the ev.IO event loop the listener runs in
        -- sock:  the socket at this end of a requested connection
        -- addr:  NOT used currently; the address of the peer socket
        --        (at the other end of the connection)
        function self.server:acceptor(loop, sock, addr)
            local sbuf = net.streambuf(sock)
            local termstr = ffi.new("uint8_t[1]", {})

            -- Callback used by StreamBuffer to close itself.
            function sbuf:close_cb(loop, dir, source)
                -- print(self.socket, 'shutdown', dir, source)
                self:close()
            end

            -- Callback used by StreamBuffer to read from socket.
            -- Note that the returned value is used by StreamBuffer
            -- to remove consumed bytes from the read buffer.
            function sbuf:read_cb(loop)
                local consumed = 0
                local term_offset = self.readbuf:search(0, termstr, 1)
                if term_offset and term_offset > 0 then
                    local bytes_read, packet = self:read(term_offset)
                    if bytes_read > 0 then
                        local json =
                            cjson.decode(ffi.string(packet, bytes_read))
                        proxy:handle_server(json)
                    else
                        self:logf("error", "%s - server sock streambuf read",
                            self.proxy_hostname)
                    end
                    consumed = term_offset + 1
                elseif term_offset and term_offset == 0 then
                    consumed = 1  -- get rid of extraneous byte
                end
                return consumed
            end

            sbuf:start(loop)
        end
        self.server:start(ev.Loop.default)
    end
end

function m:destroy()
    self.http:shutdown_server()
    local host = self.proxy_hostname
    local http_w = self.http_w
    armada.network:ModifyRRs({
        {'_armada-connectors', 'PTR', host},
        {host, 'SRV', '0 0 ' .. http_w.port .. ' ' .. self.module_hostname},
        {host, 'TXT', 'type: ' .. self.proxy_type},
        {host, 'TXT', 'available_channels: ' .. self.old_available_channels},
    }, {})
end


m.system_overloaded = function(msg, reboot)
    local sn = sys_notify:new()
    sn.title = "System Overload"
    sn:critical_error(msg, reboot)
end

m.hardware_error = function(msg, reboot)
    local sn = sys_notify:new()
    sn.title = "Hardware Error"
    sn:critical_error(msg, reboot)
end

m.internal_error = function(msg, reboot)
    local sn = sys_notify:new()
    sn.title = "Internal Error"
    sn:critical_error(msg, reboot)
end

m.system_error = function(msg, reboot)
    local sn = sys_notify:new()
    sn.title = "System Error"
    local m = msg:match('ERROR %- (.-)\n')
    if m then msg = m end
    sn:error(msg, reboot)
end

m.system_notification = function(msg)
    local sn = sys_notify:new()
    sn.title = "Notification"
    sn:notify(msg)
end

function m:logf(level, fmt, ...)
    armada.logger:Log('channel_proxy('..self.proxy_name..')', level,
        string.format(fmt, ...))
    if level == 'error' then
        self.system_error(string.format(fmt, ...), true)
    end
end

function m:write_metadata(w, metadata)
--[[
-- WTF?  We don't write metadata to the test engine anymore!
    local ts = "\128\000\000\000\000\000\000\000"
    local md = cjson.encode(metadata)
    -- 4-byte alignment crucial to client efficiency on PowerPC
    -- Unaligned floating point accesses are an order of magnitude slower
    -- So pad metadata to maintain data alignment
    local pad_to = 4 -- sizeof int32_t
    local pad = string.sub("        ", 1, (pad_to - (#md % pad_to)) % pad_to)
    local msg = net.packmsg(ts .. md .. pad)
    assert(#msg % pad_to == 0)
    w.streambuf:write(msg)
]]
    if metadata.error_message and self.allow_error then
        local msg = string.format("%s (%s): %s", self.proxy_hostname,
            self.spec.name or 'channel not known', metadata.error_message)
        log:log(log.ERR, msg)
        w.server.http_write_shutdown(w.server.ev_loop, w)
        self.allow_error = nil
    end
end

function m.http_handler(loop, w, http)
--print("channel_proxy http_handler, http=" .. data_dumper(http, "", true))
    local self = assert(w.channel_proxy)
    if http.uri:match("^/update_settings/") then
--print("channel_proxy UPDATE_SETTINGS: " .. data_dumper(cjson.decode(http.content), "", true))
        local result = self:update_settings(loop, w, http, cjson.decode(http.content))
        w.streambuf:write(
            "HTTP/1.0 200 OK\r\n"
            .. "Content-Type: application/json\r\n"
            .. "\r\n"
            .. cjson.encode(result)
            .. "\n"
        )
        result.acknowledged = "ACK"
        w.streambuf:write(cjson.encode(result))
        w.streambuf:write("\n")
        w.server.http_write_shutdown(loop, w)
    elseif http.uri:match("^/control") then
--print("channel_proxy CONTROL: " .. data_dumper(cjson.decode(http.content), "", true))
        local result
        if self.control then
            local lw = w
            result = self:control(loop, w, http, cjson.decode(http.content),
                function(ret)
                    lw.streambuf.write_cb = function(l, s)
                            lw.streambuf:shutdown()
                        end
                    local json = cjson.encode(ret)
                    lw.streambuf:write(
                        "HTTP/1.0 200 OK\r\n"
                        .. "Content-Length: " .. #json + 2 .. "\r\n"
                        .. "Content-Type:  application/json\r\n\r\n"
                        .. json
                        .. "\r\n"
                    )
                end)
            if not result then return end
        else
            result = {
                result = 'error',
                msg = 'Channel does not support control API',
            }
        end
        if result and result.result and result.result == 'error' and
            result.msg then
            self:logf("error", "%s - %s", self.proxy_hostname, result.msg)
        end
        local json = cjson.encode(result)
        w.streambuf:write(
            "HTTP/1.0 200 OK\r\n"
            .. "Content-Length: " .. #json + 2 .. "\r\n"
            .. "Content-Type:  application/json\r\n\r\n"
            .. json
            .. "\r\n"
        )
        w.server.http_write_shutdown(loop, w)
    elseif http.uri:match("^/comet_stream$") then
        w.streambuf:write(
            "HTTP/1.0 200 OK\r\n"
            .. "Content-Type: application/octet-stream\r\n"
            .. "\r\n"
        )
        local sink = function(timestamp, name, value)
            local o = {[name]=value}
            w.streambuf:write(cjson.encode(o) .. "\n\004")
        end
        local channels = {}
        if http.headers.content_type == "application/json" then
            channels = cjson.decode(http.content)
        end
        self:start_comet_stream(w, channels, sink) -- KLUDGE for now!!

    elseif http.uri:match("^/channel_stream$") then
        -- "somat:connector": "latitude@gps-1.ppc1",
        -- "core:start_time": "2010-11-09T23:14:39.000000000",

        local spec = cjson.decode(http.content)
        local md_sink = function(metadata)
            return self:write_metadata(w, metadata)
        end
        local sock = w.socket
        local sink = function(timestamp, payload, plen)
            local len = plen or payload:len()
            dhdr[0].len = len + dhdr_size - 4;  -- Adjusted for length field
            dhdr[0].ts = timestamp
            dhdr[0].len = self.htonl(dhdr[0].len)
            dhdr[0].ts = self.htonll(dhdr[0].ts)
            iovec[1].iov_len = len
            iovec[1].iov_base = ffi.cast('void *', payload)
            w.streambuf:writev(iovec, 2)
        end

        self.allow_error = true
        self:start_channel_stream(w, spec, md_sink, sink) -- KLUDGE for now!!

    elseif http.uri:match("^/channel_sink_stream$") then
        -- "somat:connector": "latitude@gps-1.ppc1",
        -- "core:start_time": "2010-11-09T23:14:39.000000000",

        local spec = cjson.decode(http.content)
        local md_sink = function(metadata)
            return self:write_metadata(w, metadata)
        end

        self:start_channel_sink(w, spec, md_sink)
    end
end

-- endian conversion
local htonll_data = ffi.new('union htonll_u')
if ffi.abi("be") then -- nothing to do
    function m.htonl(b) return b end
    function m.htons(b) return b end
    function m.htonll(b) return b end
    function m.htonll_double(b) return b end
else
    function m.htonl(b) return bit.bswap(b) end
    function m.htons(b) return bit.rshift(bit.bswap(b), 16) end
    function m.htonll_double(b)    -- assumes b is ffi C type double
        htonll_data.d = b
        local i1 = m.htonl(htonll_data.i[1])
        local i0 = m.htonl(htonll_data.i[0])
        htonll_data.i[1] = i0
        htonll_data.i[0] = i1
        return htonll_data.d
    end
    function m.htonll(b)    -- assumes b is ffi C type long long
        htonll_data.li = b
        local i1 = m.htonl(htonll_data.i[1])
        local i0 = m.htonl(htonll_data.i[0])
        htonll_data.i[1] = i0
        htonll_data.i[0] = i1
        return htonll_data.li
    end
end
m.ntohl = m.htonl -- reverse is the same
m.ntohs = m.htons -- reverse is the same
m.ntohll = m.htonll -- reverse is the same
m.ntohll_double = m.htonll_double -- reverse is the same

local function sql_exec(db, table_name, sql, args)
    local query = string.format(sql, table_name):gsub('^%s*(.-)%s$', '%1')
    local ok, result = pcall(function() return db:exec(query, args) end)
    if ok then
        return result, ok
    else
        return {ok=false, error=result}, ok
    end
end

function m:system_preferences(update)
    if self.sys_prefs and not update then
        return self.sys_prefs
    end
    local db = sqlite:new('/var/lib/armada/setup.db')
    local result, ok = sql_exec(db, 'tsd_users', [[
        SELECT content
        FROM "%s"
        WHERE deleted=''
            AND db_name='system'
            AND id='preferences';
    ]])
    if ok and result and result[1] then
        local prefs = cjson.decode(result[1].content)
        if type(prefs.system) == 'table' then
            self.sys_prefs = prefs.system
        end
    end
    db:close()
    return self.sys_prefs or {}
end

return m
