local dbus = require 'dbus.core'

--! \file

--? namespace dbus
--? class dbus.Connection

--! Helper function to get a callback function from \ref Callbacks.
local function get_callback_fn(callback)
    if type(callback) == 'table' and 
            (not getmetatable(callback) or
             not getmetatable(callback).__call) then
        local object, target = callback[1], callback[2]
        if type(target) == 'function' then
            return function(...) return target(object, ...) end
        else
            return function(...) return object[target](object, ...) end
        end
    else
        return callback
    end
end

--! A Lua proxy for a remote D-Bus object.
--!
--! To create a Proxy object, call dbus#Connection#new_proxy or
--! dbus#Connection#auto_proxy.
--? class dbus.Proxy
dbus.Proxy = {}
dbus.Proxy.__index = dbus.Proxy

--! Creates a new Proxy.
--!
--! This creates a new, empty proxy.
--!
--! \param destination The bus destination of the proxy.
--! \param object The object path of the proxy.
function dbus.Connection:new_proxy(destination, object)
    return setmetatable({
                            _connection = self,
                            _destination = destination,
                            _object = object,
                            _options = { }
                        }, dbus.Proxy)
end

--! A binding interface for a remote object's signal.
--? class dbus.Bindable
dbus.Bindable = {}
dbus.Bindable.__index = dbus.Bindable

--! Adds a method to a proxy.
--!
--! \param name the local proxy method name
--! \param interface the remote interface name
--! \param method the remote method name
--! \param signature the in argument signature for the method
--! \param options options
function dbus.Proxy:add_method(name, interface, method, signature, options)
    if not options then options = { } end
    self._options[name] = options
    self[name] =
        function(proxy, ...)
            local call = dbus.Message.new_method_call(
                proxy._destination, proxy._object,
                interface, method)
            call:append_args({...}, signature)
            if options.no_reply then
                call:set_no_reply(true)
                proxy._connection:send(call)
            else
                local reply, err =
                    proxy._connection:send_with_reply_and_block(call)
                if reply then
                    local rargs = reply:get_args()
                    return unpack(rargs)
                else
                    local cs, ce = err:find(": ")
                    return nil, err:sub(1, cs - 1), err:sub(ce + 1)
                end
            end
        end
    if not options.no_reply then
        local num_args = #dbus.Message.split_args(signature)
        self[name.."_async"] =
            function(proxy, ...)
                local args = {...}
                local callback = get_callback_fn(args[num_args + 1])
                args[num_args + 1] = nil
                local call = dbus.Message.new_method_call(
                    proxy._destination, proxy._object,
                    interface, method)
                call:append_args(args, signature)
                local pending = proxy._connection:send_with_reply(call)
                pending:set_notify(function(conn, msg, pending)
                                       local args = msg:get_args()
                                       table.insert(args, msg)
                                       if msg:get_type() == 'error' then
                                           callback(nil,
                                                    msg:get_error_name(),
                                                    unpack(args))
                                       else
                                           callback(unpack(args))
                                       end
                                   end)
            end
    end
end

--! Adds a signal to a proxy.  The name element in the proxy object
--! will be set to a Bindable object which may be used to register a
--! signal listener.
--!
--! \param name the local proxy signal name
--! \param interface the remote interface name
--! \param signal the remote signal name
--! \param signature the out argument signature for the signal
--! \param options options
function dbus.Proxy:add_signal(name, interface, signal, signature, options)
    if not options then options = { } end
    -- Store signal information for later binding
    self[name] = setmetatable({
                                  proxy = self,
                                  interface = interface,
                                  signal = signal,
                                  signature = signature,
                                  options = options
                              }, dbus.Bindable)
end

--! Flexible LOM iterator.  When used in a \p for statement, iterates
--! through all immediate subelements of \a el.
--!
--! \param el root element to iterate through
--! \param tag if set, only stop for elements of this type
--! \param attr (attribute name for \a value parameter)
--! \param value if set, only stop for elements where \a attr equals \a value
local function itags(el, tag, attr, value)
    return function(s, var)
               var = var + 1
               while s[var] do
                   if type(s[var]) == 'table' then
                       if (not tag or s[var].tag == tag) and
                          (not value or s[var].attr and
                           s[var].attr[attr] == value) then
                           return var, s[var]
                       end
                   end
                   var = var + 1
               end
               return nil
           end, el, 0
end

local function auto_proxy_init(proxy)
    local lom = require 'lxp.lom'
    proxy:add_method('Introspect', dbus.INTERFACE_INTROSPECTABLE,
                     'Introspect', '')
    local xml, ename, eval = proxy:Introspect()
    if xml then
        local node = lom.parse(xml)
        for i,interface in itags(node, 'interface') do
            for i,method in itags(interface, 'method') do
                local args = { }
                local options = { }
                for i,arg in itags(method, 'arg') do
                    if not arg.attr.direction or
                        arg.attr.direction == 'in' then
                        table.insert(args, arg.attr.type)
                    end
                end
                for i,annotation in
                    itags(method, 'annotation', 'name',
                          'org.freedesktop.DBus.Method.NoReply') do
                    options.no_reply = annotation.attr.value == 'true'
                end
                proxy:add_method(method.attr.name,
                                 interface.attr.name,
                                 method.attr.name,
                                 table.concat(args),
                                 options)
            end
            for i,signal in itags(interface, 'signal') do
                local args = { }
                for i,arg in itags(signal, 'arg') do
                    table.insert(args, arg.attr.type)
                end
                proxy:add_signal(signal.attr.name,
                                 interface.attr.name,
                                 signal.attr.name,
                                 table.concat(args),
                                 { type = 'signal' })
            end
        end
        return proxy
    else
        return nil, ename, eval
    end
end

--! Creates a new Proxy automatically populated by introspection.
--!
--! This method creates a new proxy and automatically adds appropriate
--! methods and signals to it by D-Bus introspection.
--!
--! This method blocks while requesting the introspection information.
--!
--! \param destination The bus destination of the proxy.
--! \param object The object path of the proxy.
function dbus.Connection:auto_proxy(destination, object)
    local proxy = self:new_proxy(destination, object)
    return auto_proxy_init(proxy)
end


--! Creates a new Proxy automatically populated by introspection,
--! blocking until the destination is available or until the
--! timeout expires.
--!
--! Unlike dbus#Connection#auto_proxy, this function will wait for
--! the bus destination to become available instead of failing out
--! immediately.
--!
--! The Connection must have been registered with libev with the
--! dbus#Connection#hook_libev_default_loop function before calling
--! this.
--!
--! \param destination The bus destination of the proxy.
--! \param object The object path of the proxy.
--! \param timeout Optional time to wait before invoking error()
function dbus.Connection:auto_proxy_wait(destination, object, timeout)
    local ev = require 'ev'
    local got_it = false
    timeout = timeout or 30 -- KLUDGE TESTING
    local timer, err
    local watcher = self:new_name_watcher(
        destination,
        function(name, new)
            if new ~= '' then
                got_it = true
            end
        end)
    if timeout then
        timer = ev.Timer.new(
            function(loop, timer, revents)
                err = "Timeout: auto_proxy_wait(" .. destination .. ")"
                got_it = true
            end, timeout, timeout)
        timer:start(ev.Loop.default)
    end
    repeat
        ev.Loop.default:loop('once')
    until got_it
    if timer then
        timer:stop(ev.Loop.default)
    end
    watcher:unregister()
    if err then
        print(err)
--        error(err)
    end
    return self:auto_proxy(destination, object)
end

--! An object that defines a Lua interface for a D-Bus interface.
--!
--! Interface objects are used in the dbus#Connection#export_object
--! method.
--!
--! Example usage:
--! \code
--! local test = dbus.Interface.new('com.somat.Test')
--! test:add_method{'Add', 'ii', 'i'}
--! test:add_method{'AddDelayed', 'iid', 'i'}
--! test:add_method{'Error', 'ii', 'i'}
--! test:add_method{'Print', 's', '', no_reply = true }
--! test:add_method{'SendSignal', 's', '', no_reply = true }
--! test:add_signal{'Signal', 's'}
--!
--! connection:export_object(object, '/com/somat/Test', { test })
--! \endcode
--? class dbus.Interface
dbus.Interface = {}
dbus.Interface.__index = dbus.Interface

--! Creates a new interface definition.
--! \param name the D-Bus name of the interface
--! \return the new Interface object
function dbus.Interface.new(name)
    return setmetatable({
                            name = name,
                            methods = { },
                            signals = { }
                        }, dbus.Interface)
end

--! Adds a method to an interface definition.
--!
--! This function takes a table as its argument, shaped like:
--! <tt>{name, in_signature, out_signature, [target], [option = value]*}</tt>
--!
--! \param name the D-Bus name of the method
--! \param in_signature the \ref Signatures "D-Bus signature" of the arguments
--! \param out_signature the \ref Signatures "D-Bus signature" of the returns
--! \param target the target Lua method to call, either a string or function,
--!               defaulting to \a name.
--!
--! If \p no_reply option is \p true, the generated introspection data
--! for the method will have the \p org.freedesktop.DBus.Method.NoReply
--! annotation set.
function dbus.Interface:add_method(args)
    local name, in_signature, out_signature, target = unpack(args)
    local method
    if not target then
        target = name
    end
    if type(target) == 'string' then
        method = { func = function(object, ...)
                              return object[target](object, ...)
                          end }
    elseif type(target) == 'function' then
        method = { func = target }
    end
    method.in_signature = in_signature
    method.out_signature = out_signature
    for k,v in pairs(args) do
        if type(k) ~= 'number' then
            method[k] = v
        end
    end
    self.methods[name] = method
end

--! Adds a signal to an interface definition.
--!
--! This function takes a table as its argument, shaped like:
--! <tt>{name, signature, [target], [option = value]*}</tt>
--!
--! \param name the D-Bus name of the signal
--! \param signature the \ref Signatures "D-Bus signature" of the arguments
--! \param target the signal method name to create in the Lua object when
--!               exporting, defaulting to \a name.
function dbus.Interface:add_signal(args)
    local name, signature, target = unpack(args)
    local signal
    if not target then
        target = name
    end
    signal = {
        name = name,
        signature = signature,
        target = target
    }
    for k,v in pairs(args) do
        if type(k) ~= 'number' then
            signal[k] = v
        end
    end
    self.signals[name] = signal
end

local introspector = dbus.Interface.new(dbus.INTERFACE_INTROSPECTABLE)
--! Generates introspection XML for an object.
local function do_introspect(obj)
    local reply = { [[
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
"http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node>
]] }
    local function ins(str) table.insert(reply, str) end

    for iname, interface in pairs(obj._dbus_export.interfaces) do
        ins(string.format('  <interface name="%s">\n', iname))
        for mname, method in pairs(interface.methods) do
            ins(string.format('    <method name="%s">\n', mname))
            local function pargs(direction, args)
                for i,v in ipairs(args) do
                    ins(string.format(
                            '      <arg direction="%s" type="%s"/>\n',
                            direction, v))
                end
            end
            pargs('in', dbus.Message.split_args(method.in_signature))
            pargs('out', dbus.Message.split_args(method.out_signature))
            if method.no_reply then
                ins('      <annotation ' ..
                    'name="org.freedesktop.DBus.Method.NoReply" ' ..
                    'value="true"/>\n')
            end
            ins('    </method>\n')
        end
        for sname, signal in pairs(interface.signals) do
            ins(string.format('    <signal name="%s">\n', sname))
            for i,v in ipairs(dbus.Message.split_args(signal.signature)) do
                ins(string.format('      <arg type="%s"/>\n', v))
            end
            ins('    </signal>\n')
        end
        ins('  </interface>\n')
    end

    ins('</node>\n')

    return table.concat(reply)
end
introspector:add_method{'Introspect', '', 's', do_introspect}

--! Exports a Lua object over D-Bus as defined by one or more Interface
--! objects.
function dbus.Connection:export_object(object, object_path, interfaces)
    if object._dbus_export then
        error('object already being exported!')
    end
    object._dbus_export = {
        object_path = object_path,
        connection = self,
        interfaces = { }
    }
    table.insert(interfaces, introspector)
    for i,interface in ipairs(interfaces) do
        object._dbus_export.interfaces[interface.name] = interface
        for sname, signal in pairs(interface.signals) do
            object[signal.target] =
                function (object, ...)
                    local sig = dbus.Message.new_signal(object_path,
                                                        interface.name,
                                                        signal.name)
                    sig:append_args({...}, signal.signature)
                    object._dbus_export.connection:send(sig)
                end
        end
    end
    self:register_object_path(
        object_path,
        function(connection, message, op)
            local interface_name = message:get_interface()
            local method_name = message:get_member()
            local args, sig = message:get_args()

            local interface = object._dbus_export.interfaces[interface_name]
            if not interface then return false end

            local method = interface.methods[method_name]
            if not method or sig ~= method.in_signature then return false end

            local function reply(...)
                local args = {...}
                if args[1] == nil and args[2] ~= nil then
                    local err = message:new_error(args[2], args[3])
                    connection:send(err)
                else
                    local ret = message:new_method_return()
                    ret:append_args(args, method.out_signature)
                    connection:send(ret)
                end
            end

            table.insert(args, reply)
            table.insert(args, message)
            table.insert(args, connection)
            local reply_args = {method.func(object, unpack(args))}

            if reply_args[1] == reply then
                -- handling its own reply
                message:ref()
            elseif not message:get_no_reply() then
                reply(unpack(reply_args))
            end

            return true
        end)
end

--! Converts a table-based match specification into a D-Bus match
--! specification.
local function to_matchrule(match)
    local rule = { }
    local keys = { }
    for k,v in pairs(match) do
        table.insert(keys, k)
    end
    table.sort(keys)
    for i,k in ipairs(keys) do
        table.insert(rule, string.format("%s='%s'", k, match[k]))
    end
    return table.concat(rule, ',')
end

--! A filtered listener for incoming D-Bus messages.
--! To create a new Listener, call dbus#Connection#register_listener or
--! dbus#Bindable#bind.
--? class dbus.Listener
dbus.Listener = { }
dbus.Listener.__index = dbus.Listener

--! A D-Bus name watcher.
--? class dbus.NameWatcher
dbus.NameWatcher = { }
dbus.NameWatcher.__index = dbus.NameWatcher

--! Registers a new name watcher.  The callback will be called with
--! two arguments; the first is the the name being watched, and the
--! second is the current unique bus name corresponding to the watched
--! name.  If the second argument is the empty string, nothing on the
--! bus currently has the watched name.  If the watched name is a
--! unique bus name and the name goes away, the watcher is
--! automatically unregistered.  The callback is guaranteed to be
--! called at least once.
--!
--! \param name the bus name to watch
--! \param callback the callback, see \ref Callbacks
--! \return a NameWatcher object
function dbus.Connection:new_name_watcher(name, callback)
    if not self.bus_proxy then
        self.bus_proxy = self:new_proxy(dbus.SERVICE_DBUS, dbus.PATH_DBUS)
        self.bus_proxy:add_method('GetNameOwner', dbus.INTERFACE_DBUS,
                                  'GetNameOwner', 's')
        self.bus_proxy:add_signal('NameOwnerChanged', dbus.INTERFACE_DBUS,
                                  'NameOwnerChanged', 'sss')
        self.names_watched = { }
    end
    callback = get_callback_fn(callback)
    local watcher = setmetatable({
                                     connection = self,
                                     name = name,
                                     callback = callback
                                 }, dbus.NameWatcher)
    local watched = self.names_watched[name]
    if watched then
        watched.targets[watcher] = true
        watched.count = watched.count + 1
        if watched.current then
            callback(name, watched.current)
        end
    else
        watched = { count = 1, targets = { [watcher] = true } }
        self.names_watched[name] = watched
        if name == dbus.SERVICE_DBUS then
            watched.current = name
            callback(name, name)
        else
            local function new_name(new)
                new = new or ''
                watched.current = new
                local c = { }
                for k,v in pairs(watched.targets) do c[#c+1] = k end
                for i,v in ipairs(c) do
                    if v.callback then
                        v.callback(name, new)
                    end
                end
                if new == '' and name:sub(1, 1) == ':' then
                    while #watched.targets > 0 do
                        watched.targets[1]:unregister()
                    end
                end
            end
            watched.owner_changed = self.bus_proxy.NameOwnerChanged:bind(
                function(name, old, new)
                    new_name(new)
                end,
                { arg0 = name })
            self.bus_proxy:GetNameOwner_async(
                name,
                function(owner)
                    new_name(owner)
                end)
        end
    end
    return watcher
end

--! Unregisters a name watcher.
function dbus.NameWatcher:unregister()
    if not self.callback then return end
    local watched = self.connection.names_watched[self.name]
    watched.targets[self] = nil
    watched.count = watched.count - 1
    if watched.count == 0 then
        if watched.owner_changed then
            watched.owner_changed:unregister()
            watched.owner_changed = nil
        end
        self.connection.names_watched[self.name] = nil
    end
    self.callback = nil
end

--! Filter routine for Listener objects.
local function filter_func(conn, msg, filt)
    local this = {
        type = msg:get_type(),
        sender = msg:get_sender(),
        interface = msg:get_interface(),
        member = msg:get_member(),
        path = msg:get_path(),
        destination = msg:get_destination()
    }
    local args

    for i,f in ipairs(conn.filters) do
        local match = true
        local need_args
        for k,v in pairs(f.match) do
            if this[k] ~= v then
                match = false
                break
            end
        end
        if match and f.match_args then
            if not args then
                args = msg:get_string_args()
                for i,v in ipairs(args) do
                    if type(v) == 'string' then
                        local k = 'arg' .. (i - 1)
                        this[k] = v
                    end
                end
            end
            for k,v in pairs(f.match_args) do
                if this[k] ~= v then
                    match = false
                    break
                end
            end
        end
        if match then
            f.callback(msg)
        end
    end
end

local function register_listener_guts(conn, self, match, callback)
    if not conn.matchfilter then
        conn.matchfilter = conn:add_filter(filter_func)
        conn.matchrules = { }
        conn.filters = { }
    end
    self.match = match
    self.rule = to_matchrule(match)
    self.connection = conn
    self.callback = get_callback_fn(callback)
    setmetatable(self, dbus.Listener)
    local match_args = { }
    for k,v in pairs(match) do
        if k:sub(1,3) == 'arg' then
            match_args[k] = v
            match[k] = nil
            self.match_args = match_args
        end
    end
    if conn.matchrules[self.rule] then
        conn.matchrules[self.rule] = conn.matchrules[self.rule] + 1
    else
        conn.matchrules[self.rule] = 1
        conn:bus_add_match(self.rule)
    end
    table.insert(conn.filters, self)
    if self.match.sender and self.match.sender ~= dbus.SERVICE_DBUS then
        self.name_watcher = conn:new_name_watcher(
            self.match.sender,
            function(name, new)
                self.match.sender = new
                if name:sub(1, 1) == ':' then
                    self:unregister()
                end
            end)
    end
    return self
end

--! Registers a filtered listener on a bus connection.
function dbus.Connection:register_listener(match, callback)
    return register_listener_guts(self, { }, match, callback)
end

--! Unregisters a listener.
function dbus.Listener:unregister()
    if not self.rule then return end
    if self.name_watcher then
        self.name_watcher:unregister()
        self.name_watcher = nil
    end
    local conn = self.connection
    conn.matchrules[self.rule] = conn.matchrules[self.rule] - 1
    if conn.matchrules[self.rule] == 0 then
        conn.matchrules[self.rule] = nil
        conn:bus_remove_match(self.rule)
    end
    self.rule = nil
    for i,v in ipairs(conn.filters) do
        if v == self then
            table.remove(conn.filters, i)
            break
        end
    end
    if #conn.filters == 0 then
        conn.matchfilter:remove()
        conn.matchfilter = nil
    end
end

local function bind_guts(self, listener, callback, options)
    callback = get_callback_fn(callback)
    local match = {
        type = 'signal',
        sender = self.proxy._destination,
        path = self.proxy._object,
        interface = self.interface,
        member = self.signal
    }
    if options then
        for k,v in pairs(options) do
            match[k] = v
        end
    end
    local function do_dispatch(msg)
        local args = msg:get_args()
        table.insert(args, msg)
        callback(unpack(args))
    end
    return register_listener_guts(self.proxy._connection, listener,
                                  match, do_dispatch)
end

--! Adds a listener to a remote object's signal.
--! \param callback the callback to receive the signal, see \ref Callbacks
--! \param options options
--! \return a Listener object
function dbus.Bindable:bind(callback, options)
    return bind_guts(self, { }, callback, options)
end

local PendingListener = { }
PendingListener.__index = PendingListener
function PendingListener:unregister()
    self._proxy._pending_binds[self] = nil
end

local PendingProxyElement = { }
PendingProxyElement.__index = PendingProxyElement
function PendingProxyElement:__call(...)
    if self._name:sub(-6) == '_async' then
        table.insert(self._proxy._pending_async_calls, { self._name, ... })
    else
        error('Tried to call PendingProxy method ' .. self._name)
    end
end
function PendingProxyElement:bind(...)
    local listener = { self._name, ... }
    listener._proxy = self._proxy
    setmetatable(listener, PendingListener)
    self._proxy._pending_binds[listener] = true
    return listener
end

--! A pending automatic proxy.
--? class dbus.PendingProxy
dbus.PendingProxy = {}
function dbus.PendingProxy:__index(name)
    return setmetatable({ _proxy = self, _name = name }, PendingProxyElement)
end

--! Creates a new PendingProxy which will be automatically populated
--! by introspection asynchronously.
--!
--! \param destination The bus destination of the proxy.
--! \param object The object path of the proxy.
--! \returns a PendingProxy object
function dbus.Connection:auto_proxy_async(destination, object)
    local proxy = self:new_proxy(destination, object)
    setmetatable(proxy, dbus.PendingProxy)
    proxy._pending_binds = {}
    proxy._pending_async_calls = {}
    proxy._name_watcher = self:new_name_watcher(
        destination,
        function(destination, new)
            if new ~= '' then
                setmetatable(proxy, dbus.Proxy)
                auto_proxy_init(proxy)
                for i,v in ipairs(proxy._pending_async_calls) do
                    local name = table.remove(v, 1)
                    proxy[name](unpack(v))
                end
                proxy._pending_async_calls = nil
                for k,v in pairs(proxy._pending_binds) do
                    local name = table.remove(k, 1)
                    bind_guts(proxy[name], k, unpack(k))
                end
                proxy._name_watcher:unregister()
                proxy._name_watcher = nil
            end
        end)
    return proxy
end

return dbus

--! \mainpage
--! See the #dbus package to get started.

--! \page Callbacks Callbacks
--!
--! \details In many cases the Lua D-Bus bindings want to register a
--! callback to be called when a signal or async method response is
--! received.  In these cases, the callback may be specified in the
--! following ways:
--!
--! If \a callback is a non-callable table (i.e. one that does not have
--! a \c __call method in its metatable), then the first element of the
--! table is taken to be the \a object and the second as the \a target.
--! If the \a target is a function, the callback is called as
--! <tt>target(object, ...)</tt>.  Otherwise, the callback is called as
--! <tt>object[target](object, ...)</tt>.
--!
--! If \a callback is anything else it is simply called as
--! <tt>callback(...)</tt>.
--!
--! Note that low-level callbacks (in particular, the handler function
--! for a dbus#Handler object and the notify function for a dbus#PendingCall
--! object) can only be specified as a function.
