--! \file

local common = require 'mx.device_common'
local underscore = require 'underscore'
local ffi = require 'ffi'
local C = ffi.C

local m = {}

-- XML used to do initial configuration of channels when proxy starts up
local default_xml = [[<?xml version="1.0" encoding="UTF-8"?>
                <Amplifier>
                    <Connector#conn_num#>
                        <Mode choice="AnalogInChannel">
                            <AnalogInChannel>
                                <Sensor>
                                    <TedsOptions>
                                        <UseTeds>2</UseTeds>
                                    </TedsOptions>
                                </Sensor>
                                <SyncSignal address="1">
                                    <OutputRate>10</OutputRate>
                                </SyncSignal>
                                <OutputUnit>
                                    <UnitId>1100</UnitId>
                                </OutputUnit>
                            </AnalogInChannel>
                        </Mode>
                    </Connector#conn_num#>
                </Amplifier>
]]

-- Function used to return the XML used to set the channel to it's default
-- configuration
function m:get_default_xml(channel_num)
    return default_xml:gsub('#conn_num', tostring(channel_num))
end

-- This is the mapping of JSON style setup parameters to XPATHs for the device
m.json2xpath = {
    excitation_range = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/Bridge/ExciteAmpl',
    excitation_mode = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/Bridge/ExciteFreq',
    _bridge_type = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/Bridge/BridgeType',
    bridge_resistance = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/Bridge/Impedance',
    _bridge_wiring = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype[@choice="Bridge"]/Bridge/Wiring',
    _pt_wiring = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype[@choice="Pt"]/Pt/Wiring',
    _pt_type = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/Pt/PtType',
}
underscore.extend(m.json2xpath, common.json2xpath)

m.phase_delay = {}
m.phase_delay.fixed = 398
m.phase_delay.bessel_iir = {}
m.phase_delay.butterworth_iir = {}
m.phase_delay.linear_phase_fir = {}
m.phase_delay.butterworth_fir = {}

m.ignore_list = {}
underscore.extend(m.ignore_list, common.ignore_list)

m.dont_verify_list = {}
underscore.extend(m.dont_verify_list, common.dont_verify_list)

function m:sensor_input_type(p, fromxml)
    return m.translate({
            ['1'] = 'quarter',
            ['2'] = 'half',
            ['4'] = 'full',
        }, p, fromxml)
end

function m:excitation_mode(p, fromxml)
    return m.translate({
            ['0'] = 'auto',
            ['1'] = 'dc',
            ['4'] = 'ac_1200_hz_square',
            ['7'] = 'auto_cf',
        }, p, fromxml)
end

function m:excitation_range(p, fromxml)
    if fromxml then
        return string.format("%.1fv", tonumber(p))
    end
    return p:match('(.+)[V|v]$')
end

m.translator = {
    sensor_input_type = m.sensor_input_type,
    excitation_mode = m.excitation_mode,
    excitation_range = m.excitation_range
}

function m:to_internal(spec)
    local spec = spec

    -- Do some processing based on sensor input mode (if it exists)
    if spec.sensor_input and spec.sensor_input.mode then

        -- Pt sensors require a specific set of parameter to be set or not set
        -- so we take care of that here!
        if spec.sensor_input.mode:find('^pt') then
            if spec.sensor_input.mode == 'pt1000' then
                spec._pt_type = 2
            else
                spec._pt_type = 1
            end
            spec.calibration.rezero.units = "°C"
            spec._pt_wiring = 4
            spec.sensor_input.mode = 'pt'
            spec.calibration.sensor_scaling.mode = 'internal'
            spec.calibration.sensor_scaling.units = "°C"
            spec.calibration.define_source_units.units = "°C"

            spec.drift_compensation = nil
            spec.teds = nil
        end

        -- If this is a bridge, set the bridge wiring
        if spec.sensor_input.mode == 'bridge' and spec.excitation_range then
            -- If this is NOT a TEDS device do this stuff
            if not spec.teds_sensor or not spec.teds_sensor.bindata then
                local bridge_max_levels = {
                    ['5.0'] = 4,
                    ['2.5'] = 8,
                    ['1.0'] = 20,
                    ['0.5'] = 40,
                }
                local er = string.format("%.1f",
                    tonumber(spec.excitation_range:match('(.*)[V|v]') or 0))
                if not bridge_max_levels[er] then
                    return nil, string.format("Invalid excitation range '%s'",
                        spec.excitation_range)
                end
                spec.excitation_range = er
                spec.sensor_input.max = bridge_max_levels[er]
                if spec.sensor_input.type == 'full' and
                    tonumber(spec.sensor_input.wires) == 4 then
                    spec.sensor_input.wires = '6'
                elseif spec.sensor_input.type == 'half' and
                    tonumber(spec.sensor_input.wires) == 3 then
                    spec.sensor_input.wires = '5'
                elseif spec.sensor_input.type == 'quarter' and
                    tonumber(spec.sensor_input.wires) == 3 and
                    not spec.sensor_input.regulation then
                    spec.sensor_input.wires = '9'
                end
            end
            spec._bridge_type =
                m.sensor_input_type(self, spec.sensor_input.type)
            spec.sensor_input.min = spec.sensor_input.max * -1
            spec._bridge_wiring = spec.sensor_input.wires
            spec.sensor_input.wires = nil
            spec.sensor_input.type = nil
        end

        -- If the sensor is a resistance sensor, make sure max is set
        if spec.sensor_input.mode == 'resistor' then
            spec.sensor_input.max = spec.sensor_input.max or 5000
        elseif spec.sensor_input.mode == 'potentiometer' then
            spec.sensor_input.max = spec.sensor_input.max or 500
            spec.sensor_input.resistance = spec.sensor_input.resistance or 500
            spec.calibration = spec.calibration or {}
            spec.calibration.define_source_units =
                spec.calibration.define_source_units or {
                    units = "mV/V",
                    applied = true,
                }
            spec.calibration.units_conversion =
                spec.calibration.units_conversion or {
                    units = "mV/V",
                    applied = true,
                }
        end
    end

    return spec
end

local set_shunt_xml_template =[[
<?xml version="1.0"?>
<SrvParams>
<Control>SHUNTCALIBRATION</Control>
<ControlParams>&lt;?xml version="1.0"
encoding="UTF-8"?&gt;&lt;ControlParams&gt;&lt;Par1&gt;#channel&lt;/Par1&gt;&lt;Par2&gt;#state&lt;/Par2&gt;&lt;/ControlParams&gt;</ControlParams>
</SrvParams>
]]

m.controls = {}
m.controls.channel = {}
m.controls.device = {}

-- Function used to control shunts on Quantum devices that have them.
-- Note, self in this function is the ChannelProxy object
function m.controls.channel:shunt(params, cb)
    local chan = self.index
    local enable = params and params[1]
    local xml = set_shunt_xml_template
    xml = string.gsub(xml, '#channel', chan)
    if type(enable) == 'boolean' then
        xml = string.gsub(xml, '#state', (enable and 1 or 0))
    else
        xml = string.gsub(xml, '#state', enable)
    end
    self.mxdev.mx.control:_send_transaction_request(C.TR_SCHEMA, C.TR_XML,
        function(obj, ecode, data, len)
            cb{result = 'success', msg = 'ok'}
        end, xml)
end

-- Device level shunt control that enables/disables shunts on all channels
function m.controls.device:shunt(params, cb)
    local enable = params and params[1]
    local xml = set_shunt_xml_template
    xml = string.gsub(xml, '#channel', 0)
    if type(enable) == 'boolean' then
        xml = string.gsub(xml, '#state', (enable and 1 or 0))
    else
        xml = string.gsub(xml, '#state', enable)
    end
    self.mx.control:_send_transaction_request(C.TR_SCHEMA, C.TR_XML,
        function(obj, ecode, data, len)
            cb{result = 'success', msg = 'ok'}
        end, xml)
end

function m:to_external(spec)
    local spec = spec

    -- Don't return this stuff to the external entities
    spec.wiring = nil

    if spec.sensor_input.mode and spec.sensor_input.mode ~= 'bridge' then
        spec.bridge_bridge_factor = nil
        spec.bridge_gage_factor = nil
        spec.bridge_resistance = nil
        spec.excitation_mode = nil
        spec.excitation_range = nil
    else
        spec.sensor_input.type =
            m.sensor_input_type(self, spec._bridge_type, true)
        spec.sensor_input.wires = spec._bridge_wiring
    end

    spec._bridge_wiring = nil
    spec._bridge_type = nil

    if spec.sensor_input.mode and spec.sensor_input.mode:find('^pt') then
        spec.calibration = spec.calibration or {}
        spec.calibration.define_source_units = {
            purpose = 'define_source_units',
            units = spec.calibration.sensor_scaling.units or "°C",
            applied = true,
        }

        spec.calibration.sensor_scaling = spec.calibration.sensor_scaling or {}
        spec.calibration.sensor_scaling.purpose = 'sensor_scaling'
--[[
        spec.calibration.sensor_scaling.mode = 'defined_slope_intercept'
        spec.calibration.sensor_scaling.slope = '1'
        spec.calibration.sensor_scaling.intercept = '0'
]]
        spec.calibration.sensor_scaling.mode = 'internal'
        spec.calibration.sensor_scaling.slope = nil
        spec.calibration.sensor_scaling.intercept = nil
        spec.calibration.sensor_scaling.applied = true
        spec.calibration.sensor_scaling.units =
            spec.calibration.define_source_units.units

        spec.sensor_input.wires = spec._pt_wiring or 4
        if spec._pt_type and tonumber(spec._pt_type) and
            tonumber(spec._pt_type) == 2 then
            spec.sensor_input.mode = 'pt1000'
        else
            spec.sensor_input.mode = 'pt100'
        end
    end

    spec._pt_wiring = nil
    spec._pt_type = nil
    return spec
end

-- Specification of the default settings of this channel type
m.default_spec = {
    sample_rate = '10',
    sensor_input = {
        mode = 'volt',
        min = '-15',
        max = '15.000000',
    },
    calibration = {
        define_source_units = {
            purpose = 'define_source_units',
            applied = true,
            units = 'V',
        },
        sensor_scaling = {
            purpose = 'sensor_scaling',
            mode = 'defined_slope_intercept',
            slope = '1',
            intercept = '0',
            applied = true,
            units = 'V',
        },
        units_conversion = {
            purpose = 'units_conversion',
            mode = 'defined_slope_intercept',
            applied = true,
            units = 'V',
            slope = '1',
            intercept = '0',
        },
        rezero = {
            purpose = 'rezero',
            mode = 'defined_slope_intercept',
            intercept = '0.00000',
            applied = true,
        },
    },
    digital_filter_type = 'none',
    sensor = {
        use_teds = 2,
    },
}

return m
