--! powerMicro package
--! \file
--!
--! \mainpage
--! Provides an API to the PDB power-micro
--!
--! API Functions:
--!  power_micro:new()
--!

local armada = require 'armada'
local ev = require 'ev'
local ffi = require 'ffi'
local bit = require 'bit'
local dbus = require 'dbus'
local C = ffi.C
local zlib = require 'zlib'
local sys_notify = require 'sys_notify'
local config = require 'system_config'
local cjson = require 'cjson.safe'
local status_lib = require 'status_lib'

local status_cb = function() end

local power_micro = class('powerMicro')

ffi.cdef[[
enum {
    CRC_MASK = 0xff,
};

enum {
    PDB_CMD_READ_VOLTAGE = 0,
    PDB_CMD_SET_LED_STATE,
    PDB_CMD_SEND_1_WIRE,
    PDB_CMD_SET_PARAM,
    PDB_CMD_READ_STATE,
    PDB_CMD_CH_POWER,
    PDB_CMD_SYSTEM_SHUTDOWN,
    PDB_CMD_GET_TEMP,
    PDB_CMD_STATUS_TIME_SET,

    PDB_CMD_PB_GET_CREDENTIAL_RESET_STATE,

    PDB_CMD_GET_VERSION = 0xd,

    PDB_CMD_POWER_LOST = 0x20,
    PDB_CMD_PUSHBUTTON_DEPRESSED = 0x21,
    PDB_CMD_PUSHBUTTON_RELEASED = 0x22,
    PDB_CMD_REMOTE_SWITCH_CHANGED = 0x23,
    PDB_CMD_CAP_BANK_CHARGED = 0x24,
    PDB_CMD_CAP_BANK_DEPLETED = 0x25,
    PDB_CMD_CAP_BANK_LEVEL = 0x26,

    PDB_CMD_STATUS = 0x28,

    // Firmware update commands
    PDB_CMD_START_FLASH_SEQUENCE = 0x70,
    PDB_CMD_ERASE_FLASH = 0x71,
    PDB_CMD_FLASH_LINE = 0x72,
    PDB_CMD_PROGRAM_FLASH = 0x73,
    PDB_CMD_FLASH_CRC = 0x74,
    PDB_CMD_FLASH_DONE = 0x75,
    PDB_CMD_SET_APPLICATION_STATE = 0x76,
    PDB_CMD_RESET_POWER_MICRO = 0x77,

    PDB_CMD_ACK = 0x80,
};

enum {
    PDB_SUB_LED_RED = 0x1,
    PDB_SUB_LED_GREEN = 0x2,
    PDB_SUB_LED_YELLOW = 0x4,
    PDB_SUB_LED_BLUE = 0x8,
};

enum {
    PDB_SUB_LED_OFF = 0,
    PDB_SUB_LED_ON,
    PDB_SUB_LED_0_5HZ,
    PDB_SUB_LED_1_HZ,
    PDB_SUB_LED_2_HZ,
    PDB_SUB_LED_4_HZ,
    PDB_SUB_LED_8_HZ,
};

enum {
    PDB_RSP_PWR_LOST = 0x20,
    PDB_RSP_LBTN_DOWN,
    PDB_RSP_LBTN_UP,
    PDB_RSP_RBTN_DOWN,
    PDB_RSP_RBTN_UP,
    PDB_RSP_CAP_CHARGED,
    PDB_RSP_CAP_DEPLETED,
    PDB_RSP_CAP_LEVEL
};

enum {
    PDB_APPLICATION_STATE = 0,
    PDB_LOADER_STATE = 1,
};

// Error codes for the PM flashing stuff
enum {
    PDB_FLASH_ERROR_SUCCESSFUL = 0,
    PDB_FLASH_ERROR_FLASH_BUSY = 1,
    PDB_FLASH_ERROR_PROGRAMMING_ERROR = 2,
    PDB_FLASH_ERROR_WRITE_PROTECTED = 3,
    PDB_FLASH_ERROR_OPERATION_COMPLETE = 4,
    PDB_FLASH_ERROR_OPERATION_TIMEOUT = 5,
    PDB_FLASH_ERROR_VERIFY_FAILURE = 10,
    PDB_FLASH_ERROR_LINE_LENGTH_ERROR = 11,
    PDB_FLASH_ERROR_LINE_CRC_ERROR = 12
};

enum {
    PDB_MAGIC_NUM = 0xa5,
    MAX_PDB_DATA_LENGTH = 253,
};

struct comm_hdr {
    unsigned char       magic;
    unsigned char       len;
    unsigned char       cmd;
} __attribute__((packed));

struct read_temperature_response {
    struct comm_hdr     hdr;
    int                 temp;
    unsigned char       crc8;
} __attribute__((packed));

struct read_voltages_response {
    struct comm_hdr     hdr;
    unsigned int        ivolts;
    unsigned int        cvolts;
    unsigned int        volts_3_3;
    unsigned int        volts_5_0;
    unsigned int        volts_1_0;
    unsigned int        volts_1_9;
    unsigned char       crc_8;
} __attribute__((packed));

struct led_state_response {
    struct comm_hdr     hdr;
    unsigned char       leds;
    unsigned char       state;
    unsigned char       crc_8;
} __attribute__((packed));

struct status_response {
    struct comm_hdr     hdr;
    unsigned char       cap_charger_state;
    unsigned char       cap_charge_level;
    unsigned char       led_state;
    int                 temp;
    unsigned int        ivolts;
    unsigned int        cvolts;
    unsigned int        volts_3_3;
    unsigned int        volts_5_0;
    unsigned int        volts_1_0;
    unsigned int        volts_1_9;
    unsigned char       crc_8;
} __attribute__((packed));

struct pb_released_response {
    struct comm_hdr     hdr;
    unsigned int        time_ms;
} __attribute__((packed));

struct power_lost_response {
    struct comm_hdr     hdr;
    unsigned char       over_under;
} __attribute__((packed));

struct remote_switch_response {
    struct comm_hdr     hdr;
    unsigned char       state;
} __attribute__((packed));

struct cap_discharge_level_response {
    struct comm_hdr     hdr;
    unsigned char       level;
} __attribute__((packed));

struct cred_reset_state_response {
    struct comm_hdr     hdr;
    unsigned char       state;
} __attribute__((packed));

struct shutdown_response {
    struct comm_hdr     hdr;
    unsigned char       mode;
} __attribute__((packed));


/* Firmware update stuff */
struct get_version_request {
    struct comm_hdr     hdr;
    unsigned char       crc8;
} __attribute__((packed));

struct get_version_response {
    struct comm_hdr     hdr;
    unsigned char       application_state;
    unsigned short      major_version;
    unsigned short      minor_version;
    unsigned char       crc8;
} __attribute__((packed));

struct start_flash_sequence_request {
    struct comm_hdr     hdr;
    unsigned char       crc8;
} __attribute__((packed));

struct start_flash_sequence_response {
    struct comm_hdr     hdr;
    unsigned char       error_code;
    unsigned char       crc8;
} __attribute__((packed));

struct erase_flash_request {
    struct comm_hdr     hdr;
    unsigned char       crc8;
} __attribute__((packed));

struct erase_flash_response {
    struct comm_hdr     hdr;
    unsigned char       error_code;
    unsigned char       crc8;
} __attribute__((packed));

struct flash_line_request {
    struct comm_hdr     hdr;
    unsigned char       bytes[240];
    unsigned char       crc8;
} __attribute__((packed));

struct flash_line_response {
    struct comm_hdr     hdr;
    unsigned char       error_code;
    unsigned char       crc8;
} __attribute__((packed));

struct program_flash_request {
    struct comm_hdr     hdr;
    unsigned char       crc8;
} __attribute__((packed));

struct program_flash_response {
    struct comm_hdr     hdr;
    unsigned char       error_code;
    unsigned char       crc8;
} __attribute__((packed));

struct flash_crc_request {
    struct comm_hdr     hdr;
    unsigned long       crc32;
    unsigned char       crc8;
} __attribute__((packed));

struct flash_crc_response {
    struct comm_hdr     hdr;
    unsigned char       error_code;
    unsigned char       crc8;
} __attribute__((packed));

struct flash_done_request {
    struct comm_hdr     hdr;
    unsigned char       crc8;
} __attribute__((packed));

struct flash_done_response {
    struct comm_hdr     hdr;
    unsigned char       error_code;
    unsigned char       crc8;
} __attribute__((packed));

struct set_application_state_request {
    struct comm_hdr     hdr;
    unsigned char       new_state;
    unsigned char       crc8;
} __attribute__((packed));

struct set_application_state_response {
    struct comm_hdr     hdr;
    unsigned char       error_code;
    unsigned char       crc8;
} __attribute__((packed));

struct reset_power_micro_request {
    struct comm_hdr     hdr;
    unsigned char       seconds;
    unsigned char       crc8;
} __attribute__((packed));

struct reset_power_micro_response {
    struct comm_hdr     hdr;
    unsigned char       error_code;
    unsigned char       crc8;
} __attribute__((packed));


typedef int ssize_t;

int open(const char *pathname, int flags, int);
ssize_t write(int fd, const void *buf, size_t count);
ssize_t read(int fd, void *buf, size_t count);

typedef unsigned char   cc_t;
typedef unsigned int    speed_t;
typedef unsigned int    tcflag_t;

enum {
    NCCS = 32,
};

struct termios {
    tcflag_t c_iflag;   /* input mode flags */
    tcflag_t c_oflag;   /* output mode flags */
    tcflag_t c_cflag;   /* control mode flags */
    tcflag_t c_lflag;   /* local mode flags */
    cc_t c_line;        /* line discipline */
    cc_t c_cc[NCCS];    /* control characters */
    speed_t c_ispeed;   /* input speed */
    speed_t c_ospeed;   /* output speed */
};

enum {
    TCSANOW = 0,
    TCIFLUSH = TCSANOW,
    B115200 = 00021,
    CS8 = 00001400,
    CSTOPB = 00002000,
    PARITYON = 0,
    PARITY = 0,
    CLOCAL = 00100000,
    CREAD = 00004000,
    IGNPAR = 0000004,
    VMIN = 6,
    VTIME = 5,
};

int tcsetattr(int fd, int optional_actions, const struct termios *termios_p);
int tcgetattr(int fd, struct termios *termios_p);
int tcflush(int fd, int queue_selector);
int cfsetispeed(struct termios *termios_p, speed_t speed);
int cfsetospeed(struct termios *termios_p, speed_t speed);
void *malloc(size_t size);
void *memmove(void *dest, const void *src, size_t size);
void swab(const void *from, void *to, ssize_t n);
]]

local network_settings_restored

local crc_8_table = ffi.new('unsigned char[256]', {
    0x00, 0x5E, 0xBC, 0xE2, 0x61, 0x3F, 0xDD, 0x83,
    0xC2, 0x9C, 0x7E, 0x20, 0xA3, 0xFD, 0x1F, 0x41,
    0x9D, 0xC3, 0x21, 0x7F, 0xFC, 0xA2, 0x40, 0x1E,
    0x5F, 0x01, 0xE3, 0xBD, 0x3E, 0x60, 0x82, 0xDC,
    0x23, 0x7D, 0x9F, 0xC1, 0x42, 0x1C, 0xFE, 0xA0,
    0xE1, 0xBF, 0x5D, 0x03, 0x80, 0xDE, 0x3C, 0x62,
    0xBE, 0xE0, 0x02, 0x5C, 0xDF, 0x81, 0x63, 0x3D,
    0x7C, 0x22, 0xC0, 0x9E, 0x1D, 0x43, 0xA1, 0xFF,
    0x46, 0x18, 0xFA, 0xA4, 0x27, 0x79, 0x9B, 0xC5,
    0x84, 0xDA, 0x38, 0x66, 0xE5, 0xBB, 0x59, 0x07,
    0xDB, 0x85, 0x67, 0x39, 0xBA, 0xE4, 0x06, 0x58,
    0x19, 0x47, 0xA5, 0xFB, 0x78, 0x26, 0xC4, 0x9A,
    0x65, 0x3B, 0xD9, 0x87, 0x04, 0x5A, 0xB8, 0xE6,
    0xA7, 0xF9, 0x1B, 0x45, 0xC6, 0x98, 0x7A, 0x24,
    0xF8, 0xA6, 0x44, 0x1A, 0x99, 0xC7, 0x25, 0x7B,
    0x3A, 0x64, 0x86, 0xD8, 0x5B, 0x05, 0xE7, 0xB9,
    0x8C, 0xD2, 0x30, 0x6E, 0xED, 0xB3, 0x51, 0x0F,
    0x4E, 0x10, 0xF2, 0xAC, 0x2F, 0x71, 0x93, 0xCD,
    0x11, 0x4F, 0xAD, 0xF3, 0x70, 0x2E, 0xCC, 0x92,
    0xD3, 0x8D, 0x6F, 0x31, 0xB2, 0xEC, 0x0E, 0x50,
    0xAF, 0xF1, 0x13, 0x4D, 0xCE, 0x90, 0x72, 0x2C,
    0x6D, 0x33, 0xD1, 0x8F, 0x0C, 0x52, 0xB0, 0xEE,
    0x32, 0x6C, 0x8E, 0xD0, 0x53, 0x0D, 0xEF, 0xB1,
    0xF0, 0xAE, 0x4C, 0x12, 0x91, 0xCF, 0x2D, 0x73,
    0xCA, 0x94, 0x76, 0x28, 0xAB, 0xF5, 0x17, 0x49,
    0x08, 0x56, 0xB4, 0xEA, 0x69, 0x37, 0xD5, 0x8B,
    0x57, 0x09, 0xEB, 0xB5, 0x36, 0x68, 0x8A, 0xD4,
    0x95, 0xCB, 0x29, 0x77, 0xF4, 0xAA, 0x48, 0x16,
    0xE9, 0xB7, 0x55, 0x0B, 0x88, 0xD6, 0x34, 0x6A,
    0x2B, 0x75, 0x97, 0xC9, 0x4A, 0x14, 0xF6, 0xA8,
    0x74, 0x2A, 0xC8, 0x96, 0x15, 0x4B, 0xA9, 0xF7,
    0xB6, 0xE8, 0x0A, 0x54, 0xD7, 0x89, 0x6B, 0x35
})

local function get_crc_byte(_crc, _byte)
    local idx = bit.bxor(bit.band(_crc, C.CRC_MASK), _byte)
    _crc = bit.bxor(bit.band(bit.rshift(_crc, 8), C.CRC_MASK), crc_8_table[idx])
    return(_crc)
end

local function get_crc_8(p, len)
    local crc = ffi.new('unsigned int', 0)
    while len > 0 do
        crc = get_crc_byte(tonumber(crc), tonumber(p[0]))
        p = p + 1
        len = len - 1
    end
    return crc
end

local led_map = {
    red = C.PDB_SUB_LED_RED,
    green = C.PDB_SUB_LED_GREEN,
    yellow = C.PDB_SUB_LED_YELLOW,
    blue = C.PDB_SUB_LED_BLUE,
}

local led_state_map = {
    off = C.PDB_SUB_LED_OFF,
    on = C.PDB_SUB_LED_ON,
    half_hz = C.PDB_SUB_LED_0_5HZ,
    one_hz = C.PDB_SUB_LED_1_HZ,
    two_hz = C.PDB_SUB_LED_2_HZ,
    four_hz = C.PDB_SUB_LED_4_HZ,
    eight_hz = C.PDB_SUB_LED_8_HZ,
}

local PowerMicro = {}
PowerMicro.__index = PowerMicro

function PowerMicro.new()
    local self = setmetatable({ }, PowerMicro)
    return self
end

-- Internal use function only
function PowerMicro:_set_object(obj)
    self.power_micro = obj
end

-- Set the specified LED to the specified state
function PowerMicro:SetLed(state, led)
    if self.power_micro.leds_locked then return end
    self.power_micro:set_led(state, led)
end

-- Send an ack for power button depressed to PM
function PowerMicro:AckPBDepressed()
    self.power_micro:__send_pdb_packet(
        C.PDB_CMD_PUSHBUTTON_DEPRESSED + C.PDB_CMD_ACK)
end


-- Send an ack for power button released to PM
function PowerMicro:AckPBReleased()
    self.power_micro:__send_pdb_packet(
        C.PDB_CMD_PUSHBUTTON_RELEASED + C.PDB_CMD_ACK)
end


-- Send an ack for Remote Power Button to PM
function PowerMicro:AckRemote()
    self.power_micro:__send_pdb_packet(
        C.PDB_CMD_REMOTE_SWITCH_CHANGED + C.PDB_CMD_ACK)
end


-- Send an ack for power lost 
function PowerMicro:AckPowerLost()
    self.power_micro:__send_pdb_packet(
        C.PDB_CMD_POWER_LOST + C.PDB_CMD_ACK)
end


-- This is the function the PDB uses to request system shutdown.  The powerd
-- process handles everything from that point on.
function PowerMicro:RequestSystemShutdown()
    if self.power_micro.pm.SignalSoftShutdown then
        self.power_micro.pm:SignalSoftShutdown()
    end
    self.power_micro.payload[0] = 0
    self.power_micro:__send_pdb_packet(C.PDB_CMD_SYSTEM_SHUTDOWN, 1)
end


-- Send shutdown finished to the PM
function PowerMicro:ShutdownFinished()
    if not self.sf_sent then
        self.power_micro.payload[0] = 1
        self.power_micro:__send_pdb_packet(C.PDB_CMD_SYSTEM_SHUTDOWN, 1)
        self.sf_sent = true
    end
end


-- Get the level of the capacitor bank
function PowerMicro:GetCapLevel()
    return self.power_micro.cap_level or 0
end


-- Request Channel board power to be turned off
function PowerMicro:ChannelPowerOff()
    self.power_micro.payload[0] = 0
    self.power_micro.payload[1] = 0
    self.power_micro:__send_pdb_packet(C.PDB_CMD_CH_POWER, 2)
end


-- Request Channel board power to be turned on
function PowerMicro:ChannelPowerOn()
    self.power_micro.payload[0] = 0
    self.power_micro.payload[1] = 1
    self.power_micro:__send_pdb_packet(C.PDB_CMD_CH_POWER, 2)
end

function power_micro:set_logger(logger)
    self.logger = logger
end

function power_micro:log(msg, ...)
    if self.logger then
        self.logger:log(self.logger.DEBUG, string.format(msg, ...))
    end
end

function power_micro:error(msg, ...)
    local emsg
    if self.in_fwu then
        local sn = sys_notify:new("Power Micro Firmware Update Failure")
        emsg = string.format("Updating of the power micro firmware from " ..
            "version %d.%d to version %d.%d failed due to '%s'.  Reboot " ..
            "your system to retry the firmware update and if it fails " ..
            "again contact HBM customer service.",
            self.afw.major_version, self.afw.minor_version,
            self.fw.major_version, self.fw.minor_version,
            string.format(msg, ...))
        sn:error(emsg)
    else
        emsg = string.format(msg, ...)
    end
    if self.logger then self.logger:log(self.logger.ERR, emsg) end
    if self.start_cb then self.start_cb() end
end

function power_micro:_handle_pm_reset(buf, len)
    if not buf then
        local msg = string.format("Timeout waiting for 'PM_RESET' response " ..
            "to switch to state '%s'",
            self.final_state_switch and 'APPLICATION' or 'LOADER')
        self:error(msg)
        return
    end
    local resp = ffi.cast('struct reset_power_micro_response *', buf)
    if not self.final_state_switch then
        if resp.error_code ~= C.PDB_LOADER_STATE then
            self:error("PM_RESET failed:  Failed to switch to loader mode")
            return
        end
        self:_start_flash()
    else
        if resp.error_code ~= C.PDB_APPLICATION_STATE then
            self:error("PM_RESET failed:  Failed to switch to application mode")
            return
        end
        self.logger:log(self.logger.INFO, "PM firmware update complete")
        local sn = sys_notify:new("Power Micro Firmware Update Successful")
        sn:info(string.format("The firmware for the power micro was " ..
            "successfully updated to version %d.%d",
            self.fw.major_version, self.fw.minor_version))
        self.afw.major_version = self.fw.major_version
        self.afw.minor_version = self.fw.minor_version
        if self.start_cb then self.start_cb() end
    end
end

function power_micro:_do_pm_reset()
    self.payload[0] = 0
    self:__send_pdb_packet(C.PDB_CMD_RESET_POWER_MICRO, 1,
        self._handle_pm_reset, 20)
end

function power_micro:_handle_switch_state_response(buf, len)
    if not buf then
        self:error("Timeout waiting for 'SET_APPLICATION_STATE' response")
        return
    end
    local resp = ffi.cast('struct set_application_state_response *', buf)
    if resp.error_code ~= C.PDB_FLASH_ERROR_SUCCESSFUL then
        self:error("SET_APPLICATION_STATE failed %d", resp.error_code)
        return
    end
    self:_do_pm_reset()
end

-- Function used to switch the state of the power micro
function power_micro:_switch_state(state, final)
    self.payload[0] = state
    self.final_state_switch = final
    self:__send_pdb_packet(C.PDB_CMD_SET_APPLICATION_STATE,
        ffi.sizeof('unsigned char'), self._handle_switch_state_response, 1)
end

-- Function used to handle responses to FLASH DONE requests
function power_micro:_handle_flash_done_response(buf, len)
    if not buf then
        self:error("Timeout waiting for 'FLASH_DONE' response")
        return
    end
    local resp = ffi.cast('struct flash_done_response *', buf)
    if resp.error_code ~= C.PDB_FLASH_ERROR_SUCCESSFUL then
        self:error("FLASH_DONE failed %d", resp.error_code)
        return
    end
    -- FIXME:  Remove the firmware image file?
--    os.remove(self.fw.image)
    self:_switch_state(C.PDB_APPLICATION_STATE, true)
end

-- Function used to handle responses to FLASH CRC requests
function power_micro:_handle_flash_crc_response(buf, len)
    if not buf then
        self:error("Timeout waiting for 'FLASH_CRC' response")
        return
    end
    local resp = ffi.cast('struct flash_crc_response *', buf)
    if resp.error_code ~= C.PDB_FLASH_ERROR_SUCCESSFUL then
        self:error("FLASH_CRC failed %d", resp.error_code)
        return
    end
    self:__send_pdb_packet(C.PDB_CMD_FLASH_DONE, 0,
        self._handle_flash_done_response, 1.5)
end

-- Function used to send a FLASH CRC request
function power_micro:_do_set_crc()
    ffi.copy(self.payload, self.fw.crc32, ffi.sizeof(self.fw.crc32))
    self:__send_pdb_packet(C.PDB_CMD_FLASH_CRC,
        ffi.sizeof(self.fw.crc32), self._handle_flash_crc_response, 1)
end

-- Function used to handle responses to PROGRAM FLASH requests
function power_micro:_handle_program_flash(buf, len)
    if not buf then
        self.in_retry = self.in_retry or 0
        if self.in_retry < 4 then
--[[
            self.logger:log(self.logger.NOTICE,
                "Timeout waiting for 'PROGRAM_FLASH' response - retrying")
]]
            self.in_retry = self.in_retry + 1
            self:_check_firmware()
        else
            self:error("Timeout waiting for 'PROGRAM_FLASH' response")
            self.in_retry = nil
        end
        return
    end
    self.in_retyr = nil
    local resp = ffi.cast('struct program_flash_response *', buf)
    if resp.error_code ~= C.PDB_FLASH_ERROR_SUCCESSFUL then
        self:error("PROGRAM_FLASH failed %d", resp.error_code)
        return
    end
    self.fw.offset = self.fw.offset + self.fw.amount2send
    self.fw.progress = self.fw.offset / self.fw.size * 100
    if self.fw.offset < self.fw.size then
        self:_flash_line()
    else
        self:_do_set_crc()
    end
end

-- Function used to handle responses to FLASH LINE requests
function power_micro:_handle_flash_line_response(buf, len)
    if not buf then
        self.in_retry = self.in_retry or 0
        if self.in_retry < 4 then
--[[
            self.logger:log(self.logger.NOTICE,
                "Timeout waiting for 'FLASH_LINE' response - retrying")
]]
            self.in_retry = self.in_retry + 1
            self:_check_firmware()
        else
            self:error("Timeout waiting for 'FLASH_LINE' response")
            self.in_retry = nil
        end
        return
    end
    self.in_retry = nil
    local resp = ffi.cast('struct flash_line_response *', buf)
    if resp.error_code ~= C.PDB_FLASH_ERROR_SUCCESSFUL then
        self:error("FLASH_LINE failed %d", resp.error_code)
        return
    end
    self:__send_pdb_packet(C.PDB_CMD_PROGRAM_FLASH, 0,
        self._handle_program_flash, 1)
end

-- Function used to send FLASH LINE requests
function power_micro:_flash_line()
    self.fw.amount2send = self.fw.size - self.fw.offset
    if self.fw.amount2send > 240 then
        self.fw.amount2send = 240
    end
    ffi.copy(self.payload, self.fw.buffer + self.fw.offset, self.fw.amount2send)
    self:__send_pdb_packet(C.PDB_CMD_FLASH_LINE, self.fw.amount2send,
        self._handle_flash_line_response, 1.5)
end

-- Function used to handle responses to ERASE FLASH requests
function power_micro:_handle_erase_flash_response(buf, len)
    if not buf then
        self:error("Timeout waiting for 'ERASE_FLASH' response")
        return
    end
    local resp = ffi.cast('struct erase_flash_response *', buf)
    if resp.error_code ~= C.PDB_FLASH_ERROR_SUCCESSFUL then
        self:error("ERASE_FLASH failed %d", resp.error_code)
        return
    end
    self:_flash_line()
end

-- Function used to handle responses to START FLASH requests
function power_micro:_handle_start_flash_response(buf, len)
    if not buf then
        self:error("Timeout waiting for 'START_FLASH_SEQUENCE' response")
        return
    end
    local resp = ffi.cast('struct start_flash_sequence_response *', buf)
    if resp.error_code ~= C.PDB_FLASH_ERROR_SUCCESSFUL then
        self:error("START_FLASH_SEQUENCE failed %d", resp.error_code)
        return
    end
    self:__send_pdb_packet(C.PDB_CMD_ERASE_FLASH, 0,
        self._handle_erase_flash_response, 10)
end

-- Function used to send START FLASH requests
function power_micro:_start_flash()
    self.logger:log(self.logger.INFO,
        string.format("Updating PM firmware to version %d.%d",
            self.fw.major_version, self.fw.minor_version))

    self.leds_locked = true
    self:set_led(self.in_retry and "one_hz" or "eight_hz", "red", function()
            self:set_led("four_hz", "yellow", function()
                    self:set_led("four_hz", "green", function()
                            self:set_led("eight_hz", "blue", function()
                                    self:__send_pdb_packet(
                                        C.PDB_CMD_START_FLASH_SEQUENCE, 0,
                                        self._handle_start_flash_response, 1.5)
                                end)
                        end)
                end)
        end)
--[[
    self:__send_pdb_packet(C.PDB_CMD_START_FLASH_SEQUENCE, 0,
        self._handle_start_flash_response, 1.5)
]]
end

-- Functin used to handle responses to GET VERSION requests
function power_micro:_handle_get_version_response(buf, len)
    if not buf then
        self:error("Timed out waiting for reponse to get version request")
        if self.start_cb then self.start_cb() end
        return
    end
    local resp = ffi.cast('struct get_version_response *', buf)
    self.afw.major_version = resp.major_version
    self.afw.minor_version = resp.minor_version
    if config.force_pm_fw_updates then
        if resp.application_state == C.PDB_LOADER_STATE then
            self:_start_flash()
        else
            self:_switch_state(C.PDB_LOADER_STATE)
        end
    elseif self.afw.major_version == self.fw.major_version and
        self.afw.minor_version == self.fw.minor_version and
        not self.in_retry and self.start_cb then
        self.start_cb()
    elseif resp.application_state == C.PDB_LOADER_STATE then
        self:_start_flash()
    else
        self:_switch_state(C.PDB_LOADER_STATE)
    end
end

function power_micro:_check_firmware()
    self:reinit_tty()

    -- Get the firmware file and generate information needed
    self.in_fwu = true
    self.fw = self.fw or {major_version=0,minor_version=0}
    self.afw = ffi.new('struct get_version_response')

    local image_file = '/usr/share/power-micro/power-micro.bin'
    local f = io.open(image_file, "r")
    if not f then
        -- FIXME:  Maybe not treat this as error but normal condition.  E.G.
        -- after successful update, file is removed?
        self:error("Unable to open power micro firmware binary image")
        return
    end
    local fdata = f:read('*a')
    f:close()
    local size = fdata:len()
    self.fw.image = image_file
    self.fw.size = size
    if size % 4 ~= 0 then
        self.fw.size = ((size / 4) + 1) * 4
    end
    self.fw.offset = 0
    self.fw.buffer = ffi.new('unsigned char[?]', self.fw.size, fdata)
    self.fw.crc32 =
        ffi.new('unsigned int[1]', zlib.crc32(0, self.fw.buffer, self.fw.size))

    -- CRC needs to be swizzled around for PM
    local cptr = ffi.cast('unsigned char *', self.fw.crc32)
    cptr[0] = bit.bxor(tonumber(cptr[0]), tonumber(cptr[3]))
    cptr[3] = bit.bxor(tonumber(cptr[0]), tonumber(cptr[3]))
    cptr[0] = bit.bxor(tonumber(cptr[0]), tonumber(cptr[3]))
    C.swab(cptr + 1, cptr + 1, 2)

    -- Firmware version is located 0x200 bytes into the file
    -- Byte order from file is reversed from what arrives in get_version resp
    local sptr = ffi.new('unsigned short[2]')
    ffi.copy(sptr, self.fw.buffer + 0x200, 4)
    C.swab(sptr, sptr, 2)
    C.swab(sptr + 1, sptr + 1, 2)
    self.fw.major_version = sptr[0]
    self.fw.minor_version = sptr[1]
    fdata = nil

    -- Get the version of the firmware on the power micro
    self:__send_pdb_packet(C.PDB_CMD_GET_VERSION, 0,
        self._handle_get_version_response, .5)
end

function power_micro:reinit_tty()
    self.rx_watcher:stop(ev.Loop.default)

    C.close(self.fd)

    self.fd = C.open("/dev/ttyS0", C.O_RDWR, 0)
    if self.fd < 0 then
        error(string.format("Unable to open serial port %s for PM", tty))
    end
    self.oldtio = ffi.new('struct termios[1]')
    self.newtio = ffi.new('struct termios[1]')
    C.tcgetattr(self.fd, self.oldtio)

    self.newtio[0].c_iflag = C.IGNPAR
    self.newtio[0].c_oflag = 0
    self.newtio[0].c_lflag = 0
    self.newtio[0].c_cflag = C.B115200 + C.CS8 + C.CSTOPB + C.PARITYON +
        C.PARITY + C.CLOCAL + C.CREAD
    self.newtio[0].c_cc[C.VMIN] = 0
    self.newtio[0].c_cc[C.VTIME] = 0
    C.tcflush(self.fd, C.TCIFLUSH)
    C.cfsetospeed(self.newtio, C.B115200)
    C.cfsetispeed(self.newtio, C.B115200)
    C.tcsetattr(self.fd, C.TCSANOW, self.newtio)

    self.hdr = ffi.cast('struct comm_hdr *', self.pbuf)
    self.payload = self.pbuf + self.comm_hdr_size

    self.rbuf.offset = 0

    self.rx_watcher = ev.IO.new(function(loop, _io, revents)
        if self.rbuf.size - self.rbuf.offset <= 0 then
            error("Out of receive buffer space")
            _io:stop(ev.Loop.default)
            return
        end
        local count = C.read(self.fd, self.rbuf.buf + self.rbuf.offset,
            self.rbuf.size - self.rbuf.offset)
        if count < 0 then
            local e = ffi.errno()
            if e ~= C.WOULDBLOCK and e ~= C.EAGAIN then
                error(string.format("Rx error: %d ", e))
            end
        elseif count > 0 then
            self.rbuf.offset = self.rbuf.offset + count
            repeat
                local msg = self:get_message()
                if msg and self.callback then
                    self.callback(msg)
                end
            until not msg
        end
    end, self.fd, ev.READ)
    self.rx_watcher:start(ev.Loop.default)

end

function power_micro:process_status_event(event_tbl)
    if not event_tbl then return end

    if event_tbl.test and event_tbl.test.error then
        -- Red LED
        self:set_led('on', 'red')
    end

    if event_tbl.system_status then
        -- yellow LED
        if event_tbl.system_status.test then
            if event_tbl.system_status.test.running then
                self:set_led('eight_hz', 'yellow')
            else
                self:set_led('off', 'yellow')
            end
        else
            self:set_led('off', 'yellow')
        end

        -- Red LED
        if event_tbl.system_status.error.reset then
            self:set_led('off', 'red')
        elseif event_tbl.system_status.disk and
            event_tbl.system_status.disk.full then
            self:set_led('one_hz', 'red')
        elseif event_tbl.system_status.error and
            event_tbl.system_status.error.error then
            self:set_led('on', 'red')
        else
            self:set_led('off', 'red')
        end
    end

    -- Blue LED
    if event_tbl.test and event_tbl.test.error then
        self:set_led('off', 'blue')
    elseif event_tbl.timing then
        if not event_tbl.timing.regulated then
            self:set_led('one_hz', 'blue')
        elseif event_tbl.timing.source == 'GPS' then
            self:set_led('four_hz', 'blue')
        elseif event_tbl.timing.source == 'PTP' then
            self:set_led('on', 'blue')
        end
    end

    -- Green LED
    if ((event_tbl.reboot_poweroff and event_tbl.reboot_poweroff.reboot) or
        (event_tbl.test and event_tbl.test.error)) then
        self:set_led('four_hz', 'green')
    end
end

function power_micro:initialize(logger, callback, rate, tty)
    self.logger = logger
    rate = rate or 1

    self.pm = PowerMicro.new()
    self.pm:_set_object(self)
    self.callbacks = {}

    self.callback = callback
    self.comm_hdr_size = ffi.sizeof('struct comm_hdr')
    self.fd = C.open(tty or "/dev/ttyS0", C.O_RDWR, 0)
    if self.fd < 0 then
        error(string.format("Unable to open serial port %s for PM", tty))
    end
    self.oldtio = ffi.new('struct termios[1]')
    self.newtio = ffi.new('struct termios[1]')
    C.tcgetattr(self.fd, self.oldtio)

    self.newtio[0].c_iflag = C.IGNPAR
    self.newtio[0].c_oflag = 0
    self.newtio[0].c_lflag = 0
    self.newtio[0].c_cflag = C.B115200 + C.CS8 + C.CSTOPB + C.PARITYON +
        C.PARITY + C.CLOCAL + C.CREAD
    self.newtio[0].c_cc[C.VMIN] = 0
    self.newtio[0].c_cc[C.VTIME] = 0
    C.tcflush(self.fd, C.TCIFLUSH)
    C.cfsetospeed(self.newtio, C.B115200)
    C.cfsetispeed(self.newtio, C.B115200)
    C.tcsetattr(self.fd, C.TCSANOW, self.newtio)

    self.pbuf = ffi.new('unsigned char[?]', C.MAX_PDB_DATA_LENGTH + 5)
    self.hdr = ffi.cast('struct comm_hdr *', self.pbuf)
    self.payload = self.pbuf + self.comm_hdr_size

    self.rbuf = {}
    self.rbuf.size = 1024
    self.rbuf.buf = ffi.cast('unsigned char *', C.malloc(self.rbuf.size))
    self.rbuf.offset = 0

    self.rx_watcher = ev.IO.new(function(loop, _io, revents)
        if self.rbuf.size - self.rbuf.offset <= 0 then
            error("Out of receive buffer space")
            _io:stop(ev.Loop.default)
            return
        end
        local count = C.read(self.fd, self.rbuf.buf + self.rbuf.offset,
            self.rbuf.size - self.rbuf.offset)
        if count < 0 then
            local e = ffi.errno()
            if e ~= C.WOULDBLOCK and e ~= C.EAGAIN then
                error(string.format("Rx error: %d ", e))
            end
        elseif count > 0 then
--[[
io.stderr:write("RX: ")
for i = 0, count - 1 do
io.stderr:write(string.format(" %02X", self.rbuf.buf[self.rbuf.offset + i]))
end
io.stderr:write("\n")
]]
            self.rbuf.offset = self.rbuf.offset + count
            repeat
                local msg = self:get_message()
                if msg and self.callback then
                    self.callback(msg)
                end
            until not msg
        end
    end, self.fd, ev.READ)
    self.rx_watcher:start(ev.Loop.default)

    self.cap_level = 0

    self.leds_as_set = {}
    self.leds_as_replied = {}
    for k, v in pairs(led_map) do
        self.leds_as_set[k] = false
        self.leds_as_replied[k] = false
    end

    self.led_status_timer = ev.Timer.new(
        function(l, t, e)
            self:check_led_status()
        end, 0.5, 0.5)

    self.start_cb = function()

        -- Write the version information to a file for consumption by outside
        -- entities
        local fh = io.open('/tmp/power-micro-version.json', 'w')
        if fh then
            fh:write(cjson.encode({
                    running = {
                        major = self.afw.major_version,
                        minor = self.afw.minor_version
                    },
                    expected = {
                        major = self.fw.major_version,
                        minor = self.fw.minor_version,
                    }
                }))
            fh:close()
        end

        self.in_fwu = nil

        -- If the network settings were reset to factory default then the
        -- system must be rebooted.
        if network_settings_restored then
            self:log("Rebooting to make reset network settings take effect")
            os.execute('sync;sync;/sbin/reboot')
        end

        self.sl = status_lib:new('leds', function(json)
                self:process_status_event(cjson.decode(json))
            end)

        self.sl:set_status({
            blue = 'off',
            green = 'on',
            yellow = 'off',
            red = 'off',
        })

        self.start_cb = nil
        self:set_led("off", "red")
        self:set_led("off", "yellow")
        self:set_led("off", "blue")
        self:set_led("on", "green")
        self.leds_locked = false

        -- This stuff only happens if we do NOT start a firmware update
        local t_ms = ffi.cast('unsigned short *', self.payload)
        t_ms[0] = (1000/rate)
        self:__send_pdb_packet(C.PDB_CMD_STATUS_TIME_SET,
            ffi.sizeof('unsigned short'))
    end

    -- This must be deferred until the power daemon finishes its stuff!
    ev.Timer.new(function(l, t, e)
            -- Check for credential or network settings restore
--            self:reinit_tty()
            self:__send_pdb_packet(C.PDB_CMD_PB_GET_CREDENTIAL_RESET_STATE, 0,
                self._check_firmware, .5)
        end, 10, 0):start(ev.Loop.default)
end

function power_micro:check_led_status()
    local ok = true
    for k, v in pairs(self.leds_as_set) do
        if self.leds_as_replied[k] ~= v then
            self:set_led(v, k)
            ok = false
        end
    end
    if ok then
        self.led_status_timer_started = false
    else
        self.led_status_timer:again(ev.Loop.default)
    end
end

-- Internal use only function to send a packet to the PM
function power_micro:__send_pdb_packet(cmd, dlen, cb, timeout)
    dlen = dlen or 0
    if dlen > C.MAX_PDB_DATA_LENGTH then
        error(string.format("PM packet too long: %d", dlen))
        return
    end
    self.hdr.magic = C.PDB_MAGIC_NUM
    self.hdr.len = dlen + 1
    self.hdr.cmd = cmd
    self.pbuf[dlen + self.comm_hdr_size] =
        get_crc_8(self.pbuf, self.comm_hdr_size + dlen)
    if cb then
        self.callbacks = self.callbacks or {}
        self.callback_timer = self.callback_timer or {}
        self.callbacks[cmd] = cb
        self.callback_timer[cmd] = ev.Timer.new(function(l, t, e)
                if self.callbacks[cmd] then
                    cb(self)
                end
            end, timeout or .5, 0)
    end
    local amount2send = self.comm_hdr_size + self.hdr.len
    local sent = C.write(self.fd, self.pbuf, amount2send)
    if sent <= 0 then
        local e = ffi.errno()
        error(string.format("Serial port write fail for PM: %d", e))
    elseif sent ~= amount2send then
        self:log(string.format("Short write %d -> %d", amount2send, sent))
    elseif cb then
        self.callback_timer[cmd]:start(ev.Loop.default)
    end
--[[
io.stderr:write("TX: ")
for i = 0, amount2send - 1 do
io.stderr:write(string.format(" %02X", self.pbuf[i]))
end
io.stderr:write("\n")
]]
end

function power_micro:set_led(state, led, cb)
    if not led_map[led] or not led_state_map[state] then
        self:error("Illegal set_led(%s, %s)", state, led)
        if cb then cb() end
        return
    end
    if self.leds_as_replied[led] == state then
        if cb then cb() end
        return
    end
    self.leds_as_set[led] = state
    self.payload[0] = led_map[led]
    self.payload[1] = led_state_map[state]
    self:__send_pdb_packet(C.PDB_CMD_SET_LED_STATE, 2, cb)
    if not self.led_status_timer_started then
        self.led_status_timer:again(ev.Loop.default)
        self.led_status_timer_started = true
    end
end

function power_micro:get_led(led)
    return self.leds_as_replied[led] or 'unknown'
end

local cmd_map = {
    [0] = 'Voltages',
    [1] = 'LED State',
    [2] = 'One Wire',
    [3] = 'Set Parameter',
    [4] = 'Read State',
    [5] = 'Channel Power',
    [6] = 'System Shutdown',
    [7] = 'Temperature',
    [8] = 'Status Time Set',

    [9] = 'Credential Reset',

    -- Below are unsolicited events
    [32] = 'Power Lost',
    [33] = 'Button Press',
    [34] = 'Button Released',
    [35] = 'Remote Power Switch',
    [36] = 'Caps Charged',
    [37] = 'Caps Depleted',
    [38] = 'Cap Discharge Level',
    [40] = 'Status',

    [0xd] = 'Get Version',
    [0x70] = 'Start Flash Sequence',
    [0x71] = 'Erase Flash',
    [0x72] = 'Flash Line',
    [0x73] = 'Program Flash',
    [0x74] = 'Flash CRC',
    [0x75] = 'Flash Done',
    [0x76] = 'Set Application State',
    [0x77] = 'Reset Power Micro'
}

-- Function used to resynchronize the input data stream
function power_micro:_resync_input(start)
    -- Try to find the magic marker in the buffer
    local index
    for i = (start or 0), self.rbuf.offset - 1 do
        if self.rbuf.buf[i] == C.PDB_MAGIC_NUM then
            index = i
            break
        end
    end

    -- Magic not found at all, dump entire buffer
    if not index then
        self:log("No magic number found in buffer during resync")
        self.rbuf.offset = 0
        return
    end

    -- Copy data from the found magic through the end of the buffer to
    -- the bottom of buffer then try processing that data
    self:log(string.format("Magic number found at position %d", index))
    self.rbuf.offset = self.rbuf.offset - index
    C.memmove(self.rbuf.buf, self.rbuf.buf + index, self.rbuf.offset)
    return self:get_message()
end


-- Function used to get and process a message from the input buffer
function power_micro:get_message()
    while self.rbuf.offset > self.comm_hdr_size do
        local consumed = 0
        local hdr = ffi.cast('struct comm_hdr *', self.rbuf.buf)

        -- First make sure the magic number is valid, if not try to find it in
        -- the data available in the buffer.
        if hdr.magic ~= C.PDB_MAGIC_NUM then
            self:log("First byte in RX buffer is not magic magic number")
            return self:_resync_input()
        end

        -- See if we have the entire message, return if we don't
        if self.rbuf.offset < hdr.len + self.comm_hdr_size then
            return
        end

        -- Calculate the CRC8 on everyting in the message EXCEPT the last byte
        consumed = hdr.len + 2
        local crc = get_crc_8(self.rbuf.buf, consumed)

        -- If the CRC is wrong, try resyncing
        if crc ~= self.rbuf.buf[consumed] then
            self:log(string.format("Bad CRC: calculated %02X, received %02X",
                crc, self.rbuf.buf[consumed]))
            local str = 'Buffer content: '
            for i = 0, self.rbuf.offset - 1 do
                str = str .. string.format(" %02X", self.rbuf.buf[i])
            end
            self:log(str)
            return self:_resync_input(1)
        end

        -- Adjust amount consumed for the CRC byte
        consumed = consumed + 1
        local ret = {}
        local cmd = bit.band(tonumber(hdr.cmd), 0x7f)
        ret.cmd = cmd_map[cmd] or "Unknown"
        local dlen = hdr.len - 1
        if ret.cmd == 'Temperature' then
            local rtr = ffi.cast('struct read_temperature_response *',
                self.rbuf.buf)
            ret.temperature = tonumber(rtr.temp) / 1000.0
        elseif ret.cmd == 'Voltages' then
            local rvr = ffi.cast('struct read_voltages_response *',
                self.rbuf.buf)
            ret.ivolts = tonumber(rvr.ivolts) / 1000.0
            ret.cvolts = tonumber(rvr.cvolts) / 1000.0
            ret.volts_3 = tonumber(rvr.volts_3_3) / 1000.0
            ret.volts_5 = tonumber(rvr.volts_5_0) / 1000.0
            ret.volts_1 = tonumber(rvr.volts_1_0) / 1000.0
            ret.volts_1_9 = tonumber(rvr.volts_1_9) / 1000.0
        elseif ret.cmd == 'LED State' then
            local rlr = ffi.cast('struct led_state_response *', self.rbuf.buf)
            for led, v in pairs(led_map) do
                if v == bit.band(tonumber(rlr.leds), v) then
                    self.leds_as_replied[led] = false
                    for state, v1 in pairs(led_state_map) do
                        if v1 == rlr.state then
                            self.leds_as_replied[led] = state
                            if self.sl then
                                self.sl:set_status(self.leds_as_replied)
                            end
                        end
                    end
                    if not self.leds_as_replied[led] then
                        self:log("Bad LED State rlr.leds=%s, rlr.state=%s",
                            rlr.leds, rlr.state)
                    end
                end
            end
            ret.led_state = {}
            for k, v in pairs(self.leds_as_replied) do
                ret.led_state[k] = v or 'unknown'
            end
        elseif ret.cmd == 'Channel Power' and self.pm.ChannelPowerAck then
            self.pm:ChannelPowerAck()
        elseif ret.cmd == 'Button Press' and self.pm.ButtonDepressed then
            self.pm:ButtonDepressed()
        elseif ret.cmd == 'System Shutdown' and self.pm.ShutdownAck then
            local sts = ffi.cast('struct shutdown_response *', self.rbuf.buf)
            self.pm:ShutdownAck(tonumber(sts.mode))
        elseif ret.cmd == 'Status' then
            local sts = ffi.cast('struct status_response *', self.rbuf.buf)
            ret.ivolts = tonumber(sts.ivolts) / 1000.0
            ret.cvolts = tonumber(sts.cvolts) / 1000.0
            ret.volts_3 = tonumber(sts.volts_3_3) / 1000.0
            ret.volts_5 = tonumber(sts.volts_5_0) / 1000.0
            ret.volts_1 = tonumber(sts.volts_1_0) / 1000.0
            ret.volts_1_9 = tonumber(sts.volts_1_9) / 1000.0
            ret.temperature = tonumber(sts.temp) / 1000.0
-- NOTE:  This state information is simply whether or not the LED was ON or OFF
-- at the time the status message was generated on the PM.
--[[
            for led, v in pairs(led_map) do
                if v == bit.band(tonumber(sts.led_state), v) then
                    for state, v1 in pairs(led_state_map) do
                        if v1 == sts.led_state then
                            self.leds_as_replied[led] = state
                        end
                    end
                end
            end
]]
            ret.cap_level = tonumber(sts.cap_charge_level)
            self.cap_level = ret.cap_level
            ret.charger_state = tonumber(sts.cap_charger_state)
        elseif ret.cmd == 'Power Lost' and self.pm.PowerLost then
            local sts = ffi.cast('struct power_lost_response *', self.rbuf.buf)
            local mode = 'over'
            if sts.over_under == 0 then
                mode = 'under'
            end
            self.pm:PowerLost(mode)
        elseif ret.cmd == 'Remote Power Switch' and
            self.pm.RemotePowerStateChange then
            local sts = ffi.cast('struct remote_switch_response *',
                self.rbuf.buf)
            local mode = 'not grounded'
            if sts.state == 0 then
                mode = 'grounded'
            end
            self.pm:RemotePowerStateChange(mode)
        elseif ret.cmd == 'Cap Discharge Level' then
            local sts = ffi.cast('struct cap_discharge_level_response *',
                self.rbuf.buf)
            self.cap_level = math.floor(tonumber(sts.level))
        elseif ret.cmd == 'Caps Charged' then
            self.cap_level = 100
        elseif ret.cmd == 'Caps Depleted' then
            -- If we get this message we have < 500us to do anything!
        elseif ret.cmd == 'Credential Reset' and
            self.pm.RestoreAdminCredentials then
            local sts = ffi.cast('struct cred_reset_state_response *',
                self.rbuf.buf)
            if tonumber(sts.state) == 1 then
                self.pm:RestoreAdminCredentials()
            elseif tonumber(sts.state) == 2 then
                network_settings_restored = self.pm:ResetNetworkParameters()
            end
        elseif ret.cmd == 'Status Time Set' then
            -- Ignore this message
        elseif (ret.cmd == 'Get Version' or ret.cmd == 'Start Flash Sequence' or
            ret.cmd == 'Erase Flash' or ret.cmd == 'Flash Line' or
            ret.cmd == 'Program Flash' or ret.cmd == 'Flash CRC' or
            ret.cmd == 'Flash Done' or ret.cmd == 'Set Application State' or
            ret.cmd == 'Reset Power Micro') and
            self.callbacks and self.callbacks[cmd] then
            if self.callback_timer and self.callback_timer[cmd] then
                self.callback_timer[cmd]:stop(ev.Loop.default)
                self.callback_timer[cmd]:clear_pending(ev.Loop.default)
                self.callback_timer[cmd] = nil
            end
            local cb = self.callbacks[cmd]
            self.callbacks[cmd] = nil
            cb(self, self.rbuf.buf, self.comm_hdr_size + dlen)
        else
            self:log(string.format("Unknown command %02X", tonumber(hdr.cmd)))
            ret.data = ffi.new('char[?]', dlen)
            C.memmove(ret.data, self.rbuf.buf + self.comm_hdr_size, dlen)
        end
        self.rbuf.offset = self.rbuf.offset - consumed
        C.memmove(self.rbuf.buf, self.rbuf.buf + consumed, self.rbuf.offset)
        return ret
    end
end

return power_micro
