--[[
	Lua module for importing sdbx file
--]]

local i =  {}
i.schema_info = require "schema_info"
log = require "log"

function i:run( filename)
	local sdbx_text = self:file_validation( filename)
	local sdbx_rows ,trash_count = self:parse( sdbx_text)

	if trash_count > 0 then
		domovoi:trash_alert( trash_count)
	end

	return sdbx_rows ,trash_count
end


function i:file_validation( file_name)
	local sdbx_text
	if not file_name then
		domovoi:failed( "No input file given for mdb_translation." , true)
	end

	if file_name:match( ".sdbx" ) then
		sdbx_text = domovoi:read( file_name )
		if not file_name:match( "storage") then
			domovoi:make_storage()
			os.execute( 'mv "'..file_name..'" /storage/sdbx/.' )
		end
	elseif file_name:match( ".sdb" ) then
		domovoi:failed( 
			"Mdb-translation does not currently support .sdb files. File rejected." ,true)
	else
		domovoi:failed( 
			"No .sdbx file found! Check selected file type." ,true)
	end

	if not sdbx_text then
		local msg = "Couldn't read input file for mdb-translation, not found or improperly formatted."
		domovoi:failed( msg , true )
	end

	--local header = sdbx_text:match( "(.-)%c")
	sdbx_text = sdbx_text:match( ".-\n(.+)")

	return sdbx_text
end


function i:parse( text)
	local lines , schema = {} , {}
	local garbage_count = 0

	--get list of expected tables
	for table_name,_ in pairs( self.schema_info ) do
		schema[ table_name ] = {}
	end

	for newline in text:gmatch( "[%C\t]+" ) do
		newline = newline .. global_delim
		local newline_is_not_empty ,delimiter_count = self:validate_newline( newline)

		if newline_is_not_empty then
			local iterated_line = 
				self:parse_line_to_iterated_pair( newline)
			local is_schema = self:line_is_schema( iterated_line)

			if is_schema then
				schema[ is_schema ] = iterated_line

			else --not schema, but data row. Find matching table and parse accordingly
				local current_table , current_table_name = nil , nil
				local table_match , paired_line = false , {}
 				--change to exact match, for error reporting
				for table_name , table_info in pairs( self.schema_info ) do
					if delimiter_count == table_info.size or delimiter_count == table_info.size - 1 then
						current_table = schema[ table_name ]
						current_table_name = table_name
						table_match = true
						break
					end
				end

				if not current_table or not table_match then
					local name = iterated_line[1]
					if name then
						log:warn( 'SDBX sensor line "'..name
							..'" was not completley readable and was not loaded.' )
					end
					garbage_count = garbage_count + 1
				else
					self:sanitize_values( iterated_line ,paired_line ,current_table)
					lines[ paired_line] = current_table_name
				end
			end
		end
	end
	self:were_all_schema_found()
	return lines ,garbage_count
end


--===== Helpers ==========
	function i:validate_newline( newline)
		local match_count = 0
		for _ in newline:gmatch( global_delim ) do
			match_count = match_count + 1
		end
		return (match_count > 1) ,match_count
	end

	function i:parse_line_to_iterated_pair( line)
		local iterated_line = {}
		for value in line:gmatch( "(.-)"..global_delim ) do
			table.insert( iterated_line ,value)
		end
		return iterated_line
	end

	function i:sanitize_values( iterated_line ,paired_line ,current_table)
		for i,value in ipairs( iterated_line) do
			--if value > 8, then whether number or string, needs to be string for formatting.
			value = (#value > 8 and (value:match("'(.+)'") or value))
					or tonumber(value) 
					or value:match( "'(.+)'")
					or value
					or ""

			if type(value) == 'string' and value:match([[\]]) then
				value = value:gsub( [[\]] ,[[\\]])
				print( value)
			end

			local column = current_table[i]:match( '"?([%w_]+)"?')
			paired_line[ column] = value
		end
	end

	function i:line_is_schema( line )
		for _ , b in ipairs( line ) do
			for table_name , a in pairs( self.schema_info ) do
				if string.match( b ,a.sample_column ) then
					self.schema_info[ table_name ].validated = true
					return table_name
				end
			end
		end
	end

	function i:were_all_schema_found()
		for table_name ,table_info in pairs( self.schema_info ) do
			if not table_info.validated then
				domovoi:failed( 
					"Could not find all three schema. Missing at least "
					..table_name.."." )
			end
		end
	end



return i
