--[[
	Holds majority of the sxr import rules, per document SomatXR HBM SDB Sensor Import (Rev 1), 28th December 2014
	supported_bridges are the supported bridge type enums
	scale_types are the different types of sensor scaling we support. Each has it's own set of rules.
	same for ss. Each supported sensor has a number of scaling types associated with it, hence the previous data struct.
--]]
local ss

--================== Supported Bridge Bridge Enumerations =====
local supported_bridges =
	{ 362 ,363 ,364 ,365 ,366 ,367 ,368 ,369 ,371 ,372 ,350 ,351 ,352 ,353 ,354}

local mx1601 = "source/hbm_mx1601b" 
local mx1615 = "source/hbm_mx1615b"
local mx1609 = "source/hbm_mx1609b" 
local mx840 = "source/hbm_mx840b"

--loops and calls all overlaps -> categories -> fields
--for _category_specific setup
--there's some extra stuff here to check for successive failures 
--in the same field- since we can only fail a sensor once
--it's failed in the same field for each category (if multiple)
local function setup_category_specifics( sensor ,sxr_row ,args)
	sxr_row._category_specific = {} 
	local failures = {}
	local cc = 0 
	for category ,fields in pairs( ss[ sensor].overlaps) do
		sxr_row._category_specific[ category] = {}
		cc = cc + 1

		for field ,fn in pairs( fields) do
			sxr_row[ field] = nil
			local success = fn( sxr_row ,args)
			if not success then
				failures[ category] = true
			end
		end
	end
	print("Category Count: " ,cc)

	for category,_ in pairs( failures) do
		cc = cc - 1
		sxr_row._category_specific[ category] = nil
	end

	print( "Passed Categories: " ,cc)
	return cc == 0
end


--================== Supported Scale Types ======================
local scale_types = {
StrainGage = { 	enums = {0}
				,sxr_name="strain_gage"
				,required = function( cal ,l)
								local failed = false
								cal.gage_factor = l.GageFactor
								cal.bridge_factor = l.BridgeFactor
								if cal.gage_factor < 0.1
								or cal.gage_factor > 10 
								or cal.bridge_factor == 0 then
									failed = true
									print( "Failed strain gage check.")
								end
								return failed
							end
			}

,ZeroSpan = { 	enums = {0,1}
				,sxr_name="defined_zero_span" 
				,required = function( cal ,l)
								local failed = false
								cal.electrical_zero = l.P1X
								cal.electrical_span = l.P2X - l.P1X
								cal.physical_span = l.P2Y
								if cal.electrical_span == 0 or l.P2Y == 0 then
									failed = true
									print( "Failed ZeroSpan check." )
								end
								cal.max_electrical = 
									math.abs(cal.electrical_span) +
									math.abs(cal.electrical_zero)
								return failed
							end
			}

,Table2 = { 	enums = {2}
				,sxr_name = "defined_two_point"
				,required = function( cal , l )
								local failed = false
								print( "Scale data: " ,l.ScaleData1 )
								local i,e1,p1,e2,p2 = string.match( l.ScaleData1 
												,'(%d+)~(%d+)~(%d+)~(%d+)~(%d+)' )
								if i ~= '4' then
									print( 'Failed Table2 2Point check, ',i )
									return true
								end
								cal.electrical = {}
								cal.physical = {}
								cal.electrical[1] = tonumber( e1 )
								cal.electrical[2] = tonumber( e2 )
								cal.physical[1] = tonumber( p1 )
								cal.physical[2] = tonumber( p2 )
								print("CHECK: " , e1 , e2 , p1 , p2 )
								if e1 == e2 or p1 == p2 then
									failed = true
									print( "Failed Table2 check." )
								end
								return failed
							end
			}

,TwoPoint = { 	enums = {32} 
				,sxr_name="defined_two_point" 
				,required = function( cal , l )
								local failed = false
								cal.electrical = {}
								cal.electrical[1] = l.P1X
								cal.electrical[2] = l.P2X
								cal.physical = {}
								cal.physical[1] = l.P1Y
								cal.physical[2] = l.P2Y
								if l.P1X == l.P2X or l.P1Y == l.P2Y then
									failed = true
									print( "Failing TwoPoint check.")
								end
								return failed
							end
			}

,Gain =		{	enums = {64}
				,sxr_name = "defined_slope_intercept"
			,required = function( cal , l )
							print( l.ScaleData5)
							cal.intercept ,cal.slope = l.ScaleData5:match(  '([%d%.]+)~([%d%.]+)')
							cal.intercept = tonumber( cal.intercept)
							cal.slope = tonumber( cal.slope)
							return false
						end
				}

,Internal =	{	enums = {0,256,512}
				,sxr_name = "internal"
				,required = function( cal , l ) --internal has no requirements, per spec's given by Dennis
								return false
							end
				}

}



--==================== Supported Sensor Types =======================================

--Supported Sensors
ss = {}

ss.StrainGage = {	
	name = "StrainGage"
	,enum = 1
	,categories = { mx1615 ,mx840}
	,elec_units = "mV/V"
	,s_types = {
					scale_types.ZeroSpan
					,scale_types.StrainGage
					,scale_types.TwoPoint
					,scale_types.Table2
					,scale_types.Gain
				}
	,overlaps = {
			[mx1615] = {
				excitation_range = function( s ,args)
							local l = args.l
							local er
							
							if l.Excitation >= 5 then 
								er = "5.0v"
							elseif l.Excitation >= 2.5 then 
								er = "2.5v"
							elseif l.Excitation >= 1.0 then
								er = "1.0v"
							elseif l.Excitation >= 0.5 then
								er = "0.5v"
							else 
								print("Failed excitation range " .. l.Description )
								return false
							end
							s._category_specific[ mx1615].excitation_range = er
							return true		
						end
				,excitation_mode = function( s ,args)
							local l = args.l
							local em

							if l.ExcitationFrequency == -1 then
								em = 'dc'
							else
								em = 'ac_1200_hz_square'
							end
							s._category_specific[ mx1615].excitation_mode = em
							return true
						end
					}
			,[mx840] = {
				excitation_range = function( s ,args)
							return false
						end
				,excitation_mode = function( s ,args)
							return false
						end
					}
				}
	,req = function ( s ,l )
				local supported = false
				for k,v in ipairs( supported_bridges ) do
					if v == l.SensorType then
						supported = true
					end
				end
				return supported
			end

	,special_setup = function( info ,s ,l )
						s.sensor_input.type = info.type
						if l.Wiring > 0 then
							s.sensor_input.wires = tonumber( l.Wiring )
						else
							s.sensor_input.wires = info.set_wire
						end
						for i,wire in pairs( info.comp_wires ) do
							if s.sensor_input.wires == wire then
								return false
							end
						end
						print( "failed wire check "..l.Description)
						return true
					end

	,setup = function( s , l ) --this is really messy, but the rules are kind of all over the place.
		local t = l.SensorType
		local failed = false
		if ss.StrainGage.req( s , l ) then
			s.sensor_input = 
				{ mode="bridge" ,regulation=true ,type="full" ,wires=6}
			s.bridge_resistance = 350
			s.bridge_lead_wire_resistance = 0

			if t==365 or t==366 or t==367 or t==354 then
				s.sensor_input.type = 'half'
				s.sensor_input.wires = 5
			elseif t==368 or t==369 or t==371 or t==372 then
				s.sensor_input.type = 'quarter'
				if t==368 or t==369 then
					s.sensor_input.wires = 4
				else s.sensor_input.wires = 3 end
			end

			if t==362 or t==365 or t==368 or t==371 then
				s.bridge_resistance = 120
			elseif t==364 or t==367 then
				s.bridge_resistance = 700
			end

			if t==353 or t==354 or t==350 or t==351 or t==352 then
				s.bridge_resistance = l.Resistance
				if s.bridge_resistance == 0 then s.bridge_resistance = 350; end

				local special_bridge_types = 
				{   { num=350,type="full",set_wire=6,comp_wires={6}} 
					,{ num=351,type="half",set_wire=5,comp_wires={5}} 
					,{ num=352,type="quarter",set_wire=4,comp_wires={3,4}}
				}

				for i,info in pairs( special_bridge_types ) do
					if info.num == t then
						failed = ss.StrainGage.special_setup( info ,s ,l)
					end
				end
			end

			if setup_category_specifics( 'StrainGage' ,s ,{l=l}) then
				failed = true
				print( 'Failed category specific check: ' ,l.Description)
			end

			if (s.sensor_input.type == 'half' or s.sensor_input.type == 'full')
			and (s.bridge_resistance < 80 or s.bridge_resistance > 1200) then
				failed = true
				print( "Failed Bridge Resistance " .. l.Description)
			elseif s.sensor_input.type == 'quarter' 
			and ( s.bridge_resistance < 120 or s.bridge_resistance > 350 ) then
				failed = true
				print( "Failed Bridge Resistance " .. l.Description )
			end

			
		else
			failed = true
			print( "Failed general check " .. l.Description)
		end
		return failed
	end
	}

--Bridge and StrainGage are identical, save for their enums
--I'm keeping them seperate just in case something comes up in the future,
--otherwise I'll eventually merge them and adjust the system accordingly
ss.Bridge = copy_table( ss.StrainGage ) 
ss.Bridge.name = "Bridge"
ss.Bridge.enum = 2


ss.Voltage = {
	name="Voltage"
	,enum = 4
	,categories = { mx1601 ,mx1615 }
	,elec_units = "V"
	,s_types = {
					scale_types.ZeroSpan
					,scale_types.TwoPoint
					,scale_types.Table2
					,scale_types.Gain
				}
	,overlaps = {
			[mx1601] = {
				sensor_input = function( s ,args)
							local max = args.max
							local ref = {}
							if max <= 0.1 then
								max = 0.1
							elseif max <= 1 then
								max = nil
							elseif max <= 10 then
								max = 10
							elseif max <= 60 then
								max = 60
							else
								max = nil
							end
							if max then
								ref.mode = "volt"
								ref.max = max
								ref.min = max * -1
								s._category_specific[ mx1601].sensor_input = ref
								return true
							else
								return false
							end
						end
					}
			,[mx1615] = {
				sensor_input = function( s ,args)
							local max = args.max
							local ref = {}
							if max >= 0.6 and max <= 60 then
								max = 15
							else
								max = nil
							end
							if max then
								ref.mode = "volt"
								ref.max = max
								ref.min = max * -1
								s._category_specific[ mx1615].sensor_input = ref
								return true
							else
								return false
							end
						end
					}
				}
	,req = function( s , l )
				local asv = l.ActiveSupplyVoltage
				if l.SensorType == 420 then
					if type(asv) == 'string' then
						return false
					elseif l.ActiveSupplyVoltage == 0 then
						return true
					elseif l.ActiveSupplyVoltage >= 5
					and l.ActiveSupplyVoltage <= 30 then
						s._supported_categories[2] = nil
						--cut out 1615, since invalid in this range
						return true
					end
				end
				return false
			end
	,setup = function( s , l )
				local failed = false

				if ss.Voltage.req( s , l ) then
					s.transducer_power = l.ActiveSupplyVoltage
					s.calibration[3] = nil

					local max = 0
					if l.ScaleType == 1 then --ZeroSpan
						max = s.calibration[1].max_electrical
					elseif l.ScaleType == 2 or l.ScaleType == 32 then --Table2/TwoPoint
						max = s.calibration[1].electrical[2]
					elseif l.ScaleType == 64 then --Gain
						max = l.Range
					end

					if setup_category_specifics( 'Voltage' ,s ,{max=max}) then
						failed = true
						print( 'Failed category specific: ' ,l.Description)
					end

				else
					print( "Failed general check " .. l.Description )
					failed = true
				end
				return failed
			end
	}

ss.Current = {	
		name="Current"
		,enum = 5
		,categories = { mx1601 } 
		,elec_units = "mA"
		,s_types = { 
						scale_types.ZeroSpan
						,scale_types.TwoPoint 
						,scale_types.Table2 
						,scale_types.Gain 
					}
		,req = function( s , l )
						local asv = l.ActiveSupplyVoltage
						return l.SensorType == 421 
							and type(asv) == 'number'
							and ( (l.ActiveSupplyVoltage >= 5 
								and l.ActiveSupplyVoltage <= 30) 
								or l.ActiveSupplyVoltage == 0 )
				end
		,setup = function( s , l )
					local failed = false

					if ss.Current.req( s , l ) then
						s.sensor_input.mode = "4-20mA"
						s.transducer_power = l.ActiveSupplyVoltage
						s.calibration[3] = nil
					else
						failed = true
						print( "Failed general check " .. l.Description )
					end
					return failed
				end
	}

ss.Thermocouple = {
		name="Thermocouple"
		,enum = 8
		,categories = { mx1609} 
		,elec_units = "°C"
		,s_types = {
						scale_types.Internal
					}
		,req = function( s , l )
					return l.SensorType == 451 and
						( l.EngUnit:match( '°C') or l.EngUnit:match('°F') or l.EngUnit:match('°R'))
						and ( l.ScaleType == 0 or l.ScaleType == 256 )
				end
		,setup = function( s , l )
					local failed = false

					if ss.Thermocouple.req( s , l ) then
						print( s.source_physical_units 
							,s.electrical_units , s.output_units )
						s.sensor_input.mode = "thermocouple"
						s.sensor_input.type = 'k'
						s.source_physical_units = '°C'
					else
						failed = true
						print( "Thermo Failed general check " 
							.. l.Description , l.EngUnit , l.SensorType )
					end
					return failed
				end

	}

ss.RTD = {	
		name="RTD"
		,enum = 7
		,categories = { mx1615} 
		,elec_units = "°C"
		,s_types = {
						scale_types.Internal
					}
		,req = function( s , l )
					return l.SensorType == 501 and l.SADPara == 26 and
						( l.EngUnit:match('°C') or l.EngUnit:match( '°F') )
						and ( l.ScaleType == 0 or l.ScaleType == 256 )
				end
		,setup = function( s , l )
					local failed = false

					if ss.RTD.req( s , l ) then
						s.sensor_input.mode = "pt100"
					else
						failed = true
						print( "RTD Failed general check " 
							.. l.Description , l.EngUnit , l.SensorType )
					end
					return failed
				end

	}

ss.Resistance = 	{	
		name="Resistance"
		,enum = 10
		,categories = { mx1615} 
		,elec_units = "Ohm"
		,s_types = {
						scale_types.ZeroSpan
						,scale_types.TwoPoint
						,scale_types.Table2
						,scale_types.Gain
					}
		,req = function( s , l )
					return l.SensorType == 475 and l.SADPara == 26
				end
		,setup = function( s , l )
					local failed = false

					if ss.Resistance.req( s , l ) then
						s.sensor_input.mode = "resistor"
						s.calibration[3] = nil
					else
						failed = true
						print( "Failed general check")
					end
					return failed
				end
	}

ss.Potentiometric = 	{	
		name="Potentiometric"
		,enum = 14
		,categories = { mx1615}
		,elec_units = "mV/V"
		,s_types = {
						scale_types.ZeroSpan
						,scale_types.TwoPoint
						,scale_types.Table2
						,scale_types.Gain
					}
		,req = function( s , l )
					return l.SensorType == 385 and l.SADPara == 28
				end

		,setup = function( s , l )
					local failed = false

					if ss.Potentiometric.req( s , l ) then
						s.sensor_input.mode = 'potentiometer'
						s.calibration[3] = nil
						s.excitation_mode = "dc"
						s.excitation_range = "1.0v"

					else
						failed = true
						print( "Failed general check" ,l.SensorType , l.SADPara)
					end
					return failed
				end

	}

ss.IEPE = {	
		name="IEPE"
		,enum = 31
		,categories = { mx1601 }
		,elec_units = "V"
		,s_types = {
						scale_types.ZeroSpan
						,scale_types.TwoPoint
						,scale_types.Table2
						,scale_types.Gain
					}
		,req = function( s , l )
					return l.SensorType == 581
				end
		,setup = function( s , l )
					local failed = false

					if ss.IEPE.req( s , l ) then
						s.sensor_input.mode = "iepe"
						s.calibration[3] = nil
					else
						failed = true
					end
					return failed
				end
	}



return ss