#!/usr/bin/env lua
--- This is the process which listens on a TCP port,
-- which 'on connect' creates and opens a ringbuffer
-- and forwards all data read from the ringbuffer to
-- the client socket.
-- As soon as the client disconnects or some read error
-- (e.g. overrun) occurs, the ringbuffer is deleted
-- and the client socket is closed.
-- The number of clients handled simultaneously is mainly
-- limited by the system's resources CPU, bandwidth
-- and memory.
--
-- This process works single threaded but uses libev (lua-ev)
-- to do event multiplexing. As a result there are no
-- concurrency issues like races etc, but the code
-- flow indents to the right by every async ev 'event'.
-- See https://github.com/brimworks/lua-ev for more info.

--- the port this process listens on
local port = arg[1] or 7411
local pipe_time = 1 / (tonumber(arg[2] or 25))

local socket = require'socket'
local ev = require'ev'
local cjson = require'cjson'

--- call native c functions from lua jit
local ffi = require'ffi'

--- jet is needed to interact with the 'stream' to
-- create and delete ringbuffers
local jet = require'jet.peer'.new{sync=true}

--- The libc functions used by this process.
ffi.cdef[[
int open(const char* pathname, int flags);
int close(int fd);
int read(int fd, void* buf, size_t count);
int send(int sockfd, const void* buf, size_t len, int flags);
char* strerror(int errnum);
typedef int pid_t;
pid_t getpid(void);
struct sched_param
{
  int __sched_priority;
};
int sched_setscheduler(pid_t pid, int policy, const struct sched_param *param);
int sched_get_priority_max(int policy);
]]


--- frequently used table entries made local
-- to save some cpu.
local evREAD = ev.READ
local evWRITE = ev.WRITE
local loop = ev.Loop.default

local luapid = require'luapid'
luapid.write_pid()

--- raises this process' priority
local raise_priority = function()
  local SCHED_FIFO = 1
  local max_prio = ffi.C.sched_get_priority_max(SCHED_FIFO) - 1
  local pid = ffi.C.getpid()
  local sched_param = ffi.new('struct sched_param',{max_prio})
  assert(ffi.C.sched_setscheduler(pid,SCHED_FIFO,sched_param)==0)
end

--- raise priority to make ringbuffer overruns
-- less likely.
raise_priority()

local syslog = require'syslog'
local log = syslog.new
{
  ident = 'streamng-tcp',
  --   options = {syslog.CONS},
  facility = syslog.DAEMON
}


--- creates an id string made up ip and port.
-- ip:127.0.0.1 port:12345--> tcp_127_0_0_1_12345
local client_id = function(sock)
  local ip,port = sock:getpeername()
  return 'tcp_'..string.gsub(ip,'%.','_')..'_'..port
end

--- will hold all ringbuffer ids to
-- allow cleanup in sig handler
local ringbuffers = {}

--- creates a ringbuffer by calling the jet service 'stream.create'.
local create_ringbuffer = function(id, on_success, on_error)
  local rb_filename = '/dev/rb_'..id
  local file_watcher
  local timer = ev.Timer.new(function()
      file_watcher:stop(loop)
      on_error('timeout while creating ringbuffer')
    end,5)
  timer:start(loop)
  
  -- the ringbuffer module (driver) creates a dev file using
  -- hotplug and mdev with name rule defined below
  -- unforunately mdev invocation and creation of
  -- '/dev/rb_X' may take a while.
  -- ev.Stat uses inotify to watch file creation
  file_watcher = ev.Stat.new(function()
      timer:stop(loop)
      file_watcher:stop(loop)
      local O_NONBLOCK = 2048
      local fd = ffi.C.open(rb_filename,O_NONBLOCK)
      if fd < 0 then
        jet:call('stream/delete',id)
        log:err('opening ringbuffer failed')
        on_error('Could not open '..rb_filename)
      end
      ringbuffers[id] = fd
      on_success(fd)
    end,rb_filename)
  file_watcher:start(loop)
  local ok,err = pcall(jet.call, jet, 'stream/create',{id})
  if not ok then
    timer:stop(loop)
    file_watcher:stop(loop)
    on_error('stream/create failed')
  end
end

--- creates a ringbuffer by calling the jet service 'stream.delete'.
local delete_ringbuffer = function(id)
  if ringbuffers[id] then
    local fd = ringbuffers[id]
    local ok,err = pcall(ffi.C.close,fd)
    if not ok then
      log:err('could not close fd',fd,err)
    end
    ok,err = pcall(
      function()
        jet:call('stream/delete',{id})
      end)
    if not ok then
      log:err('delete rb',id,'failed',err.message)
    end
    ringbuffers[id] = nil
  end
end

local sbind = function(host,port)
  if socket.tcp6 then
    local server = socket.tcp6()
    local _,err = server:setoption('ipv6-v6only',false)
    if err then
      server:close()
      return nil,err
    end
    _,err = server:setoption("reuseaddr",true)
    if err then
      server:close()
      return nil,err
    end
    _,err = server:bind(host,port)
    if err then
      server:close()
      return nil,err
    end
    _,err = server:listen()
    if err then
      server:close()
      return nil,err
    end
    return server
  else
    return socket.bind(host,port)
  end
end

local EAGAIN = 11

--- creates and returns the (ev) io which listens on all interfaces
local listen_io = function()
  local server = assert(sbind("*", port))
  -- enable nonblocking operation
  server:settimeout(0)
  local server_fd = server:getfd()
  assert(server_fd > -1)
  return ev.IO.new(
    -- called on every connect on the listening socket
    function(loop, watcher)
      -- create the client socket object
      local client = assert(server:accept())
      -- make frequently used variable a local
      local client_fd = client:getfd()
      assert(client_fd > -1)
      -- enable nonblocking operation
      client:settimeout(0)
      client:setoption('keepalive',true)
      client:setoption('tcp-nodelay',true)
      -- create unique string id
      local id = client_id(client)
      
      local on_error = function(err)
        client:close()
        print(err)
      end
      
      local closed
      
      --- create and starte an ev IO which releases all
      -- ringbuffer resources if the socket closes.
      ev.IO.new(
        function(loop,io)
          log:info(id, 'other socket end closed')
          io:stop(loop)
          delete_ringbuffer(id)
          closed = true
        end,
        client_fd,
      evREAD):start(loop)
      
      log:info('new client',id,fd,client)
      
      local on_success = function(fd)
        if closed then
          log:info('ringbuffer created but client already closed', id)
          delete_ringbuffer(id)
          return
        end
        log:info('new ringbuffer created',id,fd,client)
        
        -- the maximum size of read chunks
        local chunksize = 200*4096
        local buffer = ffi.new("uint8_t[?]",chunksize)
        
        
        --- create an ev IO which reads ringbuffer data when possible
        -- and forwards the data to the client socket.
        -- When data becomes ready for reading, the process waits 50ms
        -- to allow more data become availble before actually reading,
        -- thus having bigger chunks and less context switches.
        ev.IO.new(
          function(loop,readring_io)
            readring_io:stop(loop)
            -- create an ev Timer which postpones reading by some ms
            -- to collect bigger data chunks
            ev.Timer.new(
              function()
                -- the client may have closed the connection meanwhile
                if not ringbuffers[id] then
                  return
                end
                local bytes = ffi.C.read(fd,buffer,chunksize)
                if bytes < 1 then
                  local errno = ffi.errno()
                  if bytes == -1 then
                    if errno == EAGAIN then
                      return
                    else
                      local errstr = ffi.string(ffi.C.strerror(errno))
                      log:info(id,'read failed with',errstr,errno,fd)
                    end
                  end
                  -- other error than EAGAIN or bytes == 0--> FIN
                  log:info('closing connection (from ::read)')
                  client:shutdown()
                  client:close()
                  delete_ringbuffer(id)
                  return
                end
                local bytessent = 0
                -- create ev IO which forwards the read data, as soon as
                -- client socket becomes readable.
                ev.IO.new(
                  function(loop, sendring_io)
                    sendring_io:stop(loop)
                    -- the client may have closed the connection meanwhile
                    if not ringbuffers[id] then
                      return
                    end
                    local sent = ffi.C.send(client_fd,buffer+bytessent,bytes-bytessent,0)
                    -- since socket should be ready for writing
                    -- any error can be treated unrecoverable
                    if sent == -1 then
                      local errno = ffi.errno()
                      local errstr = ffi.string(ffi.C.strerror(errno))
                      log:info(id,'send failed with',errstr,errno)
                      log:info('closing connection (from ::send)')
                      client:shutdown()
                      client:close()
                      delete_ringbuffer(id)
                      return
                    end
                    bytessent = bytessent + sent
                    if bytessent == bytes then
                      readring_io:start(loop)
                    else
                      sendring_io:start(loop)
                    end
                  end,
                  client_fd,
                evWRITE):start(loop)
              end, pipe_time):start(loop)
          end,
          fd,
        evREAD):start(loop)
      end
      create_ringbuffer(id, on_success, on_error)
    end,
    server_fd,
  evREAD)
end

local quit = function()
  loop:unloop()
  log:info('quit')
  for id,_ in pairs(ringbuffers) do
    log:info('deleting',id,'before exit')
    delete_ringbuffer(id)
  end
  os.exit(0)
end

local SIGHUP = 1
local SIGINT = 2
local SIGKILL = 9
local SIGPIPE = 13
local SIGTERM = 15


-- Allow pipe rate to be adjusted via file and signal
--local sighup = ev.Signal.new(quit,SIGHUP)
local sighup = ev.Signal.new(function(l, s, e)
        local fh = io.open('/tmp/pipe_rate', 'r')
        if not fh then return end
        local l = fh:read()
        fh:close()
        os.remove('/tmp/pipe_rate')
        if not l or not tonumber(l) then return end
        pipe_time = 1 / tonumber(l)
    end, SIGHUP)

local sigint = ev.Signal.new(quit,SIGINT)
local sigkill = ev.Signal.new(quit,SIGKILL)
local sigterm = ev.Signal.new(quit,SIGTERM)
local ignore = function() log:err('SIGPIPE caught and ignored') end
local sigpipe = ev.Signal.new(ignore,SIGPIPE)
local sock_listen_io = listen_io()

jet:state( { path = "net/services/daqStream", value = { port = port } } )

sigpipe:start(loop)
sighup:start(loop)
sigint:start(loop)
sigkill:start(loop)
sigterm:start(loop)
sock_listen_io:start(loop)

loop:loop()
