require 'strict'

-- pulse_frequency.lua - Implements the Pulse Frequency computed channel.

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local rational = require 'rational'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_min = math.min
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Utils_frame_data_type = utils.frame_data_type
local Utils_put_frame = utils.put_frame

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

function mt:flush_output_frame()
    self.of.count = self.bufsize
    if not Utils_put_frame(self.of, self.pipe_list) then
        Frame_delete(self.of)
        self.of = nil
    end
    self.of = Frame_create(self.bufsize, self.out_frame_sample_size)
    self.out_vals = ffi_cast(Utils_frame_data_type(
        self.out_frame_sample_size), self.of.buf)
    self.of.hdr.ts = self.tot_count
    self.tot_count = self.tot_count + self.bufsize
    self.out_index = 0
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    if not Pipe_has_frames(self.in_pipe) then
        return
    end
    local input_rate    = self.sample_rate * self.output_rate_factor
    local output_rate   = self.sample_rate
    local factor        = self.output_rate_factor
    local scale         = self.output_scale_factor
    local minfreq       = self.minimum_output_frequency
    local reset_count   = (input_rate * factor)/minfreq
    local output        = self.output
    local wait          = self.wait
    local state         = self.state
    local refstate      = self.refstate
    local have_first_rising_edge  = self.have_first_rising_edge
    local have_first_falling_edge = self.have_first_falling_edge
    local rising_count  = self.rising_count
    local falling_count = self.falling_count
    local rise_ref_cnt  = self.rise_ref_cnt
    local fall_ref_cnt  = self.fall_ref_cnt

    local f = Pipe_get_frame(self.in_pipe) -- input frame
    local in_vals = ffi_cast(Utils_frame_data_type(f.element_size), f.buf)--data
    local inp_index = 0                    -- Index to input frame data
    while f ~= nil do
        while inp_index < f.count do
            rising_count  = rising_count + 1
            falling_count = falling_count + 1
            state = in_vals[inp_index]
            inp_index = inp_index + 1

            --found an edge
            if (refstate==0 and state==1) or (refstate==1 and state==0) then
                if state==0 and not have_first_rising_edge then
                    have_first_rising_edge = true
                elseif state==1 and not have_first_falling_edge then
                    have_first_falling_edge = true
                elseif state==0 then --falling edge
                    if falling_count > reset_count then
                        output = minfreq
                    else
                        output = (scale*factor*output_rate) / falling_count
                    end
                    fall_ref_cnt = falling_count
                elseif state==1 then --rising edge
                    if rising_count >= reset_count then
                        output = minfreq
                    else
                        output = (scale*factor*output_rate) / rising_count
                    end
                    rise_ref_cnt = rising_count
                end

                if state==1 then
                    rising_count = 0
                else
                    falling_count = 0
                end
            --check if pulse width longer than current output
            elseif (have_first_rising_edge  and rising_count  > rise_ref_cnt) or
                (have_first_falling_edge and falling_count > fall_ref_cnt) then
                output = math.min( ((scale*factor*output_rate) / rising_count),
                    ((scale*factor*output_rate) / falling_count) )
            end

            --Check for reset_count
            if rising_count >= reset_count or falling_count >= reset_count then
                rise_ref_cnt = reset_count
                rising_count = reset_count
                fall_ref_cnt  = reset_count
                falling_count = reset_count
                output = minfreq
            end

            refstate = state
            wait = wait - 1

            if wait==0 then
                wait = factor
                --Pulse frequency cc needs to lag by one in order to
                --match the way CX hardware's pulse frequency works.
                if self.lag_output then
                    --need to double push the first lag_output
                    if not self.did_push_first_sample then
                        self.out_vals[self.out_index] = self.lag_output
                        self.out_index  = self.out_index + 1
                        if self.out_index == self.bufsize then
                            self:flush_output_frame()
                        end
                        self.did_push_first_sample = true
                    end
                    self.out_vals[self.out_index] = self.lag_output
                    self.out_index  = self.out_index + 1
                end
                self.lag_output = output > minfreq and output or minfreq
                if self.out_index == self.bufsize then
                    self:flush_output_frame()
                end
            end
        end

        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
        if f ~= nil then
            in_vals = ffi_cast(Utils_frame_data_type(f.element_size), f.buf)
            inp_index = 0
        end
    end
    self.wait                    = wait
    self.state                   = state
    self.refstate                = refstate
    self.have_first_rising_edge  = have_first_rising_edge
    self.have_first_falling_edge = have_first_falling_edge
    self.rising_count            = rising_count
    self.falling_count           = falling_count
    self.rise_ref_cnt            = rise_ref_cnt
    self.fall_ref_cnt            = fall_ref_cnt
    self.output                  = output
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local pulse_f = setmetatable({}, {__index = mt})
    pulse_f.sie_collect_idx = first_sie_index
    utils.set_common_meta(pulse_f, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Pulse Frequency computed channel " ..
        "only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    assert(input_node.is_logical, "Input channel must be a logical channel")
        --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    pulse_f.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(pulse_f.pipe_rate and pulse_f.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(pulse_f.pipe_rate))
        --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    pulse_f.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(pulse_f.sample_rate and pulse_f.sample_rate > 0,
        "Invalid sample rate: " .. tostring(pulse_f.sample_rate))
    pulse_f.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(pulse_f.out_frame_sample_size and pulse_f.out_frame_sample_size > 0,
        "Invalid frame sample size:  "..tostring(pulse_f.out_frame_sample_size))

    --Module specific parameters
    pulse_f.output_rate_factor = tonumber(stanza.output_rate_factor)
    assert(pulse_f.output_rate_factor and
            pulse_f.output_rate_factor%1 == 0 and
            pulse_f.output_rate_factor >= 10,
            "Invalid output_rate_factor: "..
            tostring(stanza.output_rate_factor)..
            ", must be integer 10 or greater.")
    pulse_f.minimum_output_frequency = tonumber(stanza.minimum_output_frequency)
    assert(pulse_f.minimum_output_frequency and
            pulse_f.minimum_output_frequency >= 0.001 and
            pulse_f.minimum_output_frequency <= 1,
            "Invalid minimum_output_frequency: "..
            tostring(stanza.minimum_output_frequency)..
            ", must be value between 0.001 and 1.0 Hz.")
    pulse_f.output_scale_factor = tonumber(stanza.output_scale_factor)
    assert(pulse_f.output_scale_factor and
            pulse_f.output_scale_factor ~= 0,
            "Invalid output_scale_factor: "..
            tostring(stanza.output_scale_factor)..
            ", must be a non-zero number.")

    -- create input/output pipe structures
    pulse_f.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(pulse_f.in_pipe and pulse_f.in_pipe ~= nil,
        "failed to create input pipe")
    pulse_f.pipe_list = Pipe_new_pipe_list()
    assert(pulse_f.pipe_list and pulse_f.pipe_list ~= nil,
        "failed to create (output) pipe list")

    --local variables needed not set from stanza.
    pulse_f.tot_count = 0
    pulse_f.bufsize = frame.bufsize(pulse_f.sample_rate, pulse_f.pipe_rate)
    -- output frame
    pulse_f.of = Frame_create(pulse_f.bufsize, pulse_f.out_frame_sample_size)
    -- "pointer" to output frame data
    pulse_f.out_vals = ffi_cast(Utils_frame_data_type(
        pulse_f.out_frame_sample_size), pulse_f.of.buf)
    pulse_f.out_index = 0

    pulse_f.wait          = pulse_f.output_rate_factor
    pulse_f.state         = nil
    pulse_f.refstate      = nil
    pulse_f.have_first_rising_edge  = false
    pulse_f.have_first_falling_edge = false
    pulse_f.rising_count  = 0
    pulse_f.falling_count = 0
    pulse_f.rise_ref_cnt  = math.huge
    pulse_f.fall_ref_cnt  = math.huge
    pulse_f.output        = pulse_f.minimum_output_frequency

    return pulse_f, 1
end

return m