require 'strict'

-- run_stopper.lua - Implements the Run Stopper computed channel.

require('lj_include'){
    'signal',
}

local ffi = require 'ffi'
local C = ffi.C
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local rational = require 'rational'
local ev = require 'ev'
local armadb_client = require 'armadb_client'
local log = require 'log'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof
local Frame_create = frame.create
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Utils_frame_data_type = utils.frame_data_type
local Utils_put_frame = utils.put_frame

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    if not Pipe_has_frames(self.in_pipe) then
        return
    end
    
    local f  = Pipe_get_frame(self.in_pipe)   -- input frame
    -- create frame for output and get input, output pointers
    local of, in_vals, out_vals = utils.create_output_frame(f.maxsize, f,
        self.tot_count, self.out_frame_sample_size)
    while f ~= nil do
        for i = 0, f.count-1 do
            out_vals[i] = in_vals[i]
            if in_vals[i] == 1 and not self.triggered then
                self.triggered = true
                C.kill(C.getpid(), C.SIGTERM)
            end
        end
        of.count = self.bufsize
        if not Utils_put_frame(of, self.pipe_list) then
            Frame_delete(of)
            of = nil
        end
        of = Frame_create(self.bufsize, self.out_frame_sample_size)
        out_vals = ffi_cast(Utils_frame_data_type(
            self.out_frame_sample_size), of.buf)
        of.hdr.ts = self.tot_count
        self.tot_count = self.tot_count + self.bufsize
        --delete used frames
        Frame_delete(f)
        f = nil
        if Pipe_has_frames(self.in_pipe) then
            f = Pipe_get_frame(self.in_pipe)
            if f ~= nil then
                in_vals = ffi_cast(Utils_frame_data_type(f.element_size), f.buf)
            end
        end
    end
end


local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local test_stop = setmetatable({}, {__index = mt})
    test_stop.sie_collect_idx = first_sie_index
    utils.set_common_meta(test_stop, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Pulse Frequency computed channel " ..
        "only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    assert(input_node.is_logical, "Input channel must be a logical channel")
        --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    test_stop.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(test_stop.pipe_rate and test_stop.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(test_stop.pipe_rate))
        --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    test_stop.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(test_stop.sample_rate and test_stop.sample_rate > 0,
        "Invalid sample rate: " .. tostring(test_stop.sample_rate))
    test_stop.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(test_stop.out_frame_sample_size and
        test_stop.out_frame_sample_size > 0, "Invalid frame sample size:  "..
        tostring(test_stop.out_frame_sample_size))

    -- create input/output pipe structures
    test_stop.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(test_stop.in_pipe and test_stop.in_pipe ~= nil,
        "failed to create input pipe")
    test_stop.pipe_list = Pipe_new_pipe_list()
    assert(test_stop.pipe_list and test_stop.pipe_list ~= nil,
        "failed to create (output) pipe list")
        
    --local variables needed not set from stanza.
    test_stop.tot_count = 0
    test_stop.bufsize = frame.bufsize(test_stop.sample_rate, test_stop.pipe_rate)
    test_stop.triggered = nil

    return test_stop, 1
end

return m
