require 'strict'

-- timed_trigger.lua - Implements the Timed Trigger computed channel.
--
-- This module uses the single input channel's data to determine
-- when to trigger, instead of using a separate trigger channel.

local ffi = require 'ffi'
local rational = require 'rational'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'

local ffi_fill = ffi.fill
local ffi_sizeof = ffi.sizeof
local math_min = math.min
local math_floor = math.floor
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}

function mt:dump()
    return string.format('nbr to wait=%q, left to write=%q, ' ..
        'start mode=%q, delay period=%q, sustain period=%q\n',
        tostring(self.nbr_to_wait), tostring(self.left_to_write),
        tostring(self.start_mode), tostring(self.delay_period),
        tostring(self.sustain_period))
end

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Checks for a trigger condition resulting from the given input
-- sample.  When the "trigger start mode" is "On FALSE-TRUE edge"
-- or "On TRUE-FALSE edge" the immediately previous input sample
-- is also used to determine whether a trigger condition exists.
--
-- NOTE that when a trigger condition is detected this function
-- initializes certain values which are used to control how many
-- "trigger set" values are put in the output frame(s).  This means
-- that this function MUST NOT be called when a triggered condition
-- already exists!
--
-- Parameter:  'in_sample' is a data element from an input frame
-- Returns:  true if a trigger condition has occurred, false
--     otherwise.
function mt:trigger_event(in_sample)
    local triggered = false
    if self.start_mode == "when_true" then
        if in_sample ~= 0 then
            triggered = true
            self.nbr_to_wait = self.delay_period_samples
            self.left_to_write = self.sustain_period_samples
        end
    elseif self.start_mode == "on_false_to_true_edge" then
        if in_sample ~= 0 and self.previous_in_sample == 0 then
            triggered = true
            self.nbr_to_wait = self.delay_period_samples
            self.left_to_write = self.sustain_period_samples
        end
    else  -- self.start_mode == "after_true_to_false_edge"
        if in_sample == 0 and self.previous_in_sample ~= 0 then
            triggered = true
            self.nbr_to_wait = self.delay_period_samples
            self.left_to_write = self.sustain_period_samples
        end
    end
    self.previous_in_sample = in_sample
    return triggered
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    if not Pipe_has_frames(self.in_pipe) then
        return
    end

    -- Process each frame in input pipe.  There is only one input pipe.
    local count    -- length of data buf in current input frame
    local in_vals  -- "pointer" to current input frame's buf
    local of       -- current output frame
    local out_vals -- "pointer" to current output frame's buf
    local f = Pipe_get_frame(self.in_pipe)  -- current input frame
    while f ~= nil do
        count = f.count
        -- create frame for output - NOTE that the output frame
        --     is filled here with zeros as an attempt to allow
        --     greater runtime efficiency
        of, in_vals, out_vals =
            utils.create_output_frame(f.maxsize, f, self.tot_count,
                self.out_frame_sample_size)
        ffi_fill(out_vals, f.maxsize * self.out_frame_sample_size)
        -- set output values
        -- (these will be only 0 (not triggered) or 1 (triggered)
        --   NOTE - possible performance improvement:  use wmemset()
        --          to set 1's in output frame.  May not actually
        --          improve speed, though.
        --   NOTE - another possible performance improvement:
        --          use ffi.copy() to fill "triggered" values in
        --          an output frame from a prefilled buffer.
        --for buf_index = 0, count - 1 do
        local buf_index = 0
        while buf_index < count do
            if self.triggered or
                self:trigger_event(in_vals[buf_index]) then
                self.triggered = true
                if self.nbr_to_wait <= count - buf_index then
                    -- this input frame contains all remaining
                    -- samples in "delay" period
                    buf_index = buf_index + self.nbr_to_wait
                    self.nbr_to_wait = 0
                    -- put as many "triggered" values in output frame
                    -- as required or as possible.
                    for j = buf_index,
                        math_min(count,
                            buf_index + self.left_to_write) - 1 do
                        out_vals[j] = 1
                    end
                    local nbr_written = 
                        math_min(count - buf_index, self.left_to_write)
                    self.left_to_write = self.left_to_write - nbr_written
                    buf_index = buf_index + nbr_written
                    if self.left_to_write <= 0 then
                        self.triggered = false
                    end
                else
                    -- more input frames will be needed to encompass
                    -- the "delay" period - leave the remainder of
                    -- the output frame buf filled with zeros
                    self.nbr_to_wait = self.nbr_to_wait -
                        (count - buf_index)
                    break
                end
            else  -- not yet triggered, leave initialized zero at
                  --      this index in output frame
                buf_index = buf_index + 1
            end
        end
        of.count = count
        self.tot_count = self.tot_count + of.count
        -- put output frame in output pipe(s)
        if not utils.put_frame(of, self.pipe_list) then
            Frame_delete(of)
        end
        -- delete input frame, get next one
        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local timed_trigger = setmetatable({}, {__index = mt})
    timed_trigger.sie_collect_idx = first_sie_index
    utils.set_common_meta(timed_trigger, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Timed Trigger computed channel only" ..
        " accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    --    output frame element size
    assert(stanza.data_type, "Missing data type")
    timed_trigger.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(timed_trigger.out_frame_sample_size and
           timed_trigger.out_frame_sample_size > 0,
        string.format("Invalid frame sample size:  %s",
        tostring(timed_trigger.out_frame_sample_size)))
    --    trigger "start mode"
    timed_trigger.start_mode = stanza.trigger_start_mode
    assert(timed_trigger.start_mode, "Missing timed trigger start mode")
    assert(timed_trigger.start_mode == "when_true" or
           timed_trigger.start_mode == "on_false_to_true_edge" or
           timed_trigger.start_mode == "after_true_to_false_edge",
        "INVALID timed trigger start mode: " ..
        tostring(stanza.trigger_start_mode))
    timed_trigger.previous_in_sample = 0
    if timed_trigger.start_mode == "on_false_to_true_edge" then
        -- prevent spurious trigger at first input
        timed_trigger.previous_in_sample = 1
    end
    --    trigger delay period
    assert(stanza.trigger_delay_period,
        "Missing timed trigger delay period")
    timed_trigger.delay_period = tonumber(stanza.trigger_delay_period)
    assert(timed_trigger.delay_period and
        timed_trigger.delay_period >= 0,
        "INVALID timed trigger delay period: " ..
        tostring(stanza.trigger_delay_period))
    --    trigger sustain period
    assert(stanza.trigger_sustain_period,
        "Missing timed trigger sustain period")
    timed_trigger.sustain_period =
        tonumber(stanza.trigger_sustain_period)
    assert(timed_trigger.sustain_period and
        timed_trigger.sustain_period >= 0,
        "INVALID timed trigger sustain period: " ..
        tostring(stanza.trigger_sustain_period))
    --    at least one of delay or sustain period must be > 0
    assert(timed_trigger.delay_period > 0 or
        timed_trigger.sustain_period > 0,
        "At least one of timed trigger delay or sustain period " ..
        "must be greater than zero")
    --    sample rate, needed below
    local sample_rate = rational:new(stanza.sample_rate):tonumbers()
    assert(sample_rate and sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(sample_rate)))
    --    number of samples in a trigger delay period
    timed_trigger.delay_period_samples =
       math_floor(sample_rate * timed_trigger.delay_period + 0.5)
    if timed_trigger.start_mode == "after_true_to_false_edge" then
        -- account for the "after" part of this mode
        timed_trigger.delay_period_samples =
            timed_trigger.delay_period_samples + 1
    end
    --    number of "triggered" samples to output when triggered
    timed_trigger.sustain_period_samples =
        math_floor(sample_rate * timed_trigger.sustain_period + 0.5)
    if timed_trigger.sustain_period_samples <= 0 then
        timed_trigger.sustain_period_samples = 1
    end
    
    -- create input/output pipe structures
    timed_trigger.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(timed_trigger.in_pipe and timed_trigger.in_pipe ~= nil,
        "failed to create input pipe")
    timed_trigger.pipe_list = Pipe_new_pipe_list()
    assert(timed_trigger.pipe_list and timed_trigger.pipe_list ~= nil,
        "failed to create (output) pipe list")

    -- working values which need to be kept across frame boundaries
    timed_trigger.nbr_to_wait = 0
    timed_trigger.left_to_write = 0
    timed_trigger.triggered = false

    -- used for frames' time stamps (just a sample count for
    -- this module)
    timed_trigger.tot_count = 0

    return timed_trigger, 1
end

return m
