require 'strict'

-- track.lua - Implements the Min Track, Max Track, and
-- Range Track computed channel.  This computed channel requires
-- a single input channel, and accepts an optional "simple"
-- trigger channel.
--
-- When instantiated this computed channel will output one (and
-- only one) of the following for each input channel sample:
--     Max ==> the maximum value yet reached by the input channel
--     Min ==> the minimum value yet reached by the input channel
--     Range ==> the maximum range of values yet reached by the
--         input channel, which is the difference between Max
--         and Min.
--
-- The output value of this channel may be reset when triggered
-- by a trigger channel, if that option is selected.  Resetting
-- this channel results in:
--     For Max or Min, the output is set to the current input
--         sample value.
--     For Range, the output is set to zero.

local dbg = require 'dbg'
local ffi = require 'ffi'
local rational = require 'rational'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local simple_trigger = require 'test_engine/simple_trigger'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast

local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_add_frame = pipe.add_frame
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_has_frames = pipe.has_frames
local Pipe_get_frame = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list


local FLT_MAX_OF_RANGE = 0x7FFFFFFF
local FLT_MIN_OF_RANGE = -0x7FFFFFFF
local DBL_MAX_OF_RANGE = 0x7FFFFFFFFFFFFFFF
local DBL_MIN_OF_RANGE = -0x7FFFFFFFFFFFFFFF


local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    if not Pipe_has_frames(self.in_pipe) or
        (self.trig_obj and
        not Pipe_has_frames(self.trig_pipe)) then
        -- no work to do, so return
        return
    end

    -- Process each frame in input pipe.  There is only one input pipe.
    local in_count     -- length of data buf in current input frame
    local sample_type  -- C type string for frame's buf
    local in_vals      -- "pointer" to current input frame's buf
    local inbuf_index  -- sample index for current input frame
    local of           -- will reference created output frame
    local out_vals     -- will reference "pointer" to output frame's data
    local f = Pipe_get_frame(self.in_pipe)  -- next input frame
    if self.trig_obj then
        self.trig_obj:get_frame()
    end

    while f ~= nil and f.count > 0 do
        in_count = f.count
        sample_type = utils.frame_data_type(f.element_size)
        in_vals = ffi_cast(sample_type, f.buf)
        inbuf_index = 0
        of = Frame_create(self.out_frame_size, f.element_size)
        assert(of and of ~= nil, "Could not create output frame")
        of.hdr.ts = self.tot_count
        out_vals = ffi_cast(sample_type, of.buf)

        -- NOTE:  trig_obj:triggered() (if trig_obj exists) must be
        -- called for each trigger channel input value, to ensure the
        -- trig_obj state is accurate.
        local triggered
        -- Process input frame according to this module's track mode
        if self.track_mode == 'max' then
            -- Process each sample in input frame.
            for i = 0, in_count - 1 do
                triggered = self.trig_obj and self.trig_obj:triggered(i)
                if in_vals[i] > self.max_val or triggered then
                    self.max_val = in_vals[i]
                end
                out_vals[i] = self.max_val
            end
        elseif self.track_mode == 'min' then
            -- Process each sample in input frame.
            for i = 0, in_count - 1 do
                triggered = self.trig_obj and self.trig_obj:triggered(i)
                if in_vals[i] < self.min_val or triggered then
                    self.min_val = in_vals[i]
                end
                out_vals[i] = self.min_val
            end
        else -- is 'range'
            -- Process each sample in input frame.
            for i = 0, in_count - 1 do
                triggered = self.trig_obj and self.trig_obj:triggered(i)
                if triggered then
                    self.max_val = in_vals[i]
                    self.min_val = in_vals[i]
                    self.range_val = 0
                else
                    if in_vals[i] > self.max_val then
                        self.max_val = in_vals[i]
                        self.range_val = self.max_val - self.min_val
                    end
                    if in_vals[i] < self.min_val then
                        self.min_val = in_vals[i]
                        self.range_val = self.max_val - self.min_val
                    end
                end
                out_vals[i] = self.range_val
            end
        end

        -- put output frame in output pipe(s)
        of.count = in_count
        self.tot_count = self.tot_count + of.count
        if not utils.put_frame(of, self.pipe_list) then
            Frame_delete(of)
        end
        -- delete input and trigger frames
        Frame_delete(f)
        if self.trig_obj then
            self.trig_obj:delete_frame()
        end
        -- get next frames, if available
        if Pipe_has_frames(self.in_pipe) and
            (not self.trig_obj or
             Pipe_has_frames(self.trig_pipe)) then
            f = Pipe_get_frame(self.in_pipe)
            if self.trig_obj then
                self.trig_obj:get_frame()
            end
        else
            -- no work to do
            f = nil
        end
    end
end

-- this is the table which will contain the new() method.
local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local cc_obj = setmetatable({}, {__index = mt})
    cc_obj.sie_collect_idx = first_sie_index
    utils.set_common_meta(cc_obj, stanza)
    -- Validate needed metadata.
    --   input pipe (one only)
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Min/Max/Range computed channel " ..
        " only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    cc_obj.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(cc_obj.pipe_rate and cc_obj.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(cc_obj.pipe_rate))
    --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    cc_obj.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(cc_obj.sample_rate and cc_obj.sample_rate > 0,
        "Invalid sample rate: " .. tostring(cc_obj.sample_rate))
    --   calculated output frame size
    cc_obj.out_frame_size =
        frame.bufsize(cc_obj.sample_rate, cc_obj.pipe_rate)
    --   data type
    local data_type = stanza.data_type
    assert(data_type, "Missing data_type")
    --   track mode
    cc_obj.track_mode = stanza.track_mode
    assert(cc_obj.track_mode == 'min' or
        cc_obj.track_mode == 'max' or
        cc_obj.track_mode == 'range',
        string.format("Invalid tracking mode: %s",
            tostring(stanza.track_mode)))
    -- Check for optional metadata.
    local trig_reset =
        utils.xlate_meta_boolean(stanza.triggered_reset)
    if trig_reset then
        local trig_mode = stanza.trigger_mode
        assert(trig_mode, "Trigger Mode not found")
        assert(trig_mode == 'when_true' or
            trig_mode == 'on_false_to_true_edge' or
            trig_mode == 'after_true_to_false_edge',
            string.format("Invalid trigger mode: %s",
                tostring(stanza.trigger_mode)))
        local trig_channel = stanza.trigger_channel
        assert(trig_channel, "Trigger Channel not found")
        local trig_node = node_list[trig_channel]
        assert(trig_node, "Node for Trigger Channel not found")
        cc_obj.trig_pipe = Pipe_attach_pipe(trig_node.pipe_list)
        assert(cc_obj.trig_pipe and cc_obj.trig_pipe ~= nil,
            "failed to create trigger pipe")
        -- use "simple trigger" utility code
        cc_obj.trig_obj =
            simple_trigger:new(cc_obj.trig_pipe, trig_mode)
        assert(cc_obj.trig_obj, "failed to create trigger object")
    end

    -- create input/output pipe structures
    cc_obj.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(cc_obj.in_pipe and cc_obj.in_pipe ~= nil,
        "failed to create input pipe")
    cc_obj.pipe_list = Pipe_new_pipe_list()
    assert(cc_obj.pipe_list and cc_obj.pipe_list ~= nil,
        "failed to create (output) pipe list")

    -- The following are used to keep track of this channel object's
    -- current output value.  Note that the initial values for 'max'
    -- and 'min' are chosen to ensure that 'max' and 'min' are assigned
    -- the first input sample value.
    local sample_size, c_type = utils.sample_size(data_type)
    assert(sample_size == 8 or sample_size == 4,
        "Invalid data type size")
    cc_obj.out_frame_sample_size = sample_size
    if sample_size == 8 then
        cc_obj.max_val = DBL_MIN_OF_RANGE
        cc_obj.min_val = DBL_MAX_OF_RANGE
    elseif sample_size == 4 then
        cc_obj.max_val = FLT_MIN_OF_RANGE
        cc_obj.min_val = FLT_MAX_OF_RANGE
    end
    cc_obj.range_val = 0

    -- used for frames' time stamps (just a sample count for
    -- this module)
    cc_obj.tot_count = 0

    return cc_obj, 1
end

return m

