require 'strict'

local sample_types = require 'sample_types'

-- sie_schema.lua - Provides data schema information for creating
-- a SIE channel. 

-- Local definitions.

-- Table used to determine data schema "name" from
-- channel's category.  This is used in determining/creating
-- which SIE decoder to use.
local category_to_schema = {
    ['^sink/sie/time_history'] = '_sequential_',
    ['^sink/sie/burst_history'] = '_burst_',
    ['^sink/sie/peak_valley'] = '_pv_sequential_',
    ['^sink/sie/time_at_level_1d'] = '_histogram_bin_bounds_',
    ['^sink/sie/message_raw'] = '_message_raw_'
}

local m = {}

-- Module definitions.

-- Returns an identifying string for the "collect" SIE schema
-- appropriate to the channel with the given stanza.
-- Parameters:
--     'stanza' a channel's metadata table
-- Returns:  string comprising schema's name
function m.get_collect_schema_name(stanza)
    local name
    if stanza.category:find('^source/hbm_cx23r/msg_can') or
           stanza.data_type == "message_can" then
        name = '_message_can_'
    elseif stanza.data_type == "message_raw" then
        name = '_message_raw_'
    elseif stanza.sample_rate then
        name = '_sequential_'
    else
        name = '_message_numeric_'
    end
    return name
end

-- Returns an identifying string for the datamode schema
-- appropriate to the given category.
-- Parameters:
--     'category' is the category string in the module's metadata
-- Returns:  string comprising data schema's name
function m.get_dm_schema_name(category)
    local name = '_sequential_'
    for k, v in pairs(category_to_schema) do
        if category:find(k) then
            name = v
            break
        end
    end
    return name
end

-- Initializes tables containing the information needed to
-- use or create the XML for a SIE decoder.
-- Parameters:
--     'md_in' - is metadata table containing the info needed
-- Returns:  decoder info table
function m.get_schema_info(schema_name, md_in)
    assert(md_in, "Missing parameter 'md_in'")
    local schema_tab = {}

    schema_tab.name = schema_name
    schema_tab.sample_rate = md_in.sample_rate
    schema_tab.bits = md_in.nbr_data_bits
    assert(schema_tab.bits, "Missing data bit size")

    if schema_tab.name == '_burst_' then
        -- NOTE:  For now, use only 'float' for Burst History samples
        schema_tab.somat_data_format = 'float'
        m.get_burst_history_schema(schema_tab, md_in)
    elseif schema_tab.name == '_histogram_bin_bounds_' then
        -- NOTE:  For now, use only 'float' for Peak Valley samples
        schema_tab.somat_data_format = 'float'
        -- NOTE that the following returned schema table does NOT
        --   include information for the generic histogram
        --   bin min/max values lookup table, which will need to
        --   exist for use by any histograms.
        --   There will be one such generic table for each histogram
        --   with a different "bound_bits" value - in other words
        --   only one for now since we only support 64 bits now.
        --   See sie_utils.ensure_sie_bounds_channel() for the code
        --   used to write such a generic table along with its
        --   decoder.
        m.get_histogram_schema(schema_tab, md_in)
    elseif schema_tab.name == '_pv_sequential_' then
        -- NOTE:  For now, use only 'float' for Peak Valley samples
        schema_tab.somat_data_format = 'float'
        md_in.dim0_label = 'Events'
        md_in.dim0_units = 'Count'
        m.get_sequential_schema(schema_tab, md_in)
    elseif schema_tab.name == '_message_can_' then
        schema_tab.somat_data_format = 'message_can'
        m.get_message_can_schema(schema_tab, md_in)
    elseif schema_tab.name == '_message_raw_' then
        schema_tab.somat_data_format = 'raw'
        schema_tab.bits = '' -- no bits defined (octets used)
        m.get_message_raw_schema(schema_tab, md_in)
    elseif schema_tab.name == '_message_numeric_' then
        if md_in.data_type:find('float') then
            schema_tab.somat_data_format = 'float'
        elseif md_in.data_type:find('uint') then
            schema_tab.somat_data_format = 'uint'
        else
            schema_tab.somat_data_format = 'int'
        end
        m.get_message_numeric_schema(schema_tab, md_in)
    else  -- schema_tab.name is assumed to be '_sequential_' !!!
        schema_tab.somat_data_format = 'float'
        md_in.dim0_label = 'Time'
        md_in.dim0_units = 'Seconds'
        m.get_sequential_schema(schema_tab, md_in)
    end

    schema_tab.decoder_type = schema_tab.name ..
        schema_tab.somat_data_format

    return schema_tab
end

-- Initializes the given table to contain the SIE schema information
-- needed for a Burst History channel.
--   Example of the SIE XML associated with the burst schema:
--     
--   <decoder id="3">
--    <read var="v0" bits="64" type="int" endian="big"/>
--    <read var="v2" bits="64" type="int" endian="big"/>
--    <loop var="v0">
--     <read var="v1" bits="64" type="float" endian="big"/>
--     <sample/>
--     <set var="v2" value="{$v2 + 1}"/>
--    </loop>
--   </decoder>
--   <ch id="5" name="base channel" private="base">
--    <tag id="somat:data_format">float</tag>
--    <tag id="somat:version">1.0</tag>
--    <tag id="core:description"></tag>
--    <tag id="somat:datamode_type">burst_history</tag>
--    <tag id="somat:data_bits">64</tag>
--    <tag id="core:schema">somat:burst</tag>
--   </ch>
--   <ch test="0" id="6" base="5" group="9" name="clamped_input_level_1_dio_cx23r_ff0017@burst_2">
--    <tag id="core:description"></tag>
--    <tag id="core:sample_rate">100</tag>
--    <tag id="somat:input_channel">clamped_input_level_1_dio_cx23r_ff0017</tag>
--    <tag id="somat:connector">clamped_input_level_1@dio.cx23r-ff0017</tag>
--    <tag id="somat:datamode_name">burst_2</tag>
--    <dim index="0">
--     <tag id="core:units">Seconds</tag>
--     <tag id="core:label">Time</tag>
--     <data decoder="3" v="0"/>
--     <xform scale="0.01" offset="0"/>
--    </dim>
--    <dim index="1">
--     <tag id="core:label">burst_2</tag>
--     <data decoder="3" v="1"/>
--    </dim>
--    <dim index="2">
--     <tag id="core:label">Offset</tag>
--     <tag id="core:units">Sample</tag>
--     <data decoder="3" v="2"/>
--    </dim>
--   </ch>
--
-- Parameters:
--     'schema' - table in which to put the schema info
--     'md_in' - meta data for the a particular datamode channel
-- Returns:  filled in schema table
function m.get_burst_history_schema(schema, md_in)
    schema.sie_core_schema = 'somat:burst'
    schema.somat_datamode_type = 'burst_history'
    -- SIE dim info
    schema.dims = {
        {
            _dim_index = 0,
            _decoder_v = 0,
            ['core:label'] = md_in.time_label or 'Time',
            ['core:units'] = md_in.time_units or 'Seconds',
            _use_scale = true
        }, {
            _dim_index = 1,
            _decoder_v = 1,
            ['core:label'] = md_in.name or '',        --given by channel
            ['core:units'] = md_in.output_units or '' --given by channel
        }, {
            _dim_index = 2,
            _decoder_v = 2,
            ['core:label'] = 'Offset',
            ['core:units'] = 'Sample'
        }
    }
    -- info for decoder read elements which are outside a loop
    schema.nonrep_reads = {
        {
            varname = "v0",
            bits = "64",
            sie_data_type = "int"
        }, {
            varname = "v2",
            bits = "64",
            sie_data_type = "int"
        }
    }
    -- info for decoder set elements which are outside a loop
    schema.nonrep_sets = {}
    -- info for decoder loop elements
    -- Note:  the keys indicate the kind of loop - bare, short,
    --        or long.
    schema.loops = {
        short1 = {
            head = { lvarname = "v0" },
            reads = {
                {
                    varname = "v1",
                    bits = schema.bits,
                    sie_data_type = schema.somat_data_format
                }
            },
            sample = true,
            sets = {
                {
                    varname = "v2",
                    value = "{$v2 + 1}"
                }
            }
        }
    }

    return schema
end

-- Initializes the given table to contain the SIE schema information
-- needed for a Time At Level (histogram) channel.
--   Example of the SIE XML associated with the one dimensional
--   histogram schema:
--     
--   <decoder id="3">
--    <read var="v1" bits="64" type="float" endian="big"/>
--    <loop var="v0">
--     <read var="v2" bits="64" type="float" endian="big"/>
--     <sample/>
--     <set var="v1" value="{$v2}"/>
--    </loop>
--   </decoder>
--   <ch id="2" private="bin">
--    <tag id="core:description">Generic histogram bin min/max values lookup table</tag>
--    <dim index="0">
--     <data decoder="3" v="0"/>
--    </dim>
--    <dim index="1">
--     <data decoder="3" v="1"/>
--    </dim>
--    <dim index="2">
--     <data decoder="3" v="2"/>
--    </dim>
--   </ch>
--   <ch id="3" base="2" group="6" name="tal1#bin_1" private="bin"/>
--   <decoder id="4">
--    <read var="dimcount" bits="32" type="uint" endian="big"/>
--    <read var="bincount1" bits="32" type="uint" endian="big"/>
--    <loop var="v1" start="0" end="{$bincount1}">
--     <read var="v0" bits="64" type="uint" endian="big"/>
--     <sample/>
--    </loop>
--   </decoder>
--   <ch id="4" name="base channel" private="base">
--    <tag id="somat:data_format">uint</tag>
--    <tag id="somat:version">1.0</tag>
--    <tag id="core:description"></tag>
--    <tag id="somat:datamode_type">time_at_level</tag>
--    <tag id="somat:data_bits">64</tag>
--    <tag id="core:schema">somat:histogram</tag>
--   </ch>
--   <ch test="0" id="5" base="4" group="7" name="clamped_input_level_1_dio_cx23r_ff0017@tal1">
--    <tag id="core:description">Clamped Input Voltage</tag>
--    <tag id="core:sample_rate">100</tag>
--    <tag id="somat:input_channel">clamped_input_level_1_dio_cx23r_ff0017</tag>
--    <tag id="somat:connector"></tag>
--    <tag id="somat:datamode_name">tal1</tag>
--    <dim index="0">
--     <tag id="core:units"></tag>
--     <tag id="core:description"></tag>
--     <tag id="core:label">Counts</tag>
--     <data decoder="4" v="0"/>
--    </dim>
--    <dim index="1">
--     <tag id="somat:range_min">-6000</tag>
--     <tag id="somat:range_max">6000</tag>
--     <tag id="core:description">Axis 0 Bin Min</tag>
--     <tag id="core:label">clamped_input_level_1_dio_cx23r_ff0017 (mV)</tag>
--     <data decoder="4" v="1"/>
--     <xform index_ch="3" index_dim="1"/>
--    </dim>
--    <dim index="2">
--     <tag id="somat:range_min">-6000</tag>
--     <tag id="somat:range_max">6000</tag>
--     <tag id="core:description">Axis 0 Bin Max</tag>
--     <tag id="core:label">clamped_input_level_1_dio_cx23r_ff0017 (mV)</tag>
--     <data decoder="4" v="1"/>
--     <xform index_ch="3" index_dim="2"/>
--    </dim>
--   </ch>
--     
-- Parameters:
--     'schema' - table in which to put the schema info
--     'md_in' - meta data for the a particular datamode channel
-- Returns:  filled in schema table
function m.get_histogram_schema(schema, md_in)
    schema.sie_core_schema = 'somat:histogram'
    schema.somat_datamode_type = 'time_at_level'
    -- exception for histogram samples
    schema.somat_data_format = 'uint'
    -- SIE dim info
    schema.dims = {
        {
            _dim_index = 0,
            _decoder_v = 0,
            ['core:description'] = '',
            ['core:label'] = 'Counts',
            ['core:units'] = 'Counts'
        }, {
            _dim_index = 1,
            _decoder_v = 1,
            ['core:description'] = 'Axis 0 Bin Min',
            ['core:label'] = md_in.name .. ' (' .. md_in.units .. ')',
            ['core:units'] = md_in.units or '',
            ['somat:range_min'] = md_in.low_bound,
            ['somat:range_max'] = md_in.high_bound,
            _use_scale = true,
            _use_index = true,
            _index_ch = false,  -- must be assigned later
            _index_dim = "1"
        }, {
            _dim_index = 2,
            _decoder_v = 1,
            ['core:description'] = 'Axis 0 Bin Max',
            ['core:label'] = md_in.name .. ' (' .. md_in.units .. ')',
            ['core:units'] = md_in.units or '',
            ['somat:range_min'] = md_in.low_bound,
            ['somat:range_max'] = md_in.high_bound,
            _use_scale = true,
            _use_index = true,
            _index_ch = false,  -- must be assigned later
            _index_dim = "2"
        }
    }
    -- info for decoder read elements which are outside a loop
    schema.nonrep_reads = {
        {
            varname = "dimcount",
            bits = "32",
            sie_data_type = "uint"
        }, {
            varname = "bincount1",
            bits = "32",
            sie_data_type = "uint"
        }
    }
    -- info for decoder set elements which are outside a loop
    schema.nonrep_sets = {}
    -- info for decoder loop elements
    -- Note:  the keys indicate the kind of loop - bare, short,
    --        or long.
    schema.loops = {
        long1 = {
            head = {
                lvarname = "v1",
                start = "0",
                endvar = "{$bincount1}"
            },
            reads = {
                { varname = "v0",
                  bits = "64",
                  sie_data_type = "uint" }
            },
            sample = true
        }
    }

    return schema
end

-- Initializes the given table to contain the SIE schema information
-- needed for a sequential channel (such as Time History).
--   Example of the SIE XML associated with the sequential schema:
--     
--   <decoder id="2">
--    <read var="v0" bits="64" type="int" endian="big"/>
--    <loop var="v0">
--     <read var="v1" bits="64" type="float" endian="big"/>
--     <sample/>
--    </loop>
--   </decoder>
--   <ch id="0" name="base channel" private="base">
--    <tag id="somat:data_format">float</tag>
--    <tag id="somat:version">1.0</tag>
--    <tag id="core:description"></tag>
--    <tag id="somat:datamode_type">time_history</tag>
--    <tag id="somat:data_bits">64</tag>
--    <tag id="core:schema">somat:sequential</tag>
--   </ch>
--   <ch test="0" id="1" base="0" group="5" name="clamped_input_level_1_dio_cx23r_ff0017">
--    <tag id="core:description">Clamped Input Voltage</tag>
--    <tag id="core:sample_rate">100</tag>
--    <tag id="somat:input_channel">clamped_input_level_1_dio_cx23r_ff0017</tag>
--    <tag id="somat:connector">clamped_input_level_1@dio.cx23r-ff0017</tag>
--    <tag id="somat:datamode_name">dm</tag>
--    <dim index="0">
--     <tag id="core:description"></tag>
--     <data decoder="2" v="0"/>
--    </dim>
--    <dim index="1">
--     <tag id="core:units">mV</tag>
--     <tag id="core:description"></tag>
--     <tag id="core:label">clamped_input_level_1_dio_cx23r_ff0017</tag>
--     <data decoder="2" v="1"/>
--    </dim>
--   </ch>
--     
-- Parameters:
--     'schema' - table in which to put the schema info
--     'md_in' - meta data for the a particular datamode channel
-- Returns:  filled in schema table
function m.get_sequential_schema(schema, md_in)
        schema.sie_core_schema = 'somat:sequential'
        schema.somat_datamode_type = 'time_history'
        -- SIE dim info
        schema.dims = {
            {
                _dim_index = 0,
                _decoder_v = 0,
                ['core:description'] = '',
                ['core:label'] = md_in.dim0_label,
                ['core:units'] = md_in.dim0_units,
                _use_scale = true
            }, {
                _dim_index = 1,
                _decoder_v = 1,
                ['core:description'] = '',
                ['core:label'] = md_in.name,
                ['core:units'] = md_in.output_units
            }
        }
        -- info for decoder read elements which are outside a loop
        schema.nonrep_reads = {
            { varname = "v0",
              bits = "64",
              sie_data_type = "int" }
        }
        -- info for decoder set elements which are outside a loop
        schema.nonrep_sets = {}
        -- info for decoder loop elements
        -- Note:  the keys indicate the kind of loop - bare, short,
        --        or long.
        schema.loops = {
            short1 = {
                head = { lvarname = "v0" },
                reads = {
                    { varname = "v1",
                      bits = schema.bits,
                      sie_data_type = schema.somat_data_format }
                },
                sample = true,
                sets = {}
            }
        }

    return schema
end

-- Initializes the given table to contain the SIE schema information
-- needed for a raw CAN message channel.
--   Example of the SIE XML associated with the "raw CAN" schema:
--     
--   <decoder id="2">
--    <loop>
--     <read var="v0" bits="64" type="int" endian="big"/>
--     <seek from="current" offset="12"/>
--     <read var="ff" bits="8" type="uint"/>
--     <seek from="current" offset="-13"/>
--     <read var="v1" octets="{4 + ($ff &amp; 15)}" type="raw"/>
--     <seek from="current" offset="{9 - ($ff &amp; 15)}"/>
--     <sample/>
--    </loop>
--   </decoder>
--   <ch test="0" id="0" name="raw_can1_cx23r_ff0017" group="5">
--    <tag id="somat:source_sort_order">1</tag>
--    <tag id="core:description">Raw CAN messages</tag>
--    <tag id="somat:input_channel">raw_can1_cx23r_ff0017</tag>
--    <tag id="somat:datamode_name">dm</tag>
--    <tag id="data_type">message_can</tag>
--    <tag id="somat:connector">@can1.cx23r-ff0017</tag>
--    <tag id="somat:datamode_type">message_log</tag>
--    <tag id="core:schema">somat:message</tag>
--    <dim index="0">
--     <tag id="core:label">Time</tag>
--     <tag id="core:units">Seconds</tag>
--     <data decoder="2" v="0"/>
--     <xform scale="2.5e-08" offset="0"/>
--    </dim>
--    <dim index="1">
--     <tag id="core:label">raw_can1_cx23r_ff0017</tag>
--     <data decoder="2" v="1"/>
--    </dim>
--   </ch>
--     
-- Parameters:
--     'schema' - table in which to put the schema info
--     'md_in' - meta data for the a particular datamode channel
-- Returns:  filled in schema table
function m.get_message_can_schema(schema, md_in)
        schema.sie_core_schema = 'somat:message'
        schema.somat_datamode_type = 'message_log'
        -- SIE dim info
        schema.dims = {
            {
                _dim_index = 0,
                _decoder_v = 0,
                ['core:description'] = '',
                ['core:label'] = "Time",
                ['core:units'] = "Seconds",
                _use_scale = true
            }, {
                _dim_index = 1,
                _decoder_v = 1,
                ['core:description'] = '',
                ['core:label'] = md_in.name or '',
                ['core:units'] = md_in.output_units or ''
            }
        }
        -- info for decoder read elements which are outside a loop
        schema.nonrep_reads = {}
        -- info for decoder set elements which are outside a loop
        schema.nonrep_sets = {}
        -- info for decoder loop elements
        -- Note:  the keys indicate the kind of loop - bare, short,
        --        or long.
        schema.loops = {
            bare1 = {
                head = {},
                reads = {
                    { varname = "v0",
                      bits = 64,
                      sie_data_type = "int",
                      post_seek = { from = "current", offset = "12"} },
                    { varname = "ff",
                      bits = 8,
                      sie_data_type = "uint",
                      post_seek = { from = "current", offset = "-13"} },
                    { varname = "v1",
                      octets = "{4 + ($ff &amp; 15)}",
                      sie_data_type = "raw",
                      post_seek = { from = "current",
                                    offset = "{9 - ($ff &amp; 15)}"} }
                },
                sample = true,
                sets = {}
            }
        }

    return schema
end


-- Initializes the given table to contain the SIE schema information
-- needed for a raw message channel.
--   Example of the SIE XML associated with the "message raw" schema:
--     
-- +++here
--     
-- Parameters:
--     'schema' - table in which to put the schema info
--     'md_in' - meta data for the a particular datamode channel
-- Returns:  filled in schema table
function m.get_message_raw_schema(schema, md_in)
        schema.sie_core_schema = 'somat:message'
        schema.somat_datamode_type = 'message_log'
        -- SIE dim info
        schema.dims = {
            {
                _dim_index = 0,
                _decoder_v = 0,
                ['core:description'] = '',
                ['core:label'] = md_in.dim0_label or '',
                ['core:units'] = md_in.dim0_units or '',
                _use_scale = true
            }, {
                _dim_index = 1,
                _decoder_v = 1,
                ['core:description'] = '',
                ['core:label'] = md_in.name or '',
                ['core:units'] = 'message_data'
            }
        }
        -- info for decoder read elements which are outside a loop
        schema.nonrep_reads = {}
        -- info for decoder set elements which are outside a loop
        schema.nonrep_sets = {}
        -- info for decoder loop elements
        -- Note:  the keys indicate the kind of loop - bare, short,
        --        or long.
        schema.loops = {
            bare1 = {
                head = {},
                reads = {
                    { varname = "size",
                      bits = 32,
                      sie_data_type = "uint" },
                    { varname = "v0",
                      bits = 64,
                      sie_data_type = "int" },
                    { varname = "v1",
                      octets = "{$size - 8}",
                      sie_data_type = "raw" }
                },
                sample = true,
                sets = {}
            }
        }

    return schema
end

-- Initializes the given table to contain the SIE schema information
-- needed for a non-raw message channel.
--   Example of the SIE XML associated with the message-numeric schema:
--     
-- +++here
--     
-- Parameters:
--     'schema' - table in which to put the schema info
--     'md_in' - meta data for the a particular datamode channel
-- Returns:  filled in schema table
function m.get_message_numeric_schema(schema, md_in)
        schema.sie_core_schema = 'somat:message'
        schema.somat_datamode_type = 'message_log'
        -- SIE dim info
        schema.dims = {
            {
                _dim_index = 0,
                _decoder_v = 0,
                ['core:description'] = '',
                ['core:label'] = md_in.dim0_label or '',
                ['core:units'] = md_in.dim0_units or '',
                _use_scale = true
            }, {
                _dim_index = 1,
                _decoder_v = 1,
                ['core:description'] = '',
                ['core:label'] = md_in.name or '',
                ['core:units'] = md_in.output_units or ''
            }
        }
        -- info for decoder read elements which are outside a loop
        schema.nonrep_reads = {}
        -- info for decoder set elements which are outside a loop
        schema.nonrep_sets = {}
        -- info for decoder loop elements
        -- Note:  the keys indicate the kind of loop - bare, short,
        --        or long.
        schema.loops = {
            bare1 = {
                head = {},
                reads = {
                    { varname = "v0",
                      bits = 64,
                      sie_data_type = "int" },
                    { varname = "v1",
                      bits = schema.bits,
                      sie_data_type = schema.somat_data_format }
                },
                sample = true,
                sets = {}
            }
        }

    return schema
end

return m
