require 'strict'

require('lj_include'){
    'sys/types',
    'sys/stat',
    'fcntl',
    'sys/socket',
    'sys/uio',
    'unistd',
}

local __ = require '__'
local bit = require 'bit'
local cjson = require 'cjson'
local log = require 'log'
local membuf = require 'membuf'
local ffi = require 'ffi'
local rational = require 'rational'
local zlib = require 'zlib'
local tutil = require 'test_engine/test_eng_utils'
local stempls = require 'test_engine/sie_templates'
local sie_schema = require 'test_engine/sie_schema'
local sutil = require 'test_engine/sie_utils'
local Frame = require 'test_engine/frame'
local Pipe = require 'test_engine/pipe'
local notify = require 'sys_notify'

local Pipe_get_frame = Pipe.get_frame
local Pipe_flush_frames = Pipe.flush_frames
local Frame_delete = Frame.delete

require 'middleclass'

local C = ffi.C

-- Fake vector write:  does not actually write to a file
-- but does return the values a successful vector write
-- to a file would return.
-- Params:  same as for C.writev -
--   fd - file descriptor (not actually used)
--   iov - array of iovec structs
--   iovcnt - length of iov
-- Returns:  number of bytes which would be written
--   by C.writev, followed by 'true' to indicate
--   this function was called instead of C.writev
local function fake_writev(fd, iov, iovcnt)
    local wrote = 0
    for i = 0, iovcnt - 1 do
        wrote = wrote + iov[i].iov_len
    end
    return wrote, true
end

-- Use C.writev for vector writes.  fake_writev is only used
-- after the test has logged an error.
local C_writev = C.writev


local ffi_cast = ffi.cast
local ffi_new = ffi.new
local ffi_sizeof = ffi.sizeof
local ffi_copy = ffi.copy
local math_floor = math.floor

local function hexify(s)
    return string.gsub(s, "(.)", function(c)
        return string.format(" %02X", string.byte(c))
    end)
end

local htonl = ffi.abi("le") and bit.bswap or function(x) return x end
local ntohl = htonl

-- C declaration for a sie buf object's data.
ffi.cdef[[
    typedef int64_t timestamp;

    typedef struct __attribute((packed)) {
        uint32_t size;
        uint32_t group;
        uint32_t magic;
    } sie_block_header;

    typedef struct __attribute((packed)) {
        uint32_t crc;
        uint32_t size;
    } sie_block_trailer;

    typedef struct __attribute((packed)) {
        uint64_t offset;
        uint32_t group;
    } sie_index_entry;

    typedef struct sie_buffer {
        int                     idx;
        int                     fd;
        int                     group;
        struct pipe             *pipe;
        char                    *file_name;
        int                     element_size;
        int                     dm_type;  // datamode type (for Burst History)
        int                     payload_overhead;
        int                     timestamped_samples;
        bh_offset_t             bh_offset[1]; /* offset, for Burst History */
        timestamp               timestamp[2]; /* 0 = actual, 1 = scratch */
        sie_index_entry         index_entry[1]; /* Only for index buffer */
        sie_index_entry         index_b[1];
        struct sie_buffer       *index_buf;
        struct iovec            vec[1024];
        sie_block_header        header;
        struct membuf           payload;
        sie_block_trailer       trailer;
        double                  total_written;
        timestamp               stop_stamp;
    } sie_buffer;
]]

local sie_overhead = 20
local sie_magic = 0x51EDA7A0

-- Used to define an sie buf object.
local mt = {}

-- Initializes an sie buf object.
-- Params:
--   fd - file descriptor to (eventually) write buf to
--   group - identifies which group this buf's data
--       will be associated with in SIE file
--   max_payload - number of bytes allowed in payload;
--       note that the buffer size is rounded up to
--       a power of 2
--   element_size - size of data elements to be put
--       in the sie buf
--   payload_overhead - additional bytes needed for each
--       data element stored in sie buf
--   datamode_type - a code indicating what special processing
--       is needed for this sie buf (currently only BH_DATAMODE
--       or nil)
function mt:initialize(fd, group, max_payload, element_size,
                       payload_overhead, datamode_type)
    self.fd = fd
    self.group = group
    self.element_size = element_size
    self.dm_type = datamode_type and tonumber(datamode_type) or 0
    self.payload_overhead = payload_overhead
    self.header.group = htonl(group)
    self.header.magic = htonl(sie_magic)
    self.payload:initialize(max_payload)
    self.vec[0].iov_base = self.header
    self.vec[0].iov_len = ffi_sizeof(self.header)
    self.total_written = 0
end

-- Returns length of sie buf payload plus length of overhead
-- (the latter is equal to the sum of the lengths of the buffer
-- header and buffer trailer).
function mt:block_len()
    return #self.payload + sie_overhead
end

-- Appends the contents of given buffer to sie buf payload.
-- Params:
--   buf - buffer containing bytes to append
--   size - number of bytes to append
-- Returns:  false when successful (!), true otherwise
function mt:enq(buf, size)
    if size > self.payload:available() then
        -- If there's only a small amount buffered already (i.e.
        -- payload headers), and the buffer is too small to fit
        -- what's being enqueued, then expand.
        if #self.payload < 256 then
            self.payload:expand(#self.payload + size)
        else
            return true -- Enqueue FAILED
        end
    end

    self.payload:append(buf, size)
    -- Non-intuitive boolean sense here so that the caller
    -- can write "if sbuf:enq(...) then handle_enq_failure() end"
    return false -- OK, no problem with the enqueue
end

-- Appends the contents of given buffer to sie buf payload.
-- If the first attempt to append fails, the sie buf is
-- written to its associated file and the append is retried.
-- Params:
--   buf - buffer containing bytes to append
--   size - number of bytes to append;
--       if nil or false, the actual length of 'buf'
--       is used
-- Returns:  the length of the payload if successful,
--       false otherwise
function mt:append(buf, size)
    if not size then
        size = #buf
    end
    if self:enq(buf, size) then
        self:flush(true)
        if self:enq(buf, size) then
            return false
        end
    end
    return #self.payload
end

-- Sets the CRC for a SIE file block to be written.  The CRC
-- is calculated over the block header and the data in the
-- frames to be written.  Also sets the block trailer values.
-- Params:
--   frame - the head of the list of frames to process
--   block_len - the size of the sie file block to be written
--   num_samples - the number of data elements (samples) to
--       be written as part of the sie file block
function mt:finalize_frames(frame, block_len, num_samples)
    local block_size_be = htonl(block_len)
    self.header.size = block_size_be
    local crc = zlib.crc32(0, nil, 0)  -- initialize CRC
    local f = frame
    local scount = 0

    -- CRC the header, timestamp and the data
    crc = zlib.crc32(crc, ffi_cast('char *', self.header),
        ffi_sizeof(self.header))

    if self.timestamped_samples == 0 then
        crc = zlib.crc32(crc, ffi_cast('char *', self.timestamp),
            ffi_sizeof('timestamp'))
    end
    if self.dm_type == sutil.BH_DATAMODE then
        crc = zlib.crc32(crc, ffi_cast('char *', self.bh_offset),
            ffi_sizeof('bh_offset_t'))
    end

    num_samples = num_samples or 2^32
    while f ~= nil and num_samples > 0 do
        local size = math.min(f.count, num_samples) * f.element_size
        scount = scount + size
        crc = zlib.crc32(crc, f.buf, size)
        num_samples = num_samples - (size / f.element_size)
        f = f.prev
    end
    self.trailer.crc = htonl(crc)
    self.trailer.size = block_size_be
end

-- NOTE:  called only by mt:flush() and mt:write_index().
-- Sets the CRC of this sie buf object.  The CRC
-- is calculated over the block header and the payload to
-- to be written.  Also sets the block trailer values.
function mt:finalize()
    local block_size = self:block_len()
    local block_size_be = htonl(block_size)
    self.header.size = block_size_be
    local crc = zlib.crc32(0, nil, 0)  -- initialize CRC
    crc = zlib.crc32(crc, ffi_cast('char *', self.header),
        ffi_sizeof(self.header))
    crc = zlib.crc32(crc, self.payload:head(), #self.payload)
    self.trailer.crc = htonl(crc)
    self.trailer.size = block_size_be
end

-- Write all of sie buf's frames to the associated file as
-- a SIE block.
-- Returns:  the number of data samples written
function mt:flush_frames()
    local pipe = self.pipe
    local frame = pipe.last_frame

    -- If there aren't any frames, we are done
    if frame == nil then return 0 end

    -- Figure out the length to be written
    local block_len = sie_overhead

    if self.timestamped_samples == 0 then
        block_len = block_len + ffi_sizeof('timestamp')
    end

    if self.dm_type == sutil.BH_DATAMODE then
        self.timestamp[0] = ffi_cast('timestamp', frame.hdr.ts)
        block_len = block_len + ffi_sizeof('bh_offset_t')
        -- Update trigger offset value from first frame written
        self.bh_offset[0] = frame.offset
    end

    -- Calculate number of data samples to be written, and finish
    -- calculating the number of bytes to be written (subject to
    -- adjustment by "ragged end" cleanup at test stop).
    local sample_count = 0
    repeat
        block_len = block_len + (frame.element_size * frame.count)
        sample_count = sample_count + frame.count
        frame = frame.prev
    until frame == nil

    -- Check to see if we should stop writing anymore data for this channel
    -- because we have a limit on the number of samples for the channel.
    -- This is used to enforce test duration and ragged end clean up for
    -- normal test stops.
    local start_stamp = tonumber(self.timestamp[0])
    local stop_stamp = tonumber(self.stop_stamp)
    if stop_stamp > 0 and start_stamp + sample_count >= stop_stamp then
        local extra_samples = (start_stamp + sample_count) - stop_stamp
        -- Mark buffer as reaching it's stop point
        self.stop_stamp = 0
        if extra_samples < 0 then
            Pipe_flush_frames(pipe)
            return 0
        end
        -- Adjust the number of samples based on the stop point
        local frame = pipe.last_frame
        block_len = block_len - (frame.element_size * extra_samples)
        sample_count = sample_count - extra_samples
    elseif stop_stamp == 0 then
        Pipe_flush_frames(pipe)
        return 0
    end

    -- If after adjusting for the stop time there are no samples left to write
    -- just return
    if sample_count <= 0 then
        Pipe_flush_frames(pipe)
        return 0
    end

    -- Write the index
    self:write_index(self.header.group, block_len)

    -- Do CRC etc.
    self:finalize_frames(pipe.last_frame, block_len, sample_count)

    -- Write all of the data to the file
    local amt_written = self:write_block_frames(block_len, sample_count)

    -- If not written return sample count of 0
    if amt_written <= 0 then return 0 end

    -- Free up the frames
    Pipe_flush_frames(pipe)

    -- Return amount written
    return sample_count
end

-- Note:  Only used to write SIE XML and during "close SIE file"
--        processing.
-- Writes the payload to the SIE file, with an SIE block header
-- and trailer.  The SIE index for the block is written first.
-- After all the writes the payload buffer is cleared.
-- Params:  'force' - flush even if payload contains only overhead
-- Returns:  the number of bytes written to file, not including
--     the SIE index.
function mt:flush(force)
    if #self.payload <= self.payload_overhead and not force then
        return 0
    end
    local len = self:block_len()
    if len <= 0 then
        return 0
    end
    self:write_index(self.header.group, len)
    self:finalize()
    -- KLUDGE won't work non-blocking, needs error handling
    self:write_block()
    self.payload:clear()
    return len
end

-- Updates the sie buf's index and then appends the index to the
-- sie buf's payload buffer.  If the payload buffer is too full
-- to accept the index, the payload if first written to the SIE
-- file.
-- Params:
--     'group' - the SIE group number the index is for
--     'len' - the length of the SIE data to be indexed
function mt:write_index(group, len)
    if self.index_buf == nil then
        return -- Don't index the index blocks
    end
    self = self.index_buf
    assert(self.index_buf == nil)
    assert(ffi_sizeof(self.index_entry) == 12)
    self.index_entry[0].group = group
    if self.payload:available() < ffi_sizeof(self.index_entry) then
        -- flush() but without infinite loop danger
        self:finalize()
        self:write_block()
        self.index_entry[0].offset =
            self.index_entry[0].offset + self:block_len()
        self.payload:clear()
    end
    assert(self:enq(self.index_entry, ffi_sizeof(self.index_entry)) == false)
    self.index_entry[0].offset = self.index_entry[0].offset + len
end

-- NOTE:  Called only by flush_frames().
-- Writes the sie buf's frames to the sie buf's file.
-- Params:
--   'block_len' - number of bytes in the block to be written
--   'max_samples' - number of data samples in block to be written
-- Returns:  when positive, the number of bytes written to file,
--   otherwise indicates an error
--   FIXME: a return of zero indicates an error, but this can
--          be due to zero bytes to write.  Right now the only
--          call to this function assumes the return value is
--          the number of bytes written.  This will be a problem
--          if an error causes the return value to be negative.
function mt:write_block_frames(block_len, max_samples)
    -- Vector 0 = header
    -- Vector 1 = timestamp (sequential only)
    -- Vector 2 = first sample's offset from trigger, for Burst History
    -- Vector 3 - x = data
    -- Vector x + 1 = trailer
    local vec = self.vec

    local tot2write = vec[0].iov_len
    local frame = self.pipe.last_frame
    if self.timestamped_samples == 0 then
        vec[1].iov_base = self.timestamp
        vec[1].iov_len = ffi_sizeof('timestamp')
        tot2write = tot2write + vec[1].iov_len
    end
    local i = 2
    if self.dm_type == sutil.BH_DATAMODE then
        vec[i].iov_base = self.bh_offset
        vec[i].iov_len = ffi_sizeof('bh_offset_t')
        tot2write = tot2write + vec[i].iov_len
        i = i + 1
    end

    max_samples = max_samples or 2^32
    while frame ~= nil and max_samples > 0 do
        vec[i].iov_base = frame.buf
        vec[i].iov_len = frame.element_size * math.min(frame.count, max_samples)
        max_samples = max_samples - (vec[i].iov_len / frame.element_size)
        tot2write = tot2write + vec[i].iov_len
        frame = frame.prev
        i = i + 1
    end
    vec[i].iov_base = self.trailer
    vec[i].iov_len = ffi_sizeof(self.trailer)
    tot2write = tot2write + vec[i].iov_len
    local wrote, fake = C_writev(self.fd, self.vec, i + 1)

    if wrote ~= tot2write then
        local fname = ffi.string(self.file_name)
        if wrote == -1 and ffi.errno() ~= C.ENOSPC then
            log:error(__'SIE {0} write() error: {1}', fname, log:strerror())
        else
            if wrote > 0 then
                -- Short write, so undo it
                local pos = C.lseek64(self.fd, 0, C.SEEK_END)
                if pos == -1 then
                    log:error(__'SIE {0} lseek64() error: {1}',
                        fname, log:strerror())
                end
                if C.ftruncate64(self.fd, pos - wrote) ~= 0 then
                    log:error(__'SIE {0} ftruncate64() error: {1}',
                        fname, log:strerror())
                end
            end
            -- KLUDGE This is a temporary measure pending full crisis
            -- buffer support, test continuation on filesystem full, etc.
            local sn = notify:new(__'File System Full')
            local c = {
                type = 'warning',
                severity = 'none',
                msg = [[The system has run out of space for further writes to the data (SIE) file for the test.  The test will continue running and if there are any histogram data modes in the test, the data for those will be written to the SIE file when the test is stopped.  No other data will be written to the SIE file from this point on.]],
                led = 'red',
                led_mode = 'one_hz',
                disk_full = true,
            }
            log:warn(c.msg)
            sn:_send_notification(c)
            C_writev = fake_writev
            wrote = 0
        end
    end

    if not fake then self.total_written = self.total_written + wrote end

    return wrote
end

-- NOTE:  Called only by mt:flush() and mt:write_index().
-- Writes the sie buf's header, payload, and trailer to the
-- SIE file.  No buffers are cleared or altered.  Note that the
-- contents of any frames belonging to this sie buf are assumed
-- to be already copied to the payload.
-- Returns:  when positive, the number of bytes written to file,
--   otherwise indicates an error
--   FIXME: a return of zero indicates an error, but this can
--          be due to zero bytes to write.  Not a problem right
--          now since the return value is not ever used.
function mt:write_block()
    self.vec[1].iov_base = self.payload:head()
    self.vec[1].iov_len = #self.payload
    self.vec[2].iov_base = self.trailer
    self.vec[2].iov_len = ffi_sizeof(self.trailer)
    local wrote, fake = C_writev(self.fd, self.vec, 3)
    local len = self:block_len()
    if wrote ~= len then
        local fname = ffi.string(self.file_name)
        if wrote == -1 and ffi.errno() ~= C.ENOSPC then
            log:error(__'SIE {0} write() error: {1}', fname, log:strerror())
        else
            if wrote > 0 then
                -- Short write, so undo it
                local pos = C.lseek64(self.fd, 0, C.SEEK_END)
                if pos == -1 then
                    log:error(__'SIE {0} lseek64() error: {1}',
                        fname, log:strerror())
                end
                if C.ftruncate64(self.fd, pos - wrote) ~= 0 then
                    log:error(__'SIE {0} ftruncate64() error: {1}',
                        fname, log:strerror())
                end
            end
            -- KLUDGE This is a temporary measure pending full crisis
            -- buffer support, test continuation on filesystem full, etc.
            local sn = notify:new(__'File System Full')
            local c = {
                type = 'warning',
                severity = 'none',
                msg = [[The system has run out of space for further writes to the data (SIE) file for the test.  The test will continue running and if there are any histogram data modes in the test, the data for those will be written to the SIE file when the test is stopped.  No other data will be written to the SIE file from this point on.]],
                led = 'red',
                led_mode = 'one_hz',
            }
            log:warn(c.msg)
            sn:_send_notification(c)
            C_writev = fake_writev
            wrote = 0
        end
    end
    if not fake then self.total_written = self.total_written + wrote end
    return wrote
end

-- Constructor for sie buf with C-based data structure.
local sie_buffer_ctor = ffi.metatype('sie_buffer', {__index = mt})

-- Create this file's module as a class "inheriting" template.lua.
local m = class('sie_writer'):include(require 'template')


-- Creates a file to reserve sufficient filesystem space to allow
-- finishing writing to SIE file when test terminates.  This ensures
-- the file system running out of space will not prevent the
-- completion of the SIE file.
-- Params:
--   'bytes':  number of bytes to reserve with the file (note that
--       a fudge factor number of bytes is added to this)
--   'oos_callback':  function to call if "out of space" occurs
function m:reserve_space(bytes, oos_callback)
    local flags
    if not self.oos_info then
        self.oos_info = {}

        -- FIXME:  Need to reserve enough space here for index, etc.  but how
        -- to know how much to reserve?

        self.space_reserve_file =
            string.gsub(self.file_name, '^(.+/)(.+)%.sie$', '%1%.%2%.reserve')
        flags = bit.bor(C.O_CREAT, C.O_WRONLY, C.O_TRUNC, C.O_APPEND,
            C.O_LARGEFILE)

        -- FIXME:  Kludge need to "calculate" space required for indicies?
        bytes = bytes + (1024*1024)
    else
        flags = bit.bor(C.O_WRONLY, C.O_APPEND, C.O_LARGEFILE)
    end

    local mode = ffi.new('mode_t', 0x1B6) -- rw-rw-rw  (0x124 = r--r--r--)
    local fd = C.open(self.space_reserve_file, flags, mode)

    -- Extend the size of the file by the number of bytes requested
    local pos = tonumber(C.lseek64(fd, 0, C.SEEK_END))
-- FIXME!!
-- The line below *should* but it's not and I don't want to spend time on it
-- at the moment
--    C.ftruncate64(fd, pos + bytes)
    C.write(fd, ffi.new('char[?]', bytes), bytes)
    C.close(fd)
    self.oos_info[#self.oos_info + 1] = {
        bytes = bytes,
        callback = oos_callback,
    }
end

-- Initialize the new class object.
-- Params:  args - a table containing the parameters.  Currently
--   the keys should be 'file_name' for SIE file name and 'max_bufs'
--   for the number of sie bufs to be allocated.
--       Unused at this point is the key 'fd' for the file descriptor
--   of the previously opened SIE file.
function m:initialize(args)
    self.max_bufs = args.max_bufs + 5 -- XML + index + group_tag + slop
    self.bufs = ffi_new('sie_buffer[?]', self.max_bufs)
    self.last_buf = -1
    self.next_ids = {
        group = 0,
        test = 0,
        ch = 0,
        decoder = 2,
    }
    self:template_init()
    if args.fd then
        self.fd = args.fd
        self.file_name = args.file_name or string.format("fd(%d)", self.fd)
    elseif args.file_name then
        local flags = bit.bor(C.O_CREAT, C.O_WRONLY, C.O_TRUNC, C.O_APPEND,
            C.O_LARGEFILE)
        -- ffi.new because open(), open64() declared as varargs functions
        local mode = ffi.new('mode_t', 0x1B6) -- rw-rw-rw  (0x124 = r--r--r--)
        self.fd = C.open(args.file_name, flags, mode)
        assert(self.fd >= 0)
        self.file_name = args.file_name
    else
        error('file_name or fd unspecified')
    end
    self.c_file_name = ffi_new('char[?]', #self.file_name + 1, self.file_name)
    self.xml_group = self:next_id('group')
    self.index_group = self:next_id('group')

    self.xml_buf_idx = 0
    -- NOTE that all sie bufs share the same index sie buf (except the
    -- index sie buf itself).  The following sets the idx for the
    -- index sie buf.
    self.index_buf_idx = 1
    self.scratch_buf_idx = 2

    -- The index new_buffer() is first so that index_buf gets set properly
    -- in all succeeding new_buffer() calls
    self:new_buffer(self.index_buf_idx, self.index_group,
        8192, ffi_sizeof'sie_index_entry')
    -- NOTE that the following is necessary because new_buffer()
    -- automatically sets the index_buf value, but we don't want
    -- the sie index buf itself to have an index_buf value so
    -- it won't be indexed itself.
    self.bufs[self.index_buf_idx].index_buf = nil   -- KLUDGE?

    -- KLUDGE make dynamic maybe
    self:new_buffer(self.xml_buf_idx, self.xml_group, 256 * 1024)
    self:new_buffer(self.scratch_buf_idx, 0 --[[placeholder]], 1024, 1)

    self.index_entry = ffi_new('sie_index_entry')
    
    -- Add the 
    assert(args.setup and args.setup~= nil, "SIE writer missing setup info")
    local setup = args.setup
    local sort_order = '<tag id="somat:source_sxr_channel_order">'
    for i = 1, #setup do
        if setup[i].sort_order and
            (setup[i].category:match("^source/") or
             setup[i].category:match("^filter/") ) then
                sort_order = sort_order .. setup[i].sort_order ..'\t'..
                    setup[i].name..'\n'
        end
    end
    sort_order = sort_order .. "</tag>"
    
    self:xml(stempls.sie_header .. sort_order)

    -- KLUDGE prime the C cache with C.close()
    -- LuaJIT interpreter needs to run a C.close() before the compiler can
    -- compile a C.close(). Have seen trace aborts on C.close not in cache.
    do
        local fd = C.open('/dev/null', C.O_RDONLY, 0)
        C.close(fd)
    end
end

-- Returns the number of bytes written to the SIE file.
function m:amount_written()
    local total = 0
    for i = 0, self.last_buf do
        total = total + self.bufs[i].total_written
    end
    return math.floor(total)
end

-- Writes all frames to the SIE file.
function m:run()
    self:flush_these(3, self.last_buf - 2)
end

-- Flushes specified frames to SIE file.  This function is used by
-- m:run(), and by Burst History to flush frames at the end of bursts.
-- Params:
--     'first' - index of first sie buf to flush
--     'nbr_bufs' - how many sie bufs to flush (indices are sequential
--         from 'first')
function m:flush_these(first, nbr_bufs)
    for i = first, first + nbr_bufs - 1 do
        local buf = self.bufs[i]
        local pipe = buf.pipe
        if pipe ~= nil and pipe.last_frame ~= nil then
            buf.timestamp[1] = buf.timestamp[0]
            buf.timestamp[0] = tonumber(buf.timestamp[1]) + buf:flush_frames()
        end
    end
end

-- Returns the next id of the given type.
-- Params:  't' - the type of identifier to be returned
-- Returns:  the next id
function m:next_id(t)
    local id = self.next_ids[t]
    self.next_ids[t] = id + 1
    return id
end

-- Returns the index for a newly initialized sie buf.  Since
-- all the sie buf objects were preallocated, this function
-- finds the sie buf with the specified index or the next
-- sie buf in the list and then initializes it.
-- Params:
--   idx - index of the desired sie buf, or false (or nil)
--       in which case the index of the next sie buf to be
--       allocated is used
--   group - group number (in the context of the SIE file)
--   bytes - payload size in bytes
--   sample_size - size of data elements to be put
--       in the sie buf
--   payload_overhead - additional bytes needed for each
--       data element stored in sie buf
--   pipe - frame pipe for this sie buf
--   timestamped_samples - true if samples are timestamped
--   datamode_type - a code indicating what special processing
--       is needed for this sie buf (currently only BH_DATAMODE
--       or nil)
-- Returns:  the idx of the "new" sie buf
function m:new_buffer(idx, group, bytes, sample_size, payload_overhead, pipe,
    timestamped_samples, datamode_type)
    group = group or self:next_id('group')
    if not idx then
        self.last_buf = self.last_buf + 1
        idx = self.last_buf
    end
    -- ensures an already initialized sie buf is not used
    -- NOTE:  Also means the sie bufs must be initialized in order.
    self.last_buf = math.max(idx, self.last_buf)
    assert(self.last_buf < self.max_bufs)
    local buf = self.bufs[idx]
    buf:initialize(self.fd, group, bytes, sample_size or 1,
        payload_overhead or 0, datamode_type)
    buf.idx = idx
    buf.file_name = self.c_file_name
    buf.index_buf = self.index_buf_idx and self.bufs[self.index_buf_idx] or nil
    buf.pipe = pipe
    buf.timestamped_samples = timestamped_samples and 1 or 0
    buf.stop_stamp = -1
    return idx
end

-- If a parameter is provided, just flushes the accumulated SIE XML
-- text to the SIE file.  Otherwise processes the given parameter(s)
-- through template.lua's template() function and appends the result
-- to the sie buf dedicated to SIE file XML.
function m:xml(template, ...)
    if not template then
        self.bufs[self.xml_buf_idx]:flush()
        return
    end
    template = self:template({output = false}, template, ...)
    self.bufs[self.xml_buf_idx]:append(template)
end

-- Flushes remaining data to the SIE file and closes it.
-- If a "reserve space" file exists it is deleted.
function m:close()
    -- If there is a reserve space file, remove it
    if self.space_reserve_file then C.unlink(self.space_reserve_file) end

    -- If the filesystem is full and there are registered buffers that need to
    -- be flushed (time at level mostly), remove the space saver file, turn off
    -- the full marker, and flush stuff!
    local disk_full = C_writev ~= C.writev

    -- One or more entities have registered for space to be reserved
    if self.oos_info then
        -- Make sure writes actually can happen
        C_writev = C.writev

        -- Flush the xml
        self.bufs[self.xml_buf_idx]:flush()

        -- Call the function registered for entity that wanted space reserved
        for _, v in ipairs(self.oos_info) do
            v.callback()
        end

        -- If the disk is NOT full, flush all the buffers
        if not disk_full then
            -- If disk is not full flush all buffers
            for i = 2, self.last_buf do
                self.bufs[i]:flush()
            end

            -- Write the index
            self.bufs[self.index_buf_idx]:flush()
        end

    -- If the disk in not full, flush everything
    elseif not disk_full then
        self.bufs[self.xml_buf_idx]:flush()
        for i = 2, self.last_buf do
            self.bufs[i]:flush()
        end
        self.bufs[self.index_buf_idx]:flush()
    end

    assert(C.close(self.fd) == 0)
end

-- Creates the SIE XML for a test.
-- Params:
--   test_id - id for the test
--   run_number - number of the run of the test
--   run_description - test description string
--   start_time - start time of the test run
function m:test_header(test_id, run_number, run_description, start_time)
    self:xml(stempls.test_header, {
        test_id = test_id or 0,
        run_number = run_number or 1,
        run_description = run_description or 'None',
        start_time = start_time or '0000-01-01T00:00:00-0000',
    })
end

-- Used to create and write the SIE XML for a "collected" channel.
-- This may include writing SIE XML for a decoder and/or a SIE
-- base channel.
-- Params:
--   item - a clause (stanza) from the test's json setup text
--       which specifies the channel
-- Returns:  the created meta data object
function m:collect_channel(stanza)
    stanza.description = stanza.description or ""
    stanza.datamode_name = 'dm'  -- KLUDGE!!!
    stanza.sie_name = stanza.sie_name or stanza.name
    stanza.somat_input_channel = stanza.somat_input_channel or stanza.name
    stanza.nbr_data_bits = tonumber(stanza.data_type:match('^.-(%d+)$'))
    if not stanza.nbr_data_bits then
        -- KLUDGE the following is actually number of octets in this case
        stanza.nbr_data_bits = tutil.sample_size(stanza.data_type)
    end
    assert(stanza.nbr_data_bits and stanza.nbr_data_bits > 0,
        "Missing data bit size")
    local schema_name = sie_schema.get_collect_schema_name(stanza)
    local schema_info = sie_schema.get_schema_info(schema_name, stanza)
    local md = sutil.write_sie_channel(self, schema_info, stanza)
    return md
end

-- Creates the XML for an SIE file tag containing a "group"
-- attribute, with the tag followed in the file by associated data.
-- Params:
--   id - string identifying the tag
--   format - string indentifying a format for the group
--   data - data to be written after the tag
--   len - if not false or nil, the length of data to be written
--       after the tag; otherwise the computed length of 'data'
--       will be used
--   indent - how many spaces to indent the tag text in the
--       SIE file
--   group - if not nil or false, the desired group number;
--       otherwise a group number is allocated
-- Returns:  group number used
function m:group_tag(id, format, data, len, indent, group)
    local gtb = self.bufs[self.scratch_buf_idx]
    indent = indent or ' '
    group = group or self:next_id('group')
    gtb.header.group = htonl(group)

    self:xml(stempls.grp_tag, {
        i = indent,
        id = id,
        grp = group,
        fmt = format,
    })
    len = len or #data

    gtb.payload:clear()
    self.bufs[self.xml_buf_idx]:flush()

    gtb.payload:expand(len)
    gtb:enq(data, len)
    gtb:flush()

    return group
end

-- Creates an SIE file tag followed by associated data, which is
-- specified by the given file name.  Note that the tag's group
-- number is allocated by the function.
-- Params:
--   id - string identifying the tag
--   format - string indentifying a format for the tag's group
--   filename - where the data to be written after the tag
--       is located.
function m:group_tag_from_file(id, format, filename)
    local content = ''
    local fh = io.open(filename)
    if fh then
        content = fh:read('*a')
        fh:close()
    end
    self:group_tag(id, format, content)
end

return m
